%{

/* Qbib/b2q.y
 * Bison/yacc grammar file for the Qbib Bibtex parser
 *
 * Copyright (C) 1997-2001 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This file is part of Qbib.
 */

#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>

int yylex();
extern int saw_equals;
static char *terminator="\f";
%}

%pure_parser

%token STRING CSTRING
%token ALIAS VERBOSENAME
%token LPAREN RPAREN ASTERISK COMMA DOUBLEQUOTE SINGLEQUOTE EQUALS
%token NUMBER STRINGSTART RBRACE LBRACE ATSIGN

%type <String>		STRING CSTRING

%union {
    char	*String;
    double	Real;
    int		Integer;
};

%%

FileOfEntries:
    OneEntry {
    } |
    OneEntry FileOfEntries {
    };

OneEntry:
    STRINGSTART STRING EQUALS CSTRING RBRACE {
	saw_equals = 0;
	printf("{@string %s {%s}}\n", $2, $4);
	printf("%s", terminator);
    } |
    ATSIGN STRING LPAREN STRING COMMA ListOfSubEntries RPAREN {
	printf("%s %s\n", $2, $4);
	printf("%s", terminator);
    } |
    ATSIGN STRING LBRACE STRING COMMA ListOfSubEntries RBRACE {
	printf("%s %s\n", $2, $4);
	printf("%s", terminator);
    };

SubEntry:
    STRING EQUALS CSTRING {
	saw_equals = 0;
	printf("{%s {%s}}\n", $1, $3);
    } |
    STRING EQUALS STRING {
	saw_equals = 0;
	printf("{%s {%s}}\n", $1, $3);
    };

ListOfSubEntries:
    SubEntry {
    } |
    SubEntry COMMA {
    } |
    SubEntry COMMA ListOfSubEntries {
    };
%%

main(argc, argv)
    int        argc;
    char       *argv[];
{
    if (argc != 3) {
	fprintf(stderr, "Usage: %s <input file> <output file>\n", argv[0]);
	exit(1);
    }
    close(0);
    open(argv[1], O_RDONLY);
    close(1);
    open(argv[2], O_WRONLY|O_CREAT|O_TRUNC, 0666);
    yyparse();
    fflush(stdout);
    close(1);
    exit(0);
}

