%{

/* b2qlex.l
 * Uses 'flex' fast lexical analyzer.  Not compatible with lex at all!
 *
 * Copyright (C) 1997-2001 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This file is part of Qbib.
 */

#include <stdio.h>
#include "b2q.tab.h"

/* Redefine the YY_DECL from yylex() to the needed yylex() with parameters
 * for bison's pure_parser directive.  This is very flex specific code,
 * so don't try to use lex with it!!
 */

int lbrace_state = 0;
int saw_equals = 0;

#ifdef	YY_DECL
#	undef 		YY_DECL
#endif
#define YY_DECL		int yylex(yylvalp) \
                                  YYSTYPE		*yylvalp;
#ifdef yywrap
#	undef		yywrap
#endif

#if defined(YY_INPUT)
#undef YY_INPUT
#endif
#define YY_INPUT(buf,result,max_size) \
{ \
int c = getchar(); \
result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \
}

int yywrap();

int LineNumber = 1;
%}
a	[aA]
b	[bB]
c	[cC]
d	[dD]
e	[eE]
f	[fF]
g	[gG]
h	[hH]
i	[iI]
j	[jJ]
k	[kK]
l	[lL]
m	[mM]
n	[nN]
o	[oO]
p	[pP]
q	[qQ]
r	[rR]
s	[sS]
t	[tT]
u	[uU]
v	[vV]
w	[wW]
x	[xX]
y	[yY]
z	[zZ]
%%

\@{s}{t}{r}{i}{n}{g}\{          return STRINGSTART;
\}                              {
                                        return RBRACE;
                                }
\@                              return ATSIGN;

[a-zA-Z0-9\_\-:;\.\*\!\$\#\^\|\~`\']*          {
					yylvalp->String = malloc((size_t)yyleng+1);
					strcpy(yylvalp->String, yytext);
					return STRING;
				}


\"\"				{
					yylvalp->String = malloc(1);
					yylvalp->String[0] = '\0';
					return CSTRING;
				}

\"				{
                                        char	*Buf;
					int	i, Index;
                                        size_t  bufsiz;
					
					Buf = malloc(BUFSIZ);
					bufsiz = BUFSIZ;
					Index = 0;
RestartDoubleQuote:
					while ((i=input()) != '"') {
					    if (i == '\f' || i == '\r') continue;
					    if (i == EOF) {
						REJECT;
					    } else {
						if (Index >= bufsiz-3)
						    Buf = realloc(Buf, bufsiz+=BUFSIZ);
						Buf[Index] = i;
						Index++;
					    }
					}
					Buf[Index] = '\0';
					if (Index != 0 && Buf[Index-1] == '\\') {
					    Buf[Index-1] = '"';
					    goto RestartDoubleQuote;
					}
					Buf = realloc(Buf, (size_t)Index+1);
					yylvalp->String = Buf;
					return CSTRING;
				}
\{				{
                                        char	*Buf;
					int	i, Index, level;
                                        size_t  bufsiz;
					
					if (saw_equals == 0)
					    return LBRACE;
					Buf = malloc(BUFSIZ);
					bufsiz = BUFSIZ;
					Index = 0;
					level = 1;
RestartBrace:
					while ((i=input()) != '}') {
					    if (i == '\f' || i == '\r') continue;
					    if (i == '{') {
					        level++;
					    }
					    if (i == EOF) {
						REJECT;
					    } else {
						if (Index >= bufsiz-3)
						    Buf = realloc(Buf, bufsiz+=BUFSIZ);
						Buf[Index] = i;
						Index++;
					    }
					}
					level--;
					Buf[Index] = '\0';
					if (level > 0) {
					    if (Index >= bufsiz-3)
 					        Buf = realloc(Buf, bufsiz+=BUFSIZ);
					    Buf[Index] = '}';
					    Index++;					    
					    goto RestartBrace;
					}
					if (Index != 0 && Buf[Index-1] == '\\') {
					    Buf[Index-1] = '}';
					    goto RestartBrace;
					}
					Buf = realloc(Buf, (size_t)Index+1);
					yylvalp->String = Buf;
					return CSTRING;
				}
\%.*                            ;
"("				return LPAREN;
")"				return RPAREN;
"*"				return ASTERISK;
","				return COMMA;
"\""				return DOUBLEQUOTE;
"'"				return SINGLEQUOTE;
"="				{saw_equals = 1; return EQUALS;}
\n				LineNumber++;
\f				;
[\ \t\b\r;]+			;

%%

/* yywrap() - Flex define of yywrap() = 1 is overridden here.  
 */
int yywrap()
{
    LineNumber = 1;
    return 1;
}
int yyerror(str)
    char *str;
{
    fprintf(stderr, "parse error at line %d\n", LineNumber);
    exit(1);
}

#undef yyunput
#undef yy_push_state
#undef yy_pop_state
#undef yy_top_state
