
/* $RCSfile: Qddb_Schema.h,v $ $Revision: 2.2 $ $Date: 2003/03/23 17:16:14 $
 *
 * Copyright (C) 1997-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#if !defined(_QDDB_SCHEMA_H_)
#define _QDDB_SCHEMA_H_

#define MAXIMUM_SCHEMA_ENTRY		4096
#define MAXIMUM_NUMBER_OF_SCHEMA_LEVELS	16

/* Qddb_RebuildSchema() args
 */
#define QDDB_REBUILDSCHEMA_FROMNODES 0
#define QDDB_REBUILDSCHEMA_FROMTREE  1

#define QDDB_DEFAULT_SEPARATORS "\t \n\r!@#$%^&*()_+-={}[];':\",./<>?`~\\|"
#define QDDB_DEFAULT_REAL_SEPARATORS "\t \n\r!@#$%^&*()_={}[];':\",/<>?`~\\|"
#define QDDB_DEFAULT_INTEGER_SEPARATORS "\t \n\r!@#$%^&*()_={}[];':\",./<>?`~\\|"
#define QDDB_DEFAULT_DATE_SEPARATORS "\t \n\r!@#$%^&*()_={}[];':\",/<>?`~\\|"
#define QDDB_DEFAULT_TIME_SEPARATORS "\t \n\r!@#$%^&*()_={}[];'.\",/<>?`~\\|"

/* internal to the schema parser
 */
typedef struct qddb_attroption {
  int 		         option;
  char		         *value;
  char                   *value2;
  struct qddb_attroption *next;
} Qddb_AttrOption; 

/* internal and external to the parser 
 */

typedef enum {
  QAT_Typeless = 0,
  QAT_Real     = 1,
  QAT_Integer  = 2,
  QAT_String   = 3,
  QAT_Date     = 4,
  QAT_Time     = 5
} Qddb_AttributeType;

typedef enum {
  QAO_None          = 0x0000, /* None defined in X11R6.3 as 0L */
  QAO_Autoincrement = 0x0001
} Qddb_AttributeOptions; 

typedef struct qddb_schemanode {
    char		      *Name;
    char		      *VerboseName;
    char		      *Alias;
    char		      *Format;
    char		      *Separators;
    char		      *DefaultValue;
#if !defined(USE_TCL)
    char		      *Attribute;
#endif
    int                       AutoStart, AutoIncrement;
    Qddb_Boolean	      Exclude;
    Qddb_AttributeType	      Type;
    Qddb_AttributeOptions     Options;
    Qddb_Boolean	      AsteriskPresent;
    size_t                    Number;
    size_t                    AncestorNumber[MAXIMUM_NUMBER_OF_SCHEMA_LEVELS];
    size_t                    Level;
    Qddb_Boolean	      IsLeaf;
} Qddb_SchemaNode;

typedef struct qddb_schematreenode {
    Qddb_SchemaNode		*schemanode;
    struct qddb_schematreenode	**children;
} Qddb_SchemaTreeNode;

typedef char *Qddb_ReducedAttrTable;

typedef struct {
    char                  *RelationName;
    Qddb_SchemaNode       *Entries;
    Qddb_SchemaTreeNode   **Tree;
    Tcl_HashTable	  TclHashTable;
    Qddb_WordList         *word_list;
    size_t                word_list_size, word_list_number_words;
    int                   word_list_version;
    int                   HashType;
    size_t                NumberOfAttributes;
    int                   hashtable_fd, database_fd;
    Qddb_Hash             *HashTable;
    Qddb_Boolean	  UseReducedAttributeIdentifiers, UseCachedHashing;
    Qddb_Boolean          UseCachedSecondarySearch, UseExcludeWords, UseCondensedIndexing;
    Qddb_Boolean          UseLargeDatabaseOptimizations, UseProximityIndexing;
    Qddb_ReducedAttrTable *ReducedAttributeTable;
    size_t	  	  TopOfReducedAttributeTable, HashSize, CacheSize, TopOfCache;
    char		  *default_date_format;
} Qddb_Schema;

Qddb_Schema 	*Qddb_InitSchema _ANSI_ARGS_((char *));
Qddb_Schema 	*Qddb_RebuildSchema _ANSI_ARGS_((Qddb_Schema *, int));

void            Qddb_EntryParse_Set _ANSI_ARGS_((Qddb_Schema *));
Qddb_InCoreEntry *Qddb_EntryParse_Get _ANSI_ARGS_((void));
void            Qddb_ResetEntryParser _ANSI_ARGS_((void));
int             Qddb_EntryParse _ANSI_ARGS_((void));

void   		Qddb_ResetSchemaParser _ANSI_ARGS_((void));
void		Qddb_ResetSchemaLex _ANSI_ARGS_((void));
int    		Qddb_SchemaParse _ANSI_ARGS_((void));

void 		Qddb_SchemaParse_Set _ANSI_ARGS_((Qddb_Schema *));
Qddb_Schema 	*Qddb_SchemaParse_Get _ANSI_ARGS_((void));

void		Qddb_InitHash _ANSI_ARGS_((Qddb_Schema *));
void		Qddb_MakeHash _ANSI_ARGS_((Qddb_Schema *, char *));
int		Qddb_GetHashTableEntry _ANSI_ARGS_((Qddb_Schema *, size_t));

/* from Lib/LibQddb/KeyList.c 
 */
Qddb_KeyList *Qddb_KeyListOp _ANSI_ARGS_((Qddb_Schema *, Qddb_KeyList *, Qddb_KeyList *, int, int));
void *Qddb_KeyListProcess _ANSI_ARGS_((Qddb_Schema *, Qddb_KeyList *, CONST void *, int, int));
/* fast KeyList concat */
Qddb_KeyList *Qddb_KeyListConcat _ANSI_ARGS_((Qddb_KeyList *, Qddb_KeyList *, Qddb_KeyList **));

/* from Lib/LibQddb/WordList.c
 */
void Qddb_InitWordList _ANSI_ARGS_((Qddb_Schema *));
void Qddb_UpdateWordList _ANSI_ARGS_((Qddb_Schema *, Qddb_KeyListType, int, Qddb_Entry));
#endif

