
/* $RCSfile: Qddb_Sql.h,v $ $Revision: 2.0.0.1 $ $Date: 2000/11/09 15:05:57 $
 *
 * Copyright (C) 1993-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#if !defined(_QDDB_SQL_H_)
#define _QDDB_SQL_H_

#define MAXOPENRELATIONS	32
#define MAXALIASES		16
#define MAXALIASLEN		16

typedef void SqlReturn;

typedef struct {
    char		PathName[MAXPATHLEN];
    char		Name[MAXFILENAMELEN];
    Qddb_Hash		*HashTable;
    unsigned int	HashSize;
    Qddb_Schema		*Schema;
    char		**Aliases;
    Qddb_Boolean	Valid;
} Qddb_SqlRelation;

typedef struct {
    Qddb_KeyList	KeyList;
    Qddb_SqlRelation	Relation;
} Qddb_SqlKeyList;


/* structures internal to the SQL parser (and interpreter in general).
 */

/* Constants */
union qddb_constant {
    char		*String;
    int			Integer;
    double		Real;
};

typedef enum {
    STRINGCONS, INTEGERCONS, REALCONS
} ConstantType;


typedef struct {
    union constant	Val;
    ConstantType	Type;
} Constant;

/* Identifiers */
typedef struct {
    char		*TableName;
    char		*ColumnName;
    unsigned int	Column;
} Identifier;

/* CommaList: a list of integers for various uses in the interpreter */
typedef struct clistnode {
    unsigned int	Index;
    unsigned int	Next;
} ClistNode;

typedef unsigned int	Clist;

/* OpStack: a list of operations and values to perform.
 */
typedef enum {
    LESSTHAN, LESSEQUAL, 
    GREATERTHAN, GREATEREQUAL, 
    EQUAL, NOTEQUAL,
    SUBTRACT, ADD,
    MULTIPLY, DIVIDE,
    INTERSECTIONOP, UNIONOP,
    NONE
} Operation;

typedef enum {
    OPCONSTANT,
    OPIDENT,
    OPSTACK,
    OPKEYLIST
} PrimaryType;

typedef struct {
    PrimaryType		Type;
    unsigned int	Index;
} Primary;

typedef struct {
    Operation		Op;
    Primary		Val;
} OpStack;

typedef struct {
    unsigned int	Top, Bottom;
} OpStackList;

/* For select statements */
typedef struct {
    unsigned int	Select, From, Where;
} TabExp;

/* NIL is a value that cannot be an index into one of the QSQL internal
 * arrays (2^31-1)
 */
#define NIL		(2147483647)

#define CONSTANTSIZE	100
#define STRINGSIZE	50000
#define IDENTIFIERSIZE	1000
#define CLISTSSIZE	10000
#define OPSSIZE		100
#define OPSTACKLISTSIZE	100
#define TABEXPSIZE	20

#endif

