
/* $RCSfile: tclQddb.h,v $ $Revision: 2.1 $ $Date: 2003/03/23 17:16:14 $
 *
 * Copyright (C) 1997-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#if !defined(_TCLQDDB_H_)
#define _TCLQDDB_H_

typedef struct {
    char		*schema_name;
    int			fd;
    Qddb_KeyList	keylist;
    Qddb_DataTree	**datatree;
} TclQddb_Tuple;

typedef struct tclqddb_rows {
    char		*attr_name, *instance;
    struct tclqddb_rows *next;
} TclQddb_Rows;

typedef struct tclqddb_rowheader {
    char		*schema_name;
    char		*tuple_name;
    TclQddb_Rows	*row;
} TclQddb_RowHeader;

typedef struct tclqddb_keylistheader {
    char		*schema_name;
    Qddb_AttrHead       attributes;
    Qddb_KeyList	*keylist;
} TclQddb_KeyListHeader;

typedef struct {
    char		*schema_name;
    char		*tuple_name;
    int			num_attrs;
    /* view is an array of pointers to datatree elements.
     * view[i] is the view corresponding to the i'th node
     * in the schema.   Leaf nodes point directly into
     * the DataTree, intermediate nodes have a copy
     * of the DataTree node.
     */
    Qddb_DataTree	**view;
    /* tclvars[i] holds the name of the TCL variable linked
     * to view[i] (leaves only).
     */
    char		**tclvars;
    /* curinst[i] holds the current instance for the schema
     * entry indexed by i.
     */
    int			*curinst;
} TclQddb_View;

int TclQddb_Init 	_ANSI_ARGS_((Tcl_Interp *));

/* Main entry points */
#if TCL_MAJOR_VERSION > 8 && TCL_MINOR_VERSION >= 4
int TclQddb_Export   	   _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **));
int TclQddb_Import   	   _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **));
int TclQddb_SchemaProc 	   _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **));
int TclQddb_Search   	   _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **));
int TclQddb_KeyList  	   _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **));
int TclQddb_ScriptProc     _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **));
int TclQddb_RowsProc  	   _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **));
int TclQddb_TableProc 	   _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **));
int TclQddb_TupleProc 	   _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **));
int TclQddb_Instance 	   _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **));
int TclQddb_ViewProc   	   _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **));
int TclQddb_UtilProc   	   _ANSI_ARGS_((ClientData, Tcl_Interp *, int, char **));
#else
Tcl_CmdProc                TclQddb_Export;
Tcl_CmdProc                TclQddb_Import;
Tcl_CmdProc                TclQddb_SchemaProc;
Tcl_CmdProc                TclQddb_Search;
Tcl_CmdProc                TclQddb_KeyList;
Tcl_CmdProc                TclQddb_ScriptProc;
Tcl_CmdProc                TclQddb_RowsProc;
Tcl_CmdProc                TclQddb_TableProc;
Tcl_CmdProc                TclQddb_TupleProc;
Tcl_CmdProc                TclQddb_Instance;
Tcl_CmdProc                TclQddb_ViewProc;
Tcl_CmdProc                TclQddb_UtilProc;
#endif


/* Internal cross-referencing routines */
Qddb_Schema 	*TclQddb_GetSchema 	        _ANSI_ARGS_((CONST char *));
void 		TclQddb_DeleteSchemaProc        _ANSI_ARGS_((ClientData));
char            *TclQddb_TraverseSchemaTree     _ANSI_ARGS_((Tcl_Interp *, Qddb_SchemaTreeNode **));

int 		TclQddb_NewKeyList	        _ANSI_ARGS_((Tcl_Interp *, CONST char *, \
							     Qddb_KeyList *, int, int));
int 		TclQddb_GetKeyList 	        _ANSI_ARGS_((Tcl_Interp *, CONST char *, \
							     TclQddb_KeyListHeader **));
int		TclQddb_DeleteKeyList	        _ANSI_ARGS_((Tcl_Interp *, CONST char *, int));
void 		TclQddb_DeleteKeyListProc       _ANSI_ARGS_((ClientData));

TclQddb_RowHeader *TclQddb_GetRows	        _ANSI_ARGS_((Tcl_Interp *, CONST char *));
int 		TclQddb_SetRow		        _ANSI_ARGS_((Tcl_Interp *, TclQddb_RowHeader *));
int		TclQddb_DeleteRows	        _ANSI_ARGS_((Tcl_Interp *, CONST char *));
void 		TclQddb_DeleteRowProc 	        _ANSI_ARGS_((ClientData));
void 		TclQddb_DeleteTableProc         _ANSI_ARGS_((ClientData));
char		*TclQddb_NewTuple	        _ANSI_ARGS_((Tcl_Interp *, CONST char *, \
							     Qddb_DataTree **, Qddb_KeyList *));
TclQddb_Tuple 	*TclQddb_GetTuple	        _ANSI_ARGS_((Tcl_Interp *, CONST char *));
int		TclQddb_DeleteTuple	        _ANSI_ARGS_((Tcl_Interp *, CONST char *));
void 		TclQddb_DeleteTupleProc         _ANSI_ARGS_((ClientData));
int 		TclQddb_FlushTuple	        _ANSI_ARGS_((Tcl_Interp *, CONST char *));

TclQddb_View	*TclQddb_GetView 		_ANSI_ARGS_((Tcl_Interp *, CONST char *));
int             TclQddb_SetInstance 		_ANSI_ARGS_((Tcl_Interp *, Qddb_Schema *, \
							     TclQddb_Tuple *, TclQddb_View *, \
							     CONST char *, size_t));
int		TclQddb_ViewLockTuple  		_ANSI_ARGS_((Tcl_Interp *, CONST char *));
int		TclQddb_ViewUnlockTuple  	_ANSI_ARGS_((Tcl_Interp *, CONST char *));
int		TclQddb_DeleteView		_ANSI_ARGS_((Tcl_Interp *, CONST char *));
void 		TclQddb_DeleteViewProc 		_ANSI_ARGS_((ClientData));

void            TclQddb_SortRowsSameTuple 	_ANSI_ARGS_((Tcl_Interp *, TclQddb_Tuple *, \
							     TclQddb_Rows **, int, \
							     CONST char **, int, CONST char **));
int             TclQddb_SortRows 		_ANSI_ARGS_((Tcl_Interp *, int, char **, int, \
							     CONST char **, int, CONST char **));
int             TclQddb_SortSelectRows 	        _ANSI_ARGS_((Tcl_Interp *, int, CONST char **, CONST char **, \
							     int, CONST char **, int, CONST char **));
int             TclQddb_SortQueryRows 		_ANSI_ARGS_((Tcl_Interp *, Qddb_Schema *, int, CONST char **, \
							     int, CONST char **, int, CONST char **));
Qddb_DataTree   *TclQddb_DataTreeLookupNode 	_ANSI_ARGS_((Tcl_Interp *, TclQddb_Tuple *, char *, char *));
char            *TclQddb_NewTable 		_ANSI_ARGS_((Tcl_Interp *, Qddb_Table *));
int             TclQddb_TableSetHeaderInfo 	_ANSI_ARGS_((Tcl_Interp *, Qddb_TableHeaderInfo **, \
							     CONST char *, CONST char *, int));
char            *TclQddb_TableGetHeaderInfo     _ANSI_ARGS_((Qddb_TableHeaderInfo *, \
							     Qddb_TableHeaderInfo *, CONST char *));

int             TclQddb_TableConfigureType 	_ANSI_ARGS_((Tcl_Interp *, Qddb_TableNode *, int, CONST char *));
int             TclQddb_TableConfigureColType 	_ANSI_ARGS_((Tcl_Interp *, Qddb_TableNode *, int, CONST char *));
int             TclQddb_TableSort 		_ANSI_ARGS_((Tcl_Interp *, Qddb_Table *, CONST char *, \
							     CONST char *, CONST char **, int, \
							     CONST char **, int));
Qddb_Table      *TclQddb_GetTable 		_ANSI_ARGS_((Tcl_Interp *, CONST char *));
Qddb_Table      *TclQddb_GetTableByName		_ANSI_ARGS_((Tcl_Interp *, char *));
char            *TclQddb_GetCellValue 		_ANSI_ARGS_((Qddb_TableNode *, int, \
							     Qddb_TableHeaderFormat *));
int             TclQddb_TableDelete 		_ANSI_ARGS_((Tcl_Interp *, CONST char *));
char            **TclQddb_GetFormattedTable 	_ANSI_ARGS_((Tcl_Interp *, Qddb_Table *, CONST char *, int, int));

void		TclQddb_FreeArgs		_ANSI_ARGS_((int, char **));
int		TclQddb_FindAttribute   	_ANSI_ARGS_((Qddb_DataTree **, char *));
char		**TclQddb_SplitAttributeName 	_ANSI_ARGS_((CONST char *));
char		*TclQddb_ConcatAttributeName 	_ANSI_ARGS_((char **, int));
char		**TclQddb_ParseAttr		_ANSI_ARGS_((char *));
int		TclQddb_CountAttrs		_ANSI_ARGS_((char *));
char 		*TclQddb_TreeDataToString	_ANSI_ARGS_((Qddb_DataTree *));
char		*TclQddb_StringRef		_ANSI_ARGS_((TclQddb_Tuple *, TclQddb_Rows *));
Qddb_DataTree 	*TclQddb_DataTreeRef		_ANSI_ARGS_((TclQddb_Tuple *, TclQddb_Rows *));
int		TclQddb_LinkVar			_ANSI_ARGS_((Tcl_Interp *, char *, char *, int, \
							     char *, Qddb_DataTree *));
void		TclQddb_UnlinkVar		_ANSI_ARGS_((Tcl_Interp *, char *));
void            TclQddb_UpdateLinkedVar         _ANSI_ARGS_((Tcl_Interp *, char *));
void 		TclQddb_NullOutNode		_ANSI_ARGS_((Qddb_DataTree *, int));
Qddb_Boolean 	TclQddb_IsEmptyDataTree		_ANSI_ARGS_((Qddb_DataTree **));
Qddb_DataTree 	*TclQddb_StringToTreeData	_ANSI_ARGS_((Qddb_SchemaTreeNode *, char *));
int 		TclQddb_TableExprInterp 	_ANSI_ARGS_((Tcl_Interp *, Qddb_TableCalc *, \
							     Qddb_Table *, size_t, size_t, \
							     Qddb_TableDependencies **, double *));
#endif


