#!QDDB_BIN_DIR/qtclsh
# $RCSfile: qddbscript.tcl,v $ $Revision: 2.0.0.1 $ $Date: 2000/11/09 15:05:59 $
#
# Copyright (C) 1997-2000 Herrin Software Development, Inc.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# Usage:
#    qddbscript autogen <Relation> ?output_file?
#    qddbscript compile ?-compat? <Relation> ?output_file?

proc Usage {} {
    global argv0

    puts stderr "Usage: [file tail $argv0] autogen <relation_dir> ?output_file?"
    puts stderr "       [file tail $argv0] compile ?-compat? <relation_dir> ?output_file?"
    puts stderr "       [file tail $argv0] compile ?-compat? ?<script_file>? ?<output_file>?"
    exit 1
}

set type 0
switch -exact [lindex $argv 0] {
    "autogen" {
	set type 1
    }
    "compile" {
	set type 2
    }
}
if {$type == 0} {
    Usage
}
if {$type == 1} {
    if {[catch "qddb_schema open [lindex $argv 1]" s] != 0} {
	Usage
    }
    set output [qddb_script autogen $s]
    if {[string length [lindex $argv 2]] == 0} {
	set path [qddb_schema path $s]/QddbScript
    } else {
	set path [lindex $argv 2]
    }
    set fd [open $path w]
    if {$fd == -1} {Usage}
    puts $fd $output
    close $fd
} else {
    set compat 0
    if {[string compare [lindex $argv 1] "-compat"] == 0} {
	set nexti 2
	set compat 1
    } else {
	set nexti 1
    }
    set file [lindex $argv $nexti]
    incr nexti
    if {[catch "qddb_schema open $file" s] != 0} {
	;# Compiling from a file
	puts stderr "Warning: \[options autoalign\] will be ignored when compiling from file..."
	set fd [open $file r]
	if {$fd == -1} {Usage}
	set script [read $fd]
	close $fd
	set script [qddb_script compile "" $script]
	if {[string length [lindex $argv $nexti]] == 0} {
	    set file ${file}.tcl
	} else {
	    set file [lindex $argv $nexti]
	}
	set fd [open $file w]
	if {$fd == -1} {Usage}
	puts $fd {#!QDDB_BIN_DIR/qwish}
	puts $fd {lappend auto_path $qddb_library/fx}
	if {$compat} {
	    set fd2 [open $qddb_library/fx/fx_script.tcl]
	    puts $fd [read $fd2]
	    close $fd2
	}
	puts $fd $script
	close $fd
    } else {
	;# Compiling from a relation
	set file "[qddb_schema path $s]/QddbScript"
	set fd [open $file r]
	if {$fd == -1} {Usage}
	set script [read $fd]
	close $fd
	set qddbscript(schema) $s
	if {[catch [list qddb_script compile qddbscript(schema) $script] script]} {
	    puts stderr $script
	    Usage
	}
	if {[string length [lindex $argv $nexti]] == 0} {
	    set file ${file}.tcl
	} else {
	    set file [lindex $argv $nexti]
	}
	set fd [open $file w]
	if {$fd == -1} {Usage}
	puts $fd {#!QDDB_BIN_DIR/qwish}
	puts $fd {lappend auto_path $qddb_library/fx}
	if {$compat} {
	    set fd2 [open $qddb_library/fx/fx_script.tcl]
	    puts $fd [read $fd2]
	    close $fd2
	}
	puts $fd $script
	close $fd
    }
}

