
/* $RCSfile: TclExport.c,v $ $Revision: 2.2 $ $Date: 2003/03/23 17:16:16 $
 * 
 * Copyright (C) 1995-2003 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include "tcl.h"
#include "Qddb.h"
#include "tclQddb.h"

/* qddb_export csv ?options? <table_desc> ?<table_desc> ...?
 * qddb_export as ?options? <table_desc> ?<table_desc> ...?
 *
 * Options:
 *    -columntitles on|off (default: off)
 *    -concatsheets on|off (default: off)
 *    -rowtitles on|off    (default: off)
 *    -dateformat "%m/%d/%y"|days1900 etc.
 *    -separator ","       (default: ",")
 */


static int ExportCSV _ANSI_ARGS_((Tcl_Interp *, int, CONST char **));
static int ExportAS _ANSI_ARGS_((Tcl_Interp *, int, CONST char **));
static int ParseArgs _ANSI_ARGS_((Tcl_Interp *, int, CONST char **));
static char *ASColumnName _ANSI_ARGS_((int));
static int ASDateFormat _ANSI_ARGS_((void));

int TclQddb_Export(clientData, interp, argc, argv)
    ClientData			clientData;
    Tcl_Interp			*interp;
    int				argc;
    CONST char			*argv[];
{
    if (argc < 3) {
	Tcl_AppendResult(interp, argv[0], ": wrong # args", NULL);
	return TCL_ERROR;
    }
    if (ParseArgs(interp, argc, argv) != TCL_OK)
	return TCL_ERROR;
    if (strcmp(argv[1], "csv") == 0) {
	if (ExportCSV(interp, argc, argv) != TCL_OK)
	    return TCL_ERROR;
    } else if (strcmp(argv[1], "as") == 0) {
	if (ExportAS(interp, argc, argv) != TCL_OK)
	    return TCL_ERROR;
    } else {
	Tcl_AppendResult(interp, argv[0], ": bad option \"", argv[1], "\"", NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

static int rowtitles = 0, columntitles = 0, concatsheets = 0;
static char *separator, *dateformat;

static int ParseArgs(interp, argc, argv)
    Tcl_Interp                  *interp;
    int                         argc;
    CONST char                  **argv;
{
    int                         i;

    rowtitles = columntitles = concatsheets = 0;
    dateformat = qddb_default_date_format;
    separator = ",";
    for (i = 2; i < argc; i+=2) {
	if (argv[i][0] == '-') {
	    if (i == argc-1) {
		Tcl_AppendResult(interp, argv[0], " ", argv[1], ": wrong # args", NULL);
		return TCL_ERROR;
	    }
	    if (strcmp(argv[i], "-rowtitles") == 0) {
		if (Tcl_GetBoolean(interp, argv[i+1], &rowtitles) != TCL_OK) {
		    Tcl_AppendResult(interp, " ", argv[i],
				     " should be followed by on, off, 1, 0, true, or false", NULL);
		    return TCL_ERROR;
		}
	    } else if (strcmp(argv[i], "-columntitles") == 0) {
		if (Tcl_GetBoolean(interp, argv[i+1], &columntitles) != TCL_OK) {
		    Tcl_AppendResult(interp, " ", argv[i],
				     " should be followed by on, off, 1, 0, true, or false", NULL);
		    return TCL_ERROR;
		}
	    } else if (strcmp(argv[i], "-dateformat") == 0) {
		dateformat = (char *)Malloc(strlen(argv[i+1])+1);
		strcpy(dateformat, argv[i+1]);
	    } else if (strcmp(argv[i], "-separator") == 0) {
		separator = (char *)Malloc(strlen(argv[i+1])+1);
		strcpy(separator, argv[i+1]);
	    } else if (strcmp(argv[i], "-concatsheets") == 0) {
		if (Tcl_GetBoolean(interp, argv[i+1], &concatsheets) != TCL_OK) {
		    Tcl_AppendResult(interp, " ", argv[i],
				     " should be followed by on, off, 1, 0, true, or false", NULL);
		    return TCL_ERROR;
		}		
	    } else {
		Tcl_AppendResult(interp, argv[0], " ", argv[1], ": bad option ", argv[i], NULL);
		return TCL_ERROR;
	    }
	}
    }
    return TCL_OK;
}

static int ExportCSV(interp, argc, argv)
    Tcl_Interp                  *interp;
    int                         argc;
    CONST char                  *argv[];
{
    Qddb_Table                  *table;
    Qddb_TableNode              *nodes, *row0;
    Qddb_TableHeaderFormat      *format;
    int                         i, j, first_table;
    CONST char                  *table_desc;
    char                        *buf, *time, *date, *val;
    
    for (first_table = argc-1; first_table > 0; first_table--) {
	if (strncmp(argv[first_table], "qddb_table", 10) != 0)
	    break;
    }
    first_table++;
    for ( ; first_table < argc; first_table++) {
	table_desc = argv[first_table];
	if ((table = TclQddb_GetTable(interp, table_desc)) == NULL) {
	    return TCL_ERROR;
	}
	if (table->columns < 1) {
	    Tcl_AppendResult(interp, argv[0], " ", argv[1], ": no columns defined", NULL);
	    return TCL_ERROR;
	}
	Qddb_InitBuffer();
	row0 = table->table[0];
	if (columntitles != 0) {
	    Qddb_ConcatBuffer("\"");
	    val = TclQddb_TableGetHeaderInfo(row0[1].info, row0[1].info, "-title");
	    if (val != NULL) {
		Qddb_ConcatBuffer(val);
	    }
	    Qddb_ConcatBuffer("\"");
	    for (i = 2; i <= table->columns; i++) {
		Qddb_ConcatBuffer(",\"");
		val = TclQddb_TableGetHeaderInfo(row0[i].info, row0[i].info, "-title");
		if (val != NULL) {
		    Qddb_ConcatEscapeBuffer(val, '\"');
		}
		Qddb_ConcatBuffer("\"");
	    }
	    Qddb_ConcatBuffer("\n");
	}
	for (i = 1; i <= table->rows; i++) {
	    nodes = table->table[i];
	    for (j = 1; j <= table->columns; j++) {
		if (row0[j].info == NULL) {
		    format = NULL;
		} else {
		    format = row0[j].info->print;
		}
		if (j > 1) {
		    Qddb_ConcatBuffer(separator);
		}
		buf = TclQddb_GetCellValue(nodes+j, row0[j].data.type, format);
		if (nodes[j].data.type == QDDB_TABLE_TYPE_DEFAULT) {
		    switch (row0[j].data.type) {
		    case QDDB_TABLE_TYPE_DEFAULT:
		    case QDDB_TABLE_TYPE_STRING:
			Qddb_ConcatBuffer("\"");
			Qddb_ConcatEscapeBuffer(buf, (int)'\"');
			Qddb_ConcatBuffer("\"");
			break;
		    case QDDB_TABLE_TYPE_DATE:
			time = Qddb_DateStringToTime(buf);
			date = Qddb_DateTimeToString(time, dateformat);
			Free(time);
			Qddb_ConcatBuffer(date);
			Free(date);
			break;
		    case QDDB_TABLE_TYPE_INTEGER:
		    case QDDB_TABLE_TYPE_REAL:
		    case QDDB_TABLE_TYPE_CALC:
		    default:
			Qddb_ConcatBuffer(buf);
		    }
		} else {
		    switch (nodes[j].data.type) {
		    case QDDB_TABLE_TYPE_STRING:
			Qddb_ConcatBuffer("\"");
			Qddb_ConcatBuffer(buf);
			Qddb_ConcatBuffer("\"");
			break;
		    case QDDB_TABLE_TYPE_DATE:
			time = Qddb_DateStringToTime(buf);
			date = Qddb_DateTimeToString(time, dateformat);
			Free(time);
			Qddb_ConcatBuffer(date);
			Free(date);
			break;
		    case QDDB_TABLE_TYPE_INTEGER:
		    case QDDB_TABLE_TYPE_REAL:
		    case QDDB_TABLE_TYPE_CALC:
		    default:
			Qddb_ConcatBuffer(buf);
		    }
		}
	    }
	    Qddb_ConcatBuffer("\n");
	}
	buf = Qddb_GetBufferWithoutResize();
	if (concatsheets) {
	    Tcl_AppendResult(interp, buf, NULL);
	} else {
	    Tcl_AppendElement(interp, buf);
	}
	Free(buf);

    }
    return TCL_OK;
}

static char *ASColumnName(num)
    int                         num;
{
    if (num <= 0)
	return "";
    return Qddb_IntToSpreadsheetIndex(num-1);
}

static int ASDateFormat()
{
    int                         retval;

    if (strcmp(dateformat, "%B %d, %Y") == 0)
	return 1;
    /* 2: Dec 6, 1997 */
    if (strcmp(dateformat, "%d %b %y") == 0)
	return 3;
    if (strcmp(dateformat, "%m/%d/%y") == 0)
	return 4;
    if (strcmp(dateformat, "%d.%m.%y") == 0)
	return 5;
    if (strcmp(dateformat, "%Y-%m-%d") == 0)
	return 6;
    if (strcmp(dateformat, "%y-%m-%d") == 0)
	return 7;
    if (strcmp(dateformat, "%Y %m %d") == 0)
	return 8;
    if (strcmp(dateformat, "%y %m %d") == 0)
	return 9;
    if (strcmp(dateformat, "%Y%m%d") == 0)
	return 10;
    if (strcmp(dateformat, "%y%m%d") == 0)
	return 11;
    if (strcmp(dateformat, "%d/%m/%y") == 0)
	return 12;
    if (strcmp(dateformat, "%d.%m.%Y") == 0)
	return 13;
    if (strcmp(dateformat, "%b %d, %Y") == 0)
	return 14;
    if (strcmp(dateformat, "%B %Y") == 0)
	return 15;
    retval = atoi(dateformat);
    if (retval == 0) {
	if (qddb_datestyle == QDDB_DATE_STYLE_US)
	    return 4;
	else
	    return 12;
    }
    return retval;
}

static int ExportAS(interp, argc, argv)
    Tcl_Interp                  *interp;
    int                         argc;
    CONST char                  *argv[];
{
    Qddb_Table                  *table;
    Qddb_TableNode              *nodes, *row0;
    Qddb_TableHeaderFormat      *format;
    int                         i, j, first_table, start_table, tablenum, as_dateformat;
    int                         longbuf, buflen;
    CONST char                  *table_desc;
    char                        *buf, *time, *date, *val, tablename[MAXINTEGERLEN];
    char                        rowname[MAXINTEGERLEN], writebuf[BUFSIZ];
    
    as_dateformat = ASDateFormat();
    Qddb_InitBuffer();
    Qddb_ConcatBuffer("*BEGIN SPREADSHEETS VERSION=430/420 ENCODING=8BIT\n");
    Qddb_ConcatBuffer("Date Numbers Standard\n");
    for (start_table = argc-1; start_table > 0; start_table--) {
	if (strncmp(argv[start_table], "qddb_table", 10) != 0)
	    break;
    }
    start_table++;
    tablenum = 1;
    for (first_table = start_table; first_table < argc; first_table++, tablenum++) {
	if (strncmp(argv[first_table], "qddb_table", 10) != 0)
	    break;
	table_desc = argv[first_table];
	if ((table = TclQddb_GetTable(interp, table_desc)) == NULL) {
	    return TCL_ERROR;
	}
	Qddb_ConcatBuffer("View Start, Name: ~");
	strcpy(tablename, ASColumnName(tablenum));
	Qddb_ConcatBuffer(tablename);
	Qddb_ConcatBuffer(":~\n");
	sprintf(writebuf, "View Top Left: %s:A1\n", tablename);
	Qddb_ConcatBuffer(writebuf);
	sprintf(writebuf, "View Rows Visible: %s:1-%s:%d\n", tablename, tablename, 
		(int)table->rows + columntitles);
	Qddb_ConcatBuffer(writebuf);
	sprintf(writebuf, "View Columns Visible: %s:A-%s:%s\n", tablename, tablename, 
		ASColumnName((int)table->columns));
	Qddb_ConcatBuffer(writebuf);
	Qddb_ConcatBuffer("View End, Name: ~");
	strcpy(tablename, ASColumnName(tablenum));
	Qddb_ConcatBuffer(tablename);
	Qddb_ConcatBuffer(":~\n");
    }
    tablenum = 1;
    Qddb_ConcatBuffer("SHEETS TABLE\n");
    for (first_table = start_table; first_table < argc; first_table++, tablenum++) {
	if (strncmp(argv[first_table], "qddb_table", 10) != 0)
	    break;
	table_desc = argv[first_table];
	if ((table = TclQddb_GetTable(interp, table_desc)) == NULL) {
	    return TCL_ERROR;
	}
	val = TclQddb_TableGetHeaderInfo(table->info, NULL, "-title");	
	Qddb_ConcatBuffer("Sheet ");
	Qddb_ConcatBuffer(ASColumnName(tablenum));
	Qddb_ConcatBuffer(": ~");
	if (val == NULL) {
	    Qddb_ConcatBuffer(ASColumnName(tablenum));
	} else {
	    Qddb_ConcatBuffer(val);
	}
	Qddb_ConcatBuffer("~\n");
    }
    Qddb_ConcatBuffer("END SHEETS TABLE\n");
    tablenum = 1;
    for (first_table = start_table; first_table < argc; first_table++, tablenum++) {
	strcpy(tablename, ASColumnName(tablenum));
	table_desc = argv[first_table];
	if ((table = TclQddb_GetTable(interp, table_desc)) == NULL) {
	    return TCL_ERROR;
	}
	if (table->columns < 1) {
	    Tcl_AppendResult(interp, argv[0], " ", argv[1], ": no columns defined for table \"",
			     argv[first_table], "\"", NULL);
	    return TCL_ERROR;
	}
	row0 = table->table[0];
	if (columntitles != 0) {
	    val = TclQddb_TableGetHeaderInfo(row0[1].info, row0[1].info, "-title");
	    Qddb_ConcatBuffer("('DN,3||) ");
	    Qddb_ConcatBuffer(tablename);
	    Qddb_ConcatBuffer("!A1: ");
	    if (val != NULL) {
		Qddb_ConcatBuffer(val);
	    }
	    Qddb_ConcatBuffer("\n");
	    for (i = 2; i <= table->columns; i++) {
		Qddb_ConcatBuffer("('DN,3||) ");
		Qddb_ConcatBuffer(tablename);
		Qddb_ConcatBuffer("!");
		Qddb_ConcatBuffer(ASColumnName(i));
		Qddb_ConcatBuffer("1: ");
		val = TclQddb_TableGetHeaderInfo(row0[i].info, row0[i].info, "-title");
		if (val != NULL) {
		    Qddb_ConcatEscapeBuffer(val, '\"');
		}
		Qddb_ConcatBuffer("\n");
	    }
	}
	for (i = 1; i <= table->rows; i++) {
	    nodes = table->table[i];
	    sprintf(rowname, "%d: ", i+columntitles);
	    for (j = 1; j <= table->columns; j++) {
		if (row0[j].info == NULL) {
		    format = NULL;
		} else {
		    format = row0[j].info->print;
		}
		buf = TclQddb_GetCellValue(nodes+j, row0[j].data.type, format);
		if (strlen(buf) >= 60)
		    longbuf = 1;
		else
		    longbuf = 0;
		if (nodes[j].data.type == QDDB_TABLE_TYPE_DEFAULT) {
		    switch (row0[j].data.type) {
		    case QDDB_TABLE_TYPE_DEFAULT:
		    case QDDB_TABLE_TYPE_STRING:
			sprintf(writebuf, "('DN||) %s!%s%s", tablename, ASColumnName(j), rowname);
			if (longbuf) {
			    Qddb_ConcatBuffer(writebuf);
			    buflen = strlen(writebuf);
			    writebuf[0] = '\0';
			    strncat(writebuf, buf, (size_t)(80 - buflen));
			    Qddb_ConcatPrintableBuffer(writebuf);
			    buflen = 80 - buflen;
			    writebuf[0] = ' ';
			    writebuf[1] = ' ';
			    while (buflen < strlen(buf)) {
				Qddb_ConcatBuffer("\n");
				writebuf[2] = '\0';
				strncat(writebuf, buf+buflen, (size_t)78);
				Qddb_ConcatPrintableBuffer(writebuf);
				buflen += 78;
			    }
			} else {
			    Qddb_ConcatBuffer(writebuf);
			    Qddb_ConcatPrintableBuffer(buf);
			}
			break;
		    case QDDB_TABLE_TYPE_DATE:
			sprintf(writebuf, "(D%d||) ", as_dateformat);
			Qddb_ConcatBuffer(writebuf);
			Qddb_ConcatBuffer(tablename);
			Qddb_ConcatBuffer("!");
			Qddb_ConcatBuffer(ASColumnName(j));
			Qddb_ConcatBuffer(rowname);
			time = Qddb_DateStringToTime(buf);
			date = Qddb_DateTimeToString(time, "days1900");
			Free(time);
			Qddb_ConcatBuffer(date);
			Free(date);
			break;
		    case QDDB_TABLE_TYPE_INTEGER:
		    case QDDB_TABLE_TYPE_REAL:
		    case QDDB_TABLE_TYPE_CALC:
		    default:
			Qddb_ConcatBuffer("(DN||) ");
			Qddb_ConcatBuffer(tablename);
			Qddb_ConcatBuffer("!");
			Qddb_ConcatBuffer(ASColumnName(j));
			Qddb_ConcatBuffer(rowname);
			Qddb_ConcatBuffer(buf);
		    }
		} else {
		    switch (nodes[j].data.type) {
		    case QDDB_TABLE_TYPE_DEFAULT:
		    case QDDB_TABLE_TYPE_STRING:
			sprintf(writebuf, "('DN||) %s!%s%s", tablename, ASColumnName(j), rowname);
			if (longbuf) {
			    Qddb_ConcatBuffer(writebuf);
			    buflen = strlen(writebuf);
			    writebuf[0] = '\0';
			    strncat(writebuf, buf, (size_t)(80 - buflen));
			    Qddb_ConcatPrintableBuffer(writebuf);
			    buflen = 80 - buflen;
			    writebuf[0] = ' ';
			    writebuf[1] = ' ';
			    while (buflen < strlen(buf)) {
				Qddb_ConcatBuffer("\n");
				writebuf[2] = '\0';
				strncat(writebuf, buf+buflen, (size_t)78);
				Qddb_ConcatPrintableBuffer(writebuf);
				buflen += 78;
			    }
			} else {
			    Qddb_ConcatBuffer(writebuf);
			    Qddb_ConcatPrintableBuffer(buf);
			}
			break;
		    case QDDB_TABLE_TYPE_DATE:
			sprintf(writebuf, "(D%d||) ", as_dateformat);
			Qddb_ConcatBuffer(writebuf);
			Qddb_ConcatBuffer(tablename);
			Qddb_ConcatBuffer("!");
			Qddb_ConcatBuffer(ASColumnName(j));
			Qddb_ConcatBuffer(rowname);
			time = Qddb_DateStringToTime(buf);
			date = Qddb_DateTimeToString(time, "days1900");
			Free(time);
			Qddb_ConcatBuffer(date);
			Free(date);
			break;
		    case QDDB_TABLE_TYPE_INTEGER:
		    case QDDB_TABLE_TYPE_REAL:
		    case QDDB_TABLE_TYPE_CALC:
		    default:
			Qddb_ConcatBuffer("(DN||) ");
			Qddb_ConcatBuffer(tablename);
			Qddb_ConcatBuffer("!");
			Qddb_ConcatBuffer(ASColumnName(j));
			Qddb_ConcatBuffer(rowname);
			Qddb_ConcatBuffer(buf);
		    }
		}
		Qddb_ConcatBuffer("\n");
	    }
	}
	if (concatsheets) {
	    ;
	} else {
	    ;
	}
    }
    Qddb_ConcatBuffer("*END SPREADSHEETS\n");
    buf = Qddb_GetBufferWithoutResize();
    Tcl_SetResult(interp, buf, TCL_DYNAMIC);
    return TCL_OK;
}













