
/* $RCSfile: TclScript.c,v $ $Revision: 2.3 $ $Date: 2003/03/23 17:16:17 $
 *
 * Copyright (C) 1995-2003 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "tcl.h"
#include "Qddb.h"
#include "tclQddb.h"

static int TclQddb_ScriptCompile _ANSI_ARGS_((Tcl_Interp *, CONST char *, CONST char *, CONST char *));
static int TclQddb_ScriptAutogen _ANSI_ARGS_((Tcl_Interp *, Qddb_Schema *, CONST char *));
static int TclQddb_GenerateFx    _ANSI_ARGS_((Tcl_Interp *, CONST char *, CONST char *, Qddb_ScriptCell *));
static int TclQddb_TraverseCells _ANSI_ARGS_((Tcl_Interp *, CONST char *, CONST char *, Qddb_ScriptCell *));
static void TclQddb_ScriptPack         _ANSI_ARGS_((char *, Qddb_ScriptCell *));
static void TclQddb_ScriptConfigure    _ANSI_ARGS_((char *, Qddb_ScriptCell *));
static void TclQddb_ScriptConfigureFx  _ANSI_ARGS_((char *, Qddb_ScriptCell *));
static int TclQddb_AutogenProduce      _ANSI_ARGS_((Tcl_Interp *, CONST char *, CONST char *, size_t));
static void TclQddb_AddWaitStack _ANSI_ARGS_((Qddb_ScriptAttrStack **, char *, size_t, Qddb_Boolean));
static char *tabs _ANSI_ARGS_((size_t));

/* qddb_script autogen <schema_desc> - generate QddbScript from Schema
 * qddb_script compile <schema_variable> ?<config name>? <script>
 *             compile QddbScript into language dependent code
 */


int TclQddb_ScriptProc(clientData, interp, argc, argv)
    ClientData        clientData;
    Tcl_Interp        *interp;
    int               argc;
    CONST char        *argv[];
{
    Qddb_Schema       *schema;
    CONST char        *config;
    int               scriptarg;

    if (argc < 3 || argc > 5) {
	Tcl_AppendResult(interp, argv[0], ": wrong # args", NULL);
	return TCL_ERROR;
    }
    switch (argv[1][0]) {
    case 'a':
	if ((schema = TclQddb_GetSchema(argv[2])) == NULL) {
	    Tcl_AppendResult(interp, argv[0], " ", argv[1], 
			     ": bad schema descriptor \"", argv[2], "\"", NULL);
	    return TCL_ERROR;
	}
	if (strcmp(argv[1], "autogen") != 0) {
	    Tcl_AppendResult(interp, argv[0], ": invalid option \"", argv[1], "\"", NULL);	    
	    return TCL_ERROR;
	}
	if (argc != 3) {
	    Tcl_AppendResult(interp, argv[0], " ", argv[1], ": wrong # args", NULL);
	    return TCL_ERROR;
	}
	TclQddb_ScriptAutogen(interp, schema, argv[2]);
	break;
    case 'c':
	if (strcmp(argv[1], "compile") != 0) {
	    Tcl_AppendResult(interp, argv[0], ": invalid option \"", argv[1], "\"", NULL);	    
	    return TCL_ERROR;
	}
	if (argc == 4) {
	    config = "";
	    scriptarg = 3;
	} else if (argc == 5) {
	    config = argv[3];
	    scriptarg = 4;
	} else {
	    Tcl_AppendResult(interp, argv[0], " ", argv[1], ": wrong # args", NULL);
	    return TCL_ERROR;
	}
	if (TclQddb_ScriptCompile(interp, argv[2], config, argv[scriptarg]) != TCL_OK)
	    return TCL_ERROR;
	break;
    default:
	Tcl_AppendResult(interp, argv[0], ": invalid option \"", argv[1], "\"", NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}



static int TclQddb_ScriptCompile(interp, schema, config, script)
    Tcl_Interp         *interp;
    CONST char         *schema, *config, *script;
{
    int                retval;
    char               *script_buffer;
    Qddb_ScriptCell    *cells;

    script_buffer = (char *)Malloc(strlen(script)+1);
    strcpy(script_buffer, script);
    Qddb_ResetQddbScriptParser();
    Qddb_SetBuffer(script_buffer);
    if (QddbScriptParse() != 0) {
	cells = Qddb_GetQddbScriptParserResults();
	Qddb_Free(QDDB_TYPE_QDDBSCRIPTCELL, cells);
	Tcl_AppendResult(interp, "QddbScript ", qddb_errmsg, NULL);
	Free(script_buffer);
	return TCL_ERROR;
    }
    cells = Qddb_GetQddbScriptParserResults();
    retval = TCL_OK;
    if (TclQddb_GenerateFx(interp, schema, config, cells) != TCL_OK) {
	retval = TCL_ERROR;
    }
    Qddb_Free(QDDB_TYPE_QDDBSCRIPTCELL, cells);
    Free(script_buffer);
    return retval;
}

static char *tabs(number)
    size_t            number;
{
    static char       *tabs = "\t\t\t\t\t\t\t\t\t\t\
\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t";
    static char       tmp[128];

    if (number == 0) {
	return "";
    }
    strncpy(tmp, tabs, number);
    tmp[number] = '\0';
    return tmp;
}

static int TclQddb_AutogenProduce(interp, list, node, level)
    Tcl_Interp        *interp;
    CONST char        *list, *node;
    size_t            level;
{
    int               argc, thisargc, i, inentry, entrynumber;
    CONST char        **argv, **thisargv;
    char              buf[BUFSIZ];
    char              newnode[BUFSIZ];

    argc = 0; argv = NULL; inentry = 0; entrynumber = 0;
    if (Tcl_SplitList(interp, list, &argc, &argv) != TCL_OK)
	return TCL_ERROR;
    for (i = 0; i < argc; i++) {
	if (Tcl_SplitList(interp, argv[i], &thisargc, &thisargv) != TCL_OK) {
	    Free(argv);
	    return TCL_ERROR;
	}
	if (strlen(node) == 0) {
	    strcpy(newnode, thisargv[0]);
	} else {
	    sprintf(newnode, "%s.%s", node, thisargv[0]);
	}
	if (strlen(thisargv[2]) > 0) { /* not a leaf */
	    if (inentry) {
		Qddb_ConcatBuffer("[break]\n");
	    }
	    inentry = 0;
	    entrynumber = 0;
	    sprintf(buf, "%s[frame attr=%s", tabs(level), newnode);
	    Qddb_ConcatBuffer(buf);
	    if (level == 0) {
		Qddb_ConcatBuffer(" border=3 relief=groove]\n");
	    } else {
		Qddb_ConcatBuffer("]\n");
	    }
	    if (TclQddb_AutogenProduce(interp, thisargv[2], newnode, level+1) != TCL_OK)
		return TCL_ERROR;
	    sprintf(buf, "%s[end frame]\n%s[break]\n", tabs(level), tabs(level));
	    Qddb_ConcatBuffer(buf);
	} else {
	    inentry = 1;
	    sprintf(buf, "%s[entry attr=%s width=10]\n", tabs(level), newnode);
	    Qddb_ConcatBuffer(buf);
	    entrynumber++;
	    if (entrynumber >= 4) {
		sprintf(buf, "%s[break]\n", tabs(level));
		Qddb_ConcatBuffer(buf);
		entrynumber = 0;
	    }
	}
	Free(thisargv);
	thisargv = NULL;
    }
    Free(argv);
    return TCL_OK;
}

static int TclQddb_ScriptAutogen(interp, schema, schema_desc)
    Tcl_Interp        *interp;
    Qddb_Schema       *schema;
    CONST char        *schema_desc;
{
    char              *buf;
    int               retval;

    if ((buf = TclQddb_TraverseSchemaTree(interp, schema->Tree)) == NULL) {
	Tcl_AppendResult(interp, "qddb_script autogen: cannot print schema \"", 
			 schema_desc, "\" (Qddb error)", NULL);
	return TCL_ERROR;
    }
    Qddb_InitBuffer();
    Qddb_ConcatBuffer("[prologue]\n[menubar]\n[options autoalign]\n");
    retval = TclQddb_AutogenProduce(interp, buf, "", (size_t)0);
    Qddb_ConcatBuffer("[epilogue]\n");
    Free(buf);
    buf = Qddb_GetBufferWithoutResize();
    Tcl_SetResult(interp, buf, TCL_DYNAMIC);
    return retval;
}


static int TclQddb_GenerateFx(interp, schema, config, cells)
    Tcl_Interp        *interp;
    CONST char        *schema, *config;
    Qddb_ScriptCell   *cells;
{
    Qddb_ScriptCell   *cells_ptr;
    char              buf[BUFSIZ], *chptr, *bannergif, *bannerxbm, *bannertitle;

    Qddb_InitBuffer();
    cells_ptr = cells;
    if (cells_ptr->option == QDDB_SCRIPT_OPTION_CODE) {
	if (cells_ptr->values != NULL) {
	    Qddb_ConcatBuffer("\n;# [USER-INSERTED CODE]\n");
	    Qddb_ConcatBuffer(*cells_ptr->values);
	    Qddb_ConcatBuffer("\n;# [END USER-INSERTED CODE]\n");
	}
	cells_ptr = cells_ptr->next;
    }
    if (cells_ptr->option == QDDB_SCRIPT_OPTION_PROLOGUE) {
	if (cells_ptr->bannergif != NULL) {
	    bannergif = cells_ptr->bannergif;
	} else {
	    bannergif = "";
	}
	if (cells_ptr->bannerxbm != NULL) {
	    bannerxbm = cells_ptr->bannerxbm;
	} else {
	    bannerxbm = "";
	}
	if (cells_ptr->bannertitle != NULL) {
	    bannertitle = cells_ptr->bannertitle;
	} else {
	    bannertitle = "";
	}
	if (strlen(config) == 0 && cells_ptr->config != NULL) {
	    config = cells_ptr->config;
	} else {
	    config = "nxqddb";
	}
	sprintf(buf, "set qddbscript(schema) [Fx:Prologue %s {} \"%s\" \"%s\" \"%s\"]\n", 
		config, bannergif, bannerxbm, bannertitle);
	cells_ptr = cells_ptr->next;
	Qddb_ConcatBuffer(buf);
    }
    if (TclQddb_TraverseCells(interp, schema, config, cells_ptr) != TCL_OK) {
	chptr=Qddb_GetBufferWithoutResize();
	Free(chptr);
	return TCL_ERROR;
    } else {
	chptr=Qddb_GetBufferWithoutResize();
	Tcl_SetResult(interp, chptr, TCL_DYNAMIC);
    }
    return TCL_OK;
}

static int TclQddb_TraverseCells(interp, schemaname, config, cells)
    Tcl_Interp        *interp;
    CONST char        *schemaname, *config;
    Qddb_ScriptCell   *cells;
{
    int                   frames[QDDB_SCRIPT_MAXFRAMEDEPTH], depth, first_frame, first_entry;
    int                   nb_level, menubar, notebook, tab, current_tab, flags, size;
    int                   first_entry_in_frame;
    size_t                level_widths[MAXIMUM_NUMBER_OF_SCHEMA_LEVELS];
    char                  current_frame[BUFSIZ], tmp_frame[BUFSIZ], buf[BUFSIZ], **values, *ptr;
    CONST char            *ch;
    char                  current_notebook[BUFSIZ], linenumber[BUFSIZ];
    Qddb_Schema           *schema;
    Qddb_ScriptAttrStack  *waiting_stack, *attr_stack, *stack_ptr, *tmpstack_ptr;
    Qddb_Boolean          autoalign;

    if (schemaname[0] != '\0' && (ch = Tcl_GetVar(interp, schemaname, TCL_GLOBAL_ONLY)) != NULL) {
	schema = TclQddb_GetSchema(ch);
    } else {
	schema = NULL;
    }
    memset(frames, 0, sizeof(frames));
    memset(level_widths, 0, sizeof(level_widths));
    strcpy(current_frame, ".fx.f0");
    depth = 1;
    frames[0] = 1;
    nb_level = 0;
    first_frame = 1;
    first_entry = 1;
    first_entry_in_frame = 1;
    menubar = 0;
    notebook = 0;
    tab = 0;
    current_tab = 0;
    waiting_stack = NULL;
    attr_stack = NULL;
    autoalign = False;
    while (cells != NULL) {
	if (cells->autoalign == True) {
	    autoalign = True;
	} else if (cells->autoalign == Inapplicable) {
	    autoalign = False;
	}
	switch (cells->option) {
	case QDDB_SCRIPT_OPTION_OPTIONS:
	    break;
	case QDDB_SCRIPT_OPTION_FRAME:
	    if (menubar != 1) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [frame] before [menubar] at line ", 
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (notebook == 1 && tab == 0) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [frame] must be within [tab] [endtab] at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (cells->endoption == True) {
		depth -= 2;
		if (depth < 1) {
		    sprintf(linenumber, "%d", cells->linenumber);
		    Tcl_AppendResult(interp, "Error: mismatched [end frame] at line ", 
				     linenumber, "\n", NULL);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		    return TCL_ERROR;
		}
		*rindex(current_frame, '.') = '\0'; /* remove 2 levels */
		*rindex(current_frame, '.') = '\0';
            } else {
		first_entry_in_frame = 1;
		sprintf(buf, ".f%d", frames[depth]);
		strcat(current_frame, buf);
		frames[depth]++;
		depth++;
		if (cells->attr == NULL) {
		    sprintf(buf, "frame %s ; pack %s -side top -expand on -fill both\n",
			    current_frame, current_frame);
		    Qddb_ConcatBuffer(buf);
		    if (cells->side == 0) {
			cells->side = QDDB_SCRIPT_SIDE_TOP;
		    }
		    TclQddb_ScriptConfigure(current_frame, cells);
		    TclQddb_ScriptPack(current_frame, cells);
		    sprintf(buf, ".f%d", frames[depth]);
		    strcat(current_frame, buf);
		    frames[depth]++;
		    depth++;
		    sprintf(buf, "frame %s ; pack %s -side top -expand on -fill both\n",
			    current_frame, current_frame);
		    Qddb_ConcatBuffer(buf);
		} else {
		    if (first_frame) {
			sprintf(buf, "Fx_Frame =%s -w %s -side top -attr %s -setschema $qddbscript(schema) \n",
				cells->attr, current_frame, cells->attr);
			first_frame = 0;
		    } else {
			sprintf(buf, "Fx_Frame =%s -w %s -attr %s -side top \n",
				cells->attr, current_frame, cells->attr);
		    }
		    Qddb_ConcatBuffer(buf);
		    TclQddb_AddWaitStack(&waiting_stack, cells->attr, 0, False);

		    if (cells->side == 0) {
			cells->side = QDDB_SCRIPT_SIDE_TOP;
		    }
		    TclQddb_ScriptConfigure(current_frame, cells);
		    TclQddb_ScriptConfigureFx(current_frame, cells);
		    TclQddb_ScriptPack(current_frame, cells);
		    sprintf(buf, ".f%d", frames[depth]);
		    strcat(current_frame, buf);
		    frames[depth]++;
		    depth++;
		    sprintf(buf, "frame %s ; pack %s -side top -expand on -fill both\n",
			    current_frame, current_frame);
		    Qddb_ConcatBuffer(buf);
		}
	    }
	    break;
	case QDDB_SCRIPT_OPTION_ENTRY:
	    if (menubar != 1) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [entry] before [menubar] at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (notebook == 1 && tab == 0) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [entry] must be within [tab] [endtab] at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    sprintf(tmp_frame, "%s.e%d", current_frame, frames[depth]);
	    frames[depth]++;
	    if (cells->attr == NULL) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: entry missing required \"attr=\" at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (first_entry) {
		sprintf(buf, "Fx_Entry =%s -w %s -attr %s -setschema $qddbscript(schema) \\\n\
\t -searchfor_entry $qddbscript(search_entry) -side left -type Entry ",
			cells->attr, tmp_frame, cells->attr);
		first_entry = 0;
	    } else {
		sprintf(buf, "Fx_Entry =%s -w %s -attr %s \\\n\t-side left -type Entry ",
			cells->attr, tmp_frame, cells->attr);
	    }
	    Qddb_ConcatBuffer(buf);
	    if (cells->width != 0) {
		sprintf(buf, "-width %d ", cells->width);
		Qddb_ConcatBuffer(buf);
	    }
	    Qddb_ConcatBuffer("\n");
	    TclQddb_ScriptConfigure(tmp_frame, cells);
	    TclQddb_ScriptConfigureFx(tmp_frame, cells);
	    TclQddb_ScriptPack(tmp_frame, cells);
	    break;
	case QDDB_SCRIPT_OPTION_TEXT:
	    if (menubar != 1) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [text] before [menubar] at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (notebook == 1 && tab == 0) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [text] must be within [tab] [endtab] at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    sprintf(tmp_frame, "%s.t%d", current_frame, frames[depth]);
	    frames[depth]++;
	    if (cells->attr == NULL) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: entry missing required \"attr=\" at line ", 
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);		
		return TCL_ERROR;
	    }
	    if (first_entry) {
		sprintf(buf, "Fx_Entry =%s -w %s -attr %s -setschema $qddbscript(schema) \\\n\
\t -searchfor_entry $qddbscript(search_entry) -side left -type Text ",
			cells->attr, tmp_frame, cells->attr);
		first_entry = 0;
	    } else {
		sprintf(buf, "Fx_Entry =%s -w %s -attr %s \\\n\
\t-side left -type Text ",
			cells->attr, tmp_frame, cells->attr);
	    }
	    Qddb_ConcatBuffer(buf);
	    if (cells->width != 0) {
		sprintf(buf, "-width %d ", cells->width);
		Qddb_ConcatBuffer(buf);
	    }
	    if (cells->height != 0) {
		sprintf(buf, "-height %d ", cells->height);
		Qddb_ConcatBuffer(buf);
	    }
	    Qddb_ConcatBuffer("\n");
	    TclQddb_ScriptConfigure(tmp_frame, cells);
	    TclQddb_ScriptConfigureFx(tmp_frame, cells);
	    TclQddb_ScriptPack(tmp_frame, cells);
	    break;
	case QDDB_SCRIPT_OPTION_RADIOBUTTON:
	    if (menubar != 1) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [radiobutton] before [menubar] at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (notebook == 1 && tab == 0) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [radiobutton] must be within [tab] [endtab] at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    sprintf(tmp_frame, "%s.r%d", current_frame, frames[depth]);
	    frames[depth]++;
	    if (cells->attr == NULL) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: entry missing required \"attr=\" at line ", 
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (first_entry) {
		sprintf(buf, "Fx_Entry =%s -w %s -attr %s -setschema $%s \\\n\
\t -searchfor_entry $qddbscript(search_entry) -side left -type Radiobutton ",
			cells->attr, tmp_frame, cells->attr, schemaname);
		first_entry = 0;
	    } else {
		sprintf(buf, "Fx_Entry =%s -w %s -attr %s \\\n\
\t-side left -type Radiobutton ",
			cells->attr, tmp_frame, cells->attr);
	    }
	    Qddb_ConcatBuffer(buf);
	    if (cells->width != 0) {
		sprintf(buf, "-width %d ", cells->width);
		Qddb_ConcatBuffer(buf);
	    }
	    if (cells->height != 0) {
		sprintf(buf, "-height %d ", cells->height);
		Qddb_ConcatBuffer(buf);
	    }
	    if (cells->values != NULL) {
		Qddb_ConcatBuffer("\\\n\t-default_values {");
		values = cells->values;
		while (*values != NULL) {
		    sprintf(buf, " {%s}", *values);
		    Qddb_ConcatBuffer(buf);
		    values++;
		}
		Qddb_ConcatBuffer(" }");
	    }
	    Qddb_ConcatBuffer("\n");
	    TclQddb_ScriptConfigure(tmp_frame, cells);
	    TclQddb_ScriptConfigureFx(tmp_frame, cells);
	    TclQddb_ScriptPack(tmp_frame, cells);
	    break;
	case QDDB_SCRIPT_OPTION_NOTEBOOK:
	    if (menubar != 1) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [notebook] before [menubar] at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (cells->endoption == True) {
		depth--;
		nb_level--;
		if (depth < 0 || nb_level < 0) {
		    sprintf(linenumber, "%d", cells->linenumber);
		    Tcl_AppendResult(interp, "Error: mismatched [end notebook] at line ",
				     linenumber, "\n", NULL);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		    return TCL_ERROR;
		}
		strcpy(current_frame, current_notebook);
		*rindex(current_frame, '.') = '\0';
		notebook = 0;
		sprintf(buf, "%s raise tab0\n", current_notebook);
		Qddb_ConcatBuffer(buf);
		sprintf(buf, "%s compute_size\n", current_notebook);
		Qddb_ConcatBuffer(buf);
	    } else {
		if (notebook == 1) {
		    sprintf(linenumber, "%d", cells->linenumber);
		    Tcl_AppendResult(interp, "Error: nested [notebook] at line ",
				     linenumber, "\n", NULL);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		    return TCL_ERROR;
		}
		notebook = 1;
		current_tab = 0;
		sprintf(buf, ".nb%d", nb_level);
		strcat(current_frame, buf);
		frames[depth]++;
		depth++;
		nb_level++;
		sprintf(buf, "NoteBook %s -bd 2\npack %s -side top -expand on -fill both\n",
			current_frame, current_frame);
		Qddb_ConcatBuffer(buf);
		TclQddb_ScriptPack(current_frame, cells);
		strcpy(current_notebook, current_frame);
	    }
	    break;
	case QDDB_SCRIPT_OPTION_TAB:
	    if (menubar != 1) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [tab] before [menubar] at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (notebook == 0) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [tab] without [notebook] at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (cells->endoption == True) {
		depth--;
		nb_level--;
		if (depth < 0 || nb_level < 0) {
		    sprintf(linenumber, "%d", cells->linenumber);
		    Tcl_AppendResult(interp, "Error: mismatched [end tab] at line ",
				     linenumber, "\n", NULL);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		    return TCL_ERROR;
		}
		*rindex(current_frame, '.') = '\0';
		tab = 0;
	    } else {
		if (tab == 1) {
		    sprintf(linenumber, "%d", cells->linenumber);
		    Tcl_AppendResult(interp, "Error: already in [tab] at line ",
				     linenumber, "\n", NULL);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		    return TCL_ERROR;
		}
		tab = 1;
		flags = 0;
		if (cells->label == NULL) {
		    ptr = "";
		} else if (cells->label[0] == '@') {
		    ptr = cells->label;
		} else {
		    size = Tcl_ScanElement(cells->label, &flags);
		    ptr = (char *)Malloc((size_t)size);
		    Tcl_ConvertElement(cells->label, ptr, flags);
		}
		if (*ptr == '@') {
		    sprintf(buf, "%s insert end tab%d -image %s\n", current_notebook, current_tab, ptr);
		} else {
		    sprintf(buf, "%s insert end tab%d -text %s\n", current_notebook, current_tab, ptr);
		}
		Qddb_ConcatBuffer(buf);
		sprintf(buf, "uplevel \\#0 {lappend ::fx::qddbscript::tab_list {%s tab%d}}\n", current_notebook,
			current_tab);
		if (cells->label != NULL && cells->label[0] != '@')
		    Free(ptr);
		Qddb_ConcatBuffer(buf);
		sprintf(current_frame, "[%s getframe tab%d]", current_notebook, current_tab);
		first_entry_in_frame = 1;
		current_tab++;
		frames[depth]++;
		depth++;
		nb_level++;
		frames[depth]++;
		sprintf(buf, ".f%d", frames[depth]);
		strcat(current_frame, buf);
		sprintf(buf, "frame %s ; pack %s -side top -expand on -fill both\n",
			current_frame, current_frame);
		Qddb_ConcatBuffer(buf);
	    }
	    break;
	case QDDB_SCRIPT_OPTION_LABEL:
	    if (menubar != 1) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [label] before [menubar] at line ", 
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (notebook == 1 && tab == 0) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [label] must be within [tab] [endtab] at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    sprintf(tmp_frame, "%s.l%d", current_frame, frames[depth]);
	    frames[depth]++;
	    if (cells->label != NULL) {
		if (cells->label[0] == '@') {
		    sprintf(buf, "label %s -bitmap %s\npack %s -expand on -fill x\n", 
			    tmp_frame, cells->label, tmp_frame);
		} else {
		    size = Tcl_ScanElement(cells->label, &flags);
		    ptr = (char *)Malloc((size_t)size);
		    Tcl_ConvertElement(cells->label, ptr, flags);
/* labelside, etc. etc. */
		    sprintf(buf, "label %s -text %s\npack %s -expand on -fill x\n", 
			    tmp_frame, ptr, tmp_frame);
		    Free(ptr);
		}
		Qddb_ConcatBuffer(buf);
		TclQddb_ScriptConfigure(tmp_frame, cells);
		TclQddb_ScriptPack(tmp_frame, cells);
	    }
	    break;
	case QDDB_SCRIPT_OPTION_BREAK:
	    if (menubar != 1) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [break] before [menubar] at line ", 
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (notebook == 1 && tab == 0) {
		sprintf(linenumber, "%d", cells->linenumber);
		Tcl_AppendResult(interp, "Error: [break] must be within [tab] [endtab] at line ",
				 linenumber, "\n", NULL);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		return TCL_ERROR;
	    }
	    if (first_entry_in_frame) {
		break;
	    }
	    first_entry_in_frame = 1;
	    *rindex(current_frame, '.') = '\0'; /* remove 1 level */
	    frames[depth]++;
	    sprintf(buf, ".f%d", frames[depth]);
	    strcat(current_frame, buf);
	    sprintf(buf, "frame %s ; pack %s -side top -expand on -fill x\n",
		    current_frame, current_frame);   
	    Qddb_ConcatBuffer(buf);
	    TclQddb_ScriptConfigure(current_frame, cells);
	    break;
	case QDDB_SCRIPT_OPTION_PROLOGUE:
	    sprintf(linenumber, "%d", cells->linenumber);
	    Tcl_AppendResult(interp, "Error: [prologue] not first option at line ",
			     linenumber, "\n", NULL);
	    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
	    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
	    return TCL_ERROR;
	case QDDB_SCRIPT_OPTION_EPILOGUE:
	    Qddb_ConcatBuffer("Fx:Epilogue\n");
	    break;
	case QDDB_SCRIPT_OPTION_CODE:
	    if (cells->values != NULL) {
		sprintf(buf, "set qddbscript_frame %s", current_frame);
		Qddb_ConcatBuffer(buf);
		Qddb_ConcatBuffer("\n;# [USER-INSERTED CODE]\n");
		Qddb_ConcatBuffer(*cells->values);
		Qddb_ConcatBuffer("\n;# [END USER-INSERTED CODE]\n");
	    }
	    break;
	case QDDB_SCRIPT_OPTION_MENUBAR:
	    if (menubar != 0) {
		break;
	    }
	    menubar = 1;
	    Qddb_ConcatBuffer("Fx_Menubar menubar -w .menubar -schema \\\n");
	    Qddb_ConcatBuffer("\t$qddbscript(schema) -array gv_attr -config_dir \\\n");
	    sprintf(buf, "\t.%s_config -restrict $qddbscript(trans_name)\n", config);
	    Qddb_ConcatBuffer(buf);
	    Qddb_ConcatBuffer("set qddbscript(search_entry) [menubar SearchForEntry]\n");
	    Qddb_ConcatBuffer("frame .fx; pack .fx -side top -expand on -fill both\n");
	    Qddb_ConcatBuffer("frame .fx.f0; pack .fx.f0 -side top -expand on -fill both\n");
	    break;
	default:
	    Tcl_AppendResult(interp, "Unknown error, bad QddbScript [option]\n", NULL);
	    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
	    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
	    return TCL_ERROR;
	}
	switch (cells->option) {
	case QDDB_SCRIPT_OPTION_FRAME:
	    if (cells->attr == NULL)
		break;
	    if (schema != NULL) {
		int        attr_number;
		size_t     level;

		if ((attr_number=Qddb_ConvertAttributeNameToNumber(schema, cells->attr)) == -1) {
		    sprintf(linenumber, "%d", cells->linenumber);
		    Tcl_AppendResult(interp, "QddbScript error: bad attribute \"", cells->attr,
				     "\" at line number ", linenumber, NULL);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		    return TCL_ERROR;
		}
		if (autoalign == True && index(cells->attr, '.') != NULL) {
		    level = schema->Entries[attr_number].Level;
		    if (level > 1) {
			level--;
			if (cells->label != NULL) {
			    level_widths[level] = MAX(strlen(cells->label), level_widths[level]);
			} else if (schema->Entries[attr_number].VerboseName != NULL &&
				   schema->Entries[attr_number].VerboseName[0] != '\0') {
			    level_widths[level] = 
				MAX(strlen(schema->Entries[attr_number].VerboseName), level_widths[level]);
			} else if (schema->Entries[attr_number].Name != NULL) {
			    level_widths[level] = 
				MAX(strlen(schema->Entries[attr_number].Name), level_widths[level]);
			}
			TclQddb_AddWaitStack(&attr_stack, cells->attr, level, True);
		    }
		}
	    }
	    break;
	case QDDB_SCRIPT_OPTION_ENTRY:
	case QDDB_SCRIPT_OPTION_TEXT:
	case QDDB_SCRIPT_OPTION_RADIOBUTTON:
	    if (cells->attr == NULL)
		break;
	    if (schema != NULL) {
		int        attr_number;
		size_t     level;

		if ((attr_number=Qddb_ConvertAttributeNameToNumber(schema, cells->attr)) == -1) {
		    sprintf(linenumber, "%d", cells->linenumber);
		    Tcl_AppendResult(interp, "QddbScript error: bad attribute \"", cells->attr,
				     "\" at line number ", linenumber, NULL);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, waiting_stack);
		    Qddb_Free(QDDB_TYPE_QDDBSCRIPTATTRSTACK, attr_stack);
		    return TCL_ERROR;
		}
		if (first_entry_in_frame && autoalign == True) {
		    level = schema->Entries[attr_number].Level;
		    if (level > 1) {
			level--;
		    }
		    if (cells->label != NULL) {
			level_widths[level] = MAX(strlen(cells->label), level_widths[level]);
		    } else if (schema->Entries[attr_number].VerboseName != NULL &&
			       schema->Entries[attr_number].VerboseName[0] != '\0') {
			level_widths[level] = 
			    MAX(strlen(schema->Entries[attr_number].VerboseName), level_widths[level]);
		    } else if (schema->Entries[attr_number].Name != NULL) {
			level_widths[level] = 
			    MAX(strlen(schema->Entries[attr_number].Name), level_widths[level]);
		    }
		    TclQddb_AddWaitStack(&attr_stack, cells->attr, level, False);
		}
	    }
	    first_entry_in_frame = 0;
	    stack_ptr = waiting_stack;
	    while (stack_ptr != NULL) {
		sprintf(buf, "=%s configure -focus [=%s GetEntry]\n",
			stack_ptr->attr, cells->attr);
		Qddb_ConcatBuffer(buf);
		tmpstack_ptr = stack_ptr;
		stack_ptr = stack_ptr->next;
		Free(tmpstack_ptr);
	    }
	    waiting_stack = NULL;
	    break;
	default:
	    ;
	}
	cells = cells->next;
    }
    stack_ptr = attr_stack;
    while (stack_ptr != NULL) {
	int        i, levelsum;

	levelsum = 0;
	for (i = 1; i <= stack_ptr->level; i++) {
	    levelsum += (level_widths[i] + 1);
	}
	if (levelsum != 0) {
	    sprintf(buf, "\t[=%s cget -w].f_0.l configure -width %d\n", 
		    stack_ptr->attr, levelsum);
	    Qddb_ConcatBuffer(buf);
	}
	tmpstack_ptr = stack_ptr;
	stack_ptr = stack_ptr->next;
	Free(tmpstack_ptr);
    }
    return TCL_OK;
}

static void TclQddb_ScriptPack(frame, cell)
    char                *frame;
    Qddb_ScriptCell     *cell;
{
    char                buf[BUFSIZ];
    char                *anchor_opts, *expand_opts, *side_opts, pad_opts[BUFSIZ];

    /* frame expansion */
    switch (cell->noexpand) {
    case True:
	expand_opts = "-expand off -fill none";
	break;
    case False:
    default:
	expand_opts = "-expand on -fill x";
    }

    /* frame side */
    switch (cell->side) {
    case QDDB_SCRIPT_SIDE_RIGHT:
	side_opts = "-side right";
	break;
    case QDDB_SCRIPT_SIDE_TOP:
	side_opts = "-side top";
	break;
    case QDDB_SCRIPT_SIDE_BOTTOM:
	side_opts = "-side bottom";
	break;
    case QDDB_SCRIPT_SIDE_LEFT:
    default:
	side_opts = "-side left";
	break;
    }
    switch (cell->anchor) {
    case QDDB_SCRIPT_ANCHOR_W:
	anchor_opts = "-anchor w";
	break;
    case QDDB_SCRIPT_ANCHOR_E:
	anchor_opts = "-anchor e";
	break;
    case QDDB_SCRIPT_ANCHOR_N:
	anchor_opts = "-anchor n";
	break;
    case QDDB_SCRIPT_ANCHOR_S:
	anchor_opts = "-anchor s";
	break;
    case QDDB_SCRIPT_ANCHOR_NW:
	anchor_opts = "-anchor nw";
	break;
    case QDDB_SCRIPT_ANCHOR_NE:
	anchor_opts = "-anchor ne";
	break;
    case QDDB_SCRIPT_ANCHOR_SW:
	anchor_opts = "-anchor sw";
	break;
    case QDDB_SCRIPT_ANCHOR_SE:
	anchor_opts = "-anchor se";
	break;
    case QDDB_SCRIPT_ANCHOR_CENTER:
	anchor_opts = "-anchor center";
	break;
    default:
	anchor_opts = "";
    }

    /* padding */
    pad_opts[0] = '\0';
    if (cell->padx != 0) {
	sprintf(buf, "-padx %dm ", cell->padx);
	strcat(pad_opts, buf);
    }
    if (cell->pady != 0) {
	sprintf(buf, "-pady %dm ", cell->pady);
	strcat(pad_opts, buf);
    }
    if (cell->ipadx != 0) {
	sprintf(buf, "-ipadx %dm ", cell->ipadx);
	strcat(pad_opts, buf);
    }
    if (cell->ipady != 0) {
	sprintf(buf, "-ipady %dm ", cell->ipady);
	strcat(pad_opts, buf);
    }

    sprintf(buf, "pack %s %s %s %s %s\n", frame, expand_opts, side_opts, pad_opts, anchor_opts);
    Qddb_ConcatBuffer(buf);
}

static void TclQddb_ScriptConfigure(frame, cell)
    char                *frame;
    Qddb_ScriptCell     *cell;
{
    char                *relief_opt;
    char                buf[BUFSIZ];

    switch (cell->relief) {
    case QDDB_SCRIPT_RELIEF_SUNKEN:
	relief_opt = "sunken";
	break;
    case QDDB_SCRIPT_RELIEF_RAISED:
	relief_opt = "raised";
	break;
    case QDDB_SCRIPT_RELIEF_GROOVE:
	relief_opt = "groove";
	break;
    case QDDB_SCRIPT_RELIEF_FLAT:
    default:
	relief_opt = "flat";
    }
    sprintf(buf, "%s configure -relief %s -bd %d\n", frame, relief_opt, cell->border);
    Qddb_ConcatBuffer(buf);
}

static void TclQddb_ScriptConfigureFx(frame, cell)
    char                *frame;
    Qddb_ScriptCell     *cell;
{
    char                buf[BUFSIZ], *ptr;
    int                 size, flags;

    if (cell->label != NULL) {
	flags = 0;
	if (cell->label[0] != '@') {
	    size = Tcl_ScanElement(cell->label, &flags);
	    ptr = (char *)Malloc((size_t)size);
	    Tcl_ConvertElement(cell->label, ptr, flags);
	    sprintf(buf, "%s.f_0.l configure -text %s\n", frame, ptr);
	    Free(ptr);
	} else {
	    sprintf(buf, "%s.f_0.l configure -bitmap %s\n", frame, cell->label);
	}
	Qddb_ConcatBuffer(buf);
    }
    if (cell->labelwidth != 0) {
	sprintf(buf, "%s.f_0.l configure -width %d\n", frame, cell->labelwidth);
	Qddb_ConcatBuffer(buf);
    }
}

static void TclQddb_AddWaitStack(stack, string, level, isframe)
    Qddb_ScriptAttrStack        **stack;
    char                        *string;
    size_t                      level;
    Qddb_Boolean                isframe;
{
    Qddb_ScriptAttrStack        *ptr;

    if (string == NULL)
	return;
    if (*stack == NULL) {
	*stack = (Qddb_ScriptAttrStack *)Malloc(sizeof(Qddb_ScriptAttrStack));
	(*stack)->attr = string;
	(*stack)->isframe = isframe;
	(*stack)->level = level;
	(*stack)->next = NULL;
    } else {
	ptr = *stack;
	while (ptr->next != NULL) {
	    ptr = ptr->next;
	}
	ptr->next = (Qddb_ScriptAttrStack *)Malloc(sizeof(Qddb_ScriptAttrStack));
	ptr->next->attr = string;
	ptr->next->isframe = isframe;
	ptr->next->level = level;
	ptr->next->next = NULL;
    }
}

