
/* $RCSfile: TclUtil.c,v $ $Revision: 2.1 $ $Date: 2003/03/23 17:16:17 $
 *
 * Copyright (C) 1995-2003 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include "tcl.h"
#include "Qddb.h"
#include "tclQddb.h"

/* qddb_util --
 *	qddb_util isdate <value>
 *      qddb_util istime <value>
 *      qddb_util postscript ?option value ...? string
 *	qddb_util formatdate <format> <value>
 *      qddb_util formattime <format> <value>
 *      qddb_util datestyle ?us|european?
 *      qddb_util isregexp <value>
 *      qddb_util flock <filedesc> ?type?
 *      qddb_util catrelation <schema desc> ?offset length?
 *      qddb_util pr ?options? string 
 */

static char *commands_string = "isdate, istime, postscript, formatdate, \
formattime, isregexp, flock, catrelation, or pr";

static int PrUtil _ANSI_ARGS_((Tcl_Interp *, int, CONST char **));
static int GetPostscriptOptions _ANSI_ARGS_((Tcl_Interp *, int, CONST char **, \
					     Qddb_PostscriptOptions *));

int TclQddb_UtilProc(clientData, interp, argc, argv)
    ClientData			clientData;
    Tcl_Interp			*interp;
    int				argc;
    CONST char			*argv[];
{
    if (argc < 2) {
	Tcl_AppendResult(interp, argv[0], ": wrong # args", NULL);
	return TCL_ERROR;
    }
    if (strcmp(argv[1], "isdate") == 0 && argc == 3) {
	if (Qddb_CheckDate(argv[2]) != 0) {
	    Tcl_SetResult(interp, "0", TCL_STATIC);
	} else {
	    Tcl_SetResult(interp, "1", TCL_STATIC);
	}
	return TCL_OK;
    } else if (strcmp(argv[1], "istime") == 0 && argc == 3) {
	if (Qddb_CheckTime(argv[2]) != 0) {
	    Tcl_SetResult(interp, "0", TCL_STATIC);
	} else {
	    Tcl_SetResult(interp, "1", TCL_STATIC);
	}
	return TCL_OK;
    } else if (strcmp(argv[1], "postscript") == 0 && argc >= 3) {
	/* qddb_util postscript ?options? string */
	Qddb_PostscriptOptions     opt;
	char                       *retval;

	opt.lines_per_page = QDDB_POSTSCRIPT_LINES_PER_PAGE;
	opt.tab_width_chars = QDDB_POSTSCRIPT_TAB_WIDTH;
	opt.orient = QDDB_POSTSCRIPT_ORIENT_PORTRAIT;
	opt.font = QDDB_POSTSCRIPT_FONT;
	opt.bold_font = QDDB_POSTSCRIPT_BOLD_FONT;
	opt.chars_per_inch = QDDB_POSTSCRIPT_CHARS_PER_INCH;
	opt.page_length_pts = QDDB_POSTSCRIPT_PAGE_LENGTH_LETTER;
	opt.baseline_offset_pts = QDDB_POSTSCRIPT_BASELINE_OFFSET;
	opt.vertical_spacing_pts = QDDB_POSTSCRIPT_VERTICAL_SPACING;
	opt.left_margin_pts = QDDB_POSTSCRIPT_LEFT_MARGIN;
	GetPostscriptOptions(interp, argc, argv, &opt);
	opt.char_width_pts = 72.0/(double)opt.chars_per_inch;
	retval = Qddb_TextToPostscript(argv[argc-1], &opt);
	Tcl_SetResult(interp, retval, TCL_DYNAMIC);
	return TCL_OK;
    } else if (strcmp(argv[1], "formatdate") == 0 && argc == 4) {
	/* qddb_util formatdate format date */
	char		*date, *time;

	time = Qddb_DateStringToTime(argv[3]);
	if (strlen(time) == 0) {
	    Free(time);
	    Tcl_AppendResult(interp, argv[0], ": date conversion failed, bad input: ", argv[3], NULL);
	    return TCL_ERROR;
	}
	date = Qddb_DateTimeToString(time, argv[2]);
	Free(time);
	if (strlen(date) >= 1) {
	    Tcl_SetResult(interp, date, TCL_DYNAMIC);
	} else {
	    Free(date);
	    Tcl_AppendResult(interp, argv[0], ": date conversion failed, bad format: ", argv[2], NULL);
	    return TCL_ERROR;
	}
	return TCL_OK;
    } else if (strcmp(argv[1], "formattime") == 0 && argc == 4) {
	/* qddb_util formattime format time */
        char             *time;
        int              seconds;

	seconds = Qddb_TimeStringToSeconds(argv[3]);
	if (seconds < 0) {
	    Tcl_AppendResult(interp, argv[0], ": time conversion failed, bad input: ", argv[3], NULL);
	    return TCL_ERROR;
	}
	time = Qddb_SecondsToTimeString(seconds, argv[2]);
	if (strlen(time) >= 1) {
	    Tcl_SetResult(interp, time, TCL_DYNAMIC);
	} else {
	    Free(time);
	    Tcl_AppendResult(interp, argv[0], 
			     ": time conversion failed, bad format: ", 
			     argv[2], NULL);
	    return TCL_ERROR;
	}
	return TCL_OK;
    } else if (strcmp(argv[1], "datestyle") == 0) {
	if (argc == 2) {
	    if (qddb_datestyle == QDDB_DATE_STYLE_US) {
		Tcl_SetResult(interp, "us", TCL_STATIC);
		return TCL_OK;
	    } else {
		Tcl_SetResult(interp, "european", TCL_STATIC);
		return TCL_OK;
	    }
	} else if (argc == 3) {
	    if (strcmp(argv[2], "us") == 0) {
		qddb_default_date_format = "%m/%d/%y";
		qddb_datestyle = QDDB_DATE_STYLE_US;
	    } else if (strcmp(argv[2], "european") == 0) {
		qddb_default_date_format = "%d/%m/%y";
		qddb_datestyle = QDDB_DATE_STYLE_EUROPEAN;
	    } else {
		Tcl_AppendResult(interp, argv[0], " ", argv[1], ": bad date style \"", 
				 argv[2], "\"", NULL);
		return TCL_ERROR;
	    }
	    return TCL_OK;
	} else {
	    Tcl_AppendResult(interp, argv[0], " ", argv[1], ": wrong # args", NULL);
	    return TCL_ERROR;
	}
    } else if (strcmp(argv[1], "isregexp") == 0 && argc == 3) {
	Qddb_Regexp    *reg;
	char           *reg_string;
	
	reg_string = (char *)Malloc(strlen(argv[2])+1);
	strcpy(reg_string, argv[2]);
	reg = Qddb_Regcomp(reg_string);
	Free(reg_string);
	if (reg == NULL) {
	    Tcl_SetResult(interp, "0", TCL_STATIC);
	} else {
	    Free(reg);
	    Tcl_SetResult(interp, "1", TCL_STATIC);
	}
	return TCL_OK;
    } else if (strcmp(argv[1], "flock") == 0 && argc == 4) {
	FILE			*fp;
	int			flags = 0;
	char			*ptr, *tok_string;

#if TCL_MAJOR_VERSION >= 8
	if (Tcl_GetOpenFile(interp, argv[2], 0, 0, (ClientData *)&fp) != TCL_OK) {
#elif TCL_MAJOR_VERSION >= 4 && TCL_MINOR_VERSION >= 1
	if (Tcl_GetOpenFile(interp, argv[2], 0, 0, (ClientData)&fp) != TCL_OK) {
#else
	if (Tcl_GetOpenFile(interp, argv[2], 0, 0, &fp) != TCL_OK) {
#endif
	    Tcl_AppendResult(interp, argv[0], " ", argv[1], ": cannot get open file \"", 
			     argv[2], "\"", NULL);
	    return TCL_ERROR;
	}
	tok_string = (char *)Malloc(strlen(argv[3])+1);
	strcpy(tok_string, argv[3]);
	ptr = strtok(tok_string, "|");
	while (ptr != NULL) {
	    if (strcmp(ptr, "LOCK_SH") == 0) {
		flags |= LOCK_SH;
	    } else if (strcmp(ptr, "LOCK_EX") == 0) {
		flags |= LOCK_EX;
	    } else if (strcmp(ptr, "LOCK_NB") == 0) {
		flags |= LOCK_NB;
	    } else if (strcmp(ptr, "LOCK_UN") == 0) {
		flags |= LOCK_UN;
	    } else {
		Tcl_AppendResult(interp, argv[0], " ", argv[1], ": bad option \"", ptr, "\"", NULL);
		return TCL_ERROR;
	    }
	    ptr = strtok(NULL, "|");
	}
	Free(tok_string);
#if defined(HAVE_FLOCK)
	if (flock(fileno(fp), flags) != 0) {
	    Tcl_SetResult(interp, "0", TCL_STATIC);
	} else {
	    Tcl_SetResult(interp, "1", TCL_STATIC);
	}
#else
	if ((flags&LOCK_UN) != 0) {
	    if (Qddb_UnlockSection(fileno(fp), 0, 0) != 0) {
		Tcl_SetResult(interp, "0", TCL_STATIC);
	    } else {
		Tcl_SetResult(interp, "1", TCL_STATIC);
	    }
	} else {
	    if (Qddb_LockSection(fileno(fp), (flags&LOCK_SH) != 0? F_RDLCK:F_WRLCK, 0, 0, 
			    (flags&LOCK_NB) != 0? 0:1) != 0) {
		Tcl_SetResult(interp, "0", TCL_STATIC);
	    } else {
		Tcl_SetResult(interp, "1", TCL_STATIC);
	    }
	}
#endif
	return TCL_OK;
    } else if (strcmp(argv[1], "catrelation") == 0) {
	Qddb_Schema         *schema;
	Qddb_KeyList        *keylist;
	off_t               offset;
	size_t              how_many;

	if (argc != 3 && argc != 5) {
	    Tcl_AppendResult(interp, argv[0], " ", argv[1], ": wrong # args", NULL);
	    return TCL_ERROR;
	}
	if ((schema = TclQddb_GetSchema(argv[2])) == NULL) {
	    Tcl_AppendResult(interp, argv[0], " ", argv[1], ": cannot open schema \"", 
			     argv[2], "\"", NULL);
	    return TCL_ERROR;
	}
	if (argc == 5) {
	    offset = Qddb_BaseToOffset(argv[3], 10);
	    how_many = Qddb_BaseToSize(argv[4], 10);
	} else {
	    offset = 0;
	    how_many = 0;
	}
	keylist = Qddb_CatRelation(schema, offset, how_many);
	if (TclQddb_NewKeyList(interp, argv[2], keylist, -1, 0) != TCL_OK) {
	    return TCL_ERROR;
	}
	return TCL_OK;
    } else if (strcmp(argv[1], "pr") == 0 && argc >= 3) {
	/* qddb_util pr ?options? string 
	 *     -header on|off
	 *     -footer on|off
	 *     -formfeed on|off
	 */
	if (PrUtil(interp, argc, argv) != TCL_OK) {
	    return TCL_ERROR;
	}
	return TCL_OK;
    }
    Tcl_AppendResult(interp, argv[0], ": bad command \"", argv[1], "\": should be ",
		     commands_string, NULL);
    return TCL_ERROR;
}

static int GetPostscriptOptions(interp, argc, argv, opt)
    Tcl_Interp             *interp;
    int                    argc;
    CONST char             **argv;
    Qddb_PostscriptOptions *opt;
{
    int                    i;

    /* -lines
     * -cpi
     * -pagelength
     * -baseline
     * -verticalspacing
     * -leftmargin
     * -papertype
     */
    for (i = 2; i < argc; i+=2) {
	if (argv[i][0] == '-') {
	    if (strcmp(argv[i], "-lines") == 0) {
		opt->lines_per_page = strtol(argv[i+1], NULL, 10);
	    } else if (strcmp(argv[i], "-cpi") == 0) {
		opt->chars_per_inch = strtod(argv[i+1], NULL);
	    } else if (strcmp(argv[i], "-papertype") == 0) {
		if (strcmp(argv[i+1], "letter") == 0) {
		    opt->page_length_pts = QDDB_POSTSCRIPT_PAGE_LENGTH_LETTER;
		} else if (strcmp(argv[i+1], "legal") == 0) {
		    opt->page_length_pts = QDDB_POSTSCRIPT_PAGE_LENGTH_LEGAL;
		} else if (strcmp(argv[i+1], "a4") == 0) {
		    opt->page_length_pts = QDDB_POSTSCRIPT_PAGE_LENGTH_A4;
		} else {
		    Tcl_AppendResult(interp, argv[0], " ", argv[1], ": bad arg for option ", argv[i], NULL);
		    return TCL_ERROR;
		}
	    } else if (strcmp(argv[i], "-orient") == 0) {
		if (strcmp(argv[i+1], "portrait") == 0) {
		    opt->orient = QDDB_POSTSCRIPT_ORIENT_PORTRAIT;
		} else if (strcmp(argv[i+1], "landscape") == 0) {
		    opt->orient = QDDB_POSTSCRIPT_ORIENT_LANDSCAPE;
		} else {
		    Tcl_AppendResult(interp, argv[0], " ", argv[1], ": bad arg for option ", argv[i], NULL);
		    return TCL_ERROR;
		}
	    } else if (strcmp(argv[i], "-baseline") == 0) {
		opt->baseline_offset_pts = Qddb_PrinterPoints(argv[i+1]);
	    } else if (strcmp(argv[i], "-vs") == 0 || strcmp(argv[i], "-verticalspacing") == 0) {
		opt->vertical_spacing_pts = Qddb_PrinterPoints(argv[i+1]);
	    } else if (strcmp(argv[i], "-leftmargin") == 0) {
		opt->left_margin_pts = Qddb_PrinterPoints(argv[i+1]);
	    } else {
		Tcl_AppendResult(interp, argv[0], " ", argv[1], ": bad option ", argv[i], NULL);
		return TCL_ERROR;
	    }
	}
    }
    return TCL_OK;
}

static int PrUtil(interp, argc, argv)
    Tcl_Interp            *interp;
    int                   argc;
    CONST char            **argv;
{
    int                   i;
    int                   headers, footers, formfeed, linesperpage;
    size_t                pageno, lineno;
    CONST char            *dateformat, *firstofline;
    char                  *time, *time2, *date, *ch, tmpbuf[BUFSIZ];

    headers = footers = formfeed = 0;
    linesperpage = 66;
    dateformat = qddb_default_date_format;
    for (i = 2; i < argc-1; i+=2) {
	if (strcmp(argv[i], "-headers") == 0) {
	    if (Tcl_GetBoolean(interp, argv[i+1], &headers) != TCL_OK) {
		return TCL_ERROR;
	    }
	} else if (strcmp(argv[i], "-footers") == 0) {
	    if (Tcl_GetBoolean(interp, argv[i+1], &footers) != TCL_OK) {
		return TCL_ERROR;
	    }
	} else if (strcmp(argv[i], "-formfeed") == 0) {
	    if (Tcl_GetBoolean(interp, argv[i+1], &formfeed) != TCL_OK) {
		return TCL_ERROR;
	    }
	} else if (strcmp(argv[i], "-linesperpage") == 0) {
	    if (Tcl_GetInt(interp, argv[i+1], &linesperpage) != TCL_OK) {
		return TCL_ERROR;
	    }
	} else if (strcmp(argv[i], "-dateformat") == 0) {
	    if (strlen(argv[i+1]) < BUFSIZ-20) {
		dateformat = argv[i+1];
	    }
	}
    }
    pageno = 1;
    lineno = 1;
    Qddb_InitBuffer();
    time = Qddb_DateStringToTime("now");
    if (strlen(time) == 0) {
	Free(time);
	Tcl_AppendResult(interp, argv[0], ": could not get current date/time", NULL);
	return TCL_ERROR;
    }
    date = Qddb_DateTimeToString(time, dateformat);
    time2 = Qddb_DateTimeToString(time, "%H:%M");
    Free(time);
    ch = (char *)Malloc(strlen(argv[argc-1])+1);
    strcpy(ch, argv[argc-1]);
    firstofline = ch;
    for (; *ch != '\0';) {
	if (headers != 0 && lineno == 1) {
	    Qddb_ConcatBuffer("\n");
	    sprintf(tmpbuf, "%s %s  Page %d\n\n\n", date, time2, pageno);
	    Qddb_ConcatBuffer(tmpbuf);
	    lineno += 4;
	    firstofline = ch;
	} else if ((footers != 0 && lineno == linesperpage-5) || (lineno == linesperpage)) {
	    if (footers != 0) {
		if (formfeed != 0) {
		    Qddb_ConcatBuffer("\f");
		} else {
		    Qddb_ConcatBuffer("\n\n\n\n\n");
		}
	    } else if (formfeed != 0) {
		Qddb_ConcatBuffer("\f");
	    } else {
		Qddb_ConcatBuffer("\n");
	    }
	    pageno++;
	    lineno = 1;
	    firstofline = ch;
	} else {
	    if (*ch == '\n') {
		lineno++;
		*ch = '\0';
		Qddb_ConcatBuffer(firstofline);
		Qddb_ConcatBuffer("\n");
		*ch = '\n';
		ch++;
		firstofline = ch;
	    } else {
		ch++;
	    }
	}
    }
    Free(ch);
    if (formfeed != 0) {
	if (lineno < linesperpage) {
	    Qddb_ConcatBuffer("\f");
	}
    } else {
	for (i = lineno; i <= linesperpage; i++) {
	    Qddb_ConcatBuffer("\n");
	}
    }
    Free(time2);
    Free(date);
    Tcl_SetResult(interp, Qddb_GetBufferWithoutResize(), TCL_DYNAMIC);
    return TCL_OK;
}
