
/* $RCSfile: TclUtils.c,v $ $Revision: 2.1 $ $Date: 2003/03/23 17:16:17 $
 *
 * Copyright (C) 1995-2003 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include "tcl.h"
#include "Qddb.h"
#include "tclQddb.h"

void TclQddb_FreeArgs(argc, argv)
    int				argc;
    char			*argv[];
{
    int				i;

    if (argc > 0) {
	for (i = 0; i < argc; i++)
	    Free(argv[i]);
    } else {
	for (i = 0; argv[i] != NULL; i++)
	    Free(argv[i]);
    }
    Free(argv);
}

char **TclQddb_SplitAttributeName(name)
    CONST char			*name;
{
    char			**retval, *newname;
    CONST char                  *chidx;
    int				i, numdots;

    numdots = 0;
    for (chidx = name; *chidx != '\0'; chidx++)
	if (*chidx == '.')
	    numdots++;
    retval = (char **)Malloc(sizeof(char *)*(numdots+2));
    retval[numdots+1] = NULL;
    newname = Malloc(strlen(name)+1);
    strcpy(newname, name);
    chidx = strtok(newname, ".");
    if (chidx == NULL) {
	retval[0] = newname;
	retval[1] = '\0';
	return retval;
    }
    for (i = 0; i <= numdots; i++) {
	retval[i] = Malloc(strlen(chidx)+1);
	strcpy(retval[i], chidx);
	chidx = strtok(NULL, ".");
    }
    Free(newname);
    return retval;
}

char *TclQddb_ConcatAttributeName(name_list, num)
    char			**name_list;
    int				num;
{
    int				i;
    size_t			len;
    char			*retval;

    for (len = 0, i = 0; i < num; i++)
	len += strlen(name_list[i]);
    retval = Malloc(len+num);
    retval[0] = '\0';
    for (i = 0; i < num-1; i++) {
	strcat(retval, name_list[i]);
	strcat(retval, ".");
    }
    strcat(retval, name_list[i]);
    return retval;
}

int TclQddb_FindAttribute(tree, name)
    Qddb_DataTree		**tree;
    char			*name;
{
    int				i;

    for (i = 0; tree[i] != NULL; i++)
	if (strcmp(tree[i]->datatree_schema->schemanode->Name, name) == 0)
	    return i;
    return -1;
}

char **TclQddb_ParseAttr(name)
    char			*name;
{
    int				numdots, i;
    size_t			len;
    char			**retval;

    if (name == NULL)
	return NULL;
    retval = (char **)Calloc(sizeof(char *)*(MAXIMUM_NUMBER_OF_SCHEMA_LEVELS+1));
    len = strlen(name);
    for (i = 0, numdots = 0; i < len; i++) {
	if (name[i] == '.') {
	    name[i] = '\0';
	    retval[numdots] = Malloc((size_t)i+1);
	    strcpy(retval[numdots++], name);
	    name[i] = '.';
	}
    }
    retval[numdots] = Malloc(len+1);
    strcpy(retval[numdots], name);
    retval[numdots+1] = NULL;
    return retval;
}

int TclQddb_CountAttrs(attrs)
    char			*attrs;
{
    int				i, numdots;
    if (attrs[0] == '\0')
	return 0;
    for (i = 0, numdots = 0; attrs[i] != '\0'; i++)
	if (attrs[i] == '.')
	    numdots++;
    return ++numdots;
}

char *TclQddb_StringRef(tuple, rownode)
    TclQddb_Tuple		*tuple;
    TclQddb_Rows		*rownode;
{
    Qddb_DataTree		**dt;
    char			**attrs, *retval = NULL;
    size_t			inst[MAXIMUM_NUMBER_OF_SCHEMA_LEVELS];
    int				num, nextattr;
    int				i;

    attrs = TclQddb_SplitAttributeName(rownode->attr_name);
    num = Qddb_StringToInstance(rownode->instance, inst);
    dt = tuple->datatree;
    i = 0;
    while ((nextattr = TclQddb_FindAttribute(dt, attrs[i])) != -1) {
	if (i >= num) {
	    TclQddb_FreeArgs(-1, attrs);
	    return NULL;
	}
	if (dt[nextattr]->datatree_type == DATATREE_CHILDREN)
	    dt = dt[nextattr][inst[i] - 1].datatree_children;
	else {
	    retval = TclQddb_TreeDataToString(dt[nextattr] + inst[i] - 1);
	    break;
	}
	i++;
    }
    TclQddb_FreeArgs(-1, attrs);
    return retval;
}

Qddb_DataTree *TclQddb_DataTreeRef(tuple, rownode)
    TclQddb_Tuple		*tuple;
    TclQddb_Rows		*rownode;
{
    Qddb_DataTree		**dt, *retval = NULL;
    char			**attrs;
    size_t			inst[MAXIMUM_NUMBER_OF_SCHEMA_LEVELS];
    int				num, nextattr;
    int				i;

    attrs = TclQddb_SplitAttributeName(rownode->attr_name);
    num = Qddb_StringToInstance(rownode->instance, inst);
    dt = tuple->datatree;
    i = 0;
    while ((nextattr = TclQddb_FindAttribute(dt, attrs[i])) != -1) {
	if (i >= num) {
	    TclQddb_FreeArgs(-1, attrs);
	    return NULL;
	}
	if (dt[nextattr]->datatree_type == DATATREE_CHILDREN)
	    dt = dt[nextattr][inst[i] - 1].datatree_children;
	else {
	    retval = dt[nextattr] + inst[i] - 1;
	    break;
	}
	i++;
    }
    TclQddb_FreeArgs(-1, attrs);
    return retval;
}

char *TclQddb_TreeDataToString(tree_entry)
    Qddb_DataTree		*tree_entry;
{
    char			*retval;
    char			buf[4*1024];

    switch(tree_entry->datatree_type) {
    case DATATREE_STRING:
        if (tree_entry->datatree_string == NULL)
	    PANIC("type datatree_string is NULL");
        retval = Malloc(strlen(tree_entry->datatree_string)+1);
        strcpy(retval, tree_entry->datatree_string);
        break;
    case DATATREE_DATE:
	retval = Malloc(strlen(tree_entry->datatree_date)+1);
	strcpy(retval, tree_entry->datatree_date);
	break;
    case DATATREE_INT:
	if (tree_entry->modified == False) {
	    retval = Calloc(1);
	} else {
	    sprintf(buf, "%d", tree_entry->datatree_int);
	    retval = Malloc(strlen(buf)+1);
	    strcpy(retval, buf);
	}
	break;
    case DATATREE_REAL:
	if (tree_entry->modified == False) {
	    retval = Calloc(1);
	} else {
	    sprintf(buf, "%f", tree_entry->datatree_real);
	    retval = Malloc(strlen(buf)+1);
	    strcpy(retval, buf);
	}
	break;
    case DATATREE_NOVAL:
    default:
	return NULL;
    }
    return retval;
}


Qddb_DataTree *TclQddb_StringToTreeData(node, value)
    Qddb_SchemaTreeNode		*node;
    char			*value;
{
    Qddb_DataTree		*retval;

    retval = (Qddb_DataTree *)Malloc(sizeof(Qddb_DataTree));
    switch(node->schemanode->Type) {
    case QAT_String:
	retval->datatree_type = DATATREE_STRING;
	retval->datatree_string = Malloc(strlen(value)+1);
	retval->datatree_schema = node;
	strcpy(retval->datatree_string, value);
	break;
    case QAT_Date:
	retval->datatree_type = DATATREE_DATE;
	retval->datatree_date = Malloc(strlen(value)+1);
	retval->datatree_schema = node;
	strcpy(retval->datatree_date, value);
	break;
    case QAT_Time:
	retval->datatree_type = DATATREE_TIME;
	retval->datatree_time = Malloc(strlen(value)+1);
	retval->datatree_schema = node;
	strcpy(retval->datatree_time, value);
	break;
    case QAT_Integer:
	retval->datatree_type = DATATREE_DATE;
	if (value[0] == '\0') {
	    retval->modified = False;
	    retval->datatree_int = 0;
	} else {
	    retval->modified = True;
	    retval->datatree_int = Qddb_GetInt(value);
	}
	retval->datatree_schema = node;
	break;
    case QAT_Real:
	retval->datatree_type = DATATREE_REAL;
	if (value[0] == '\0') {
	    retval->modified = False;
	    retval->datatree_real = 0.00;
	} else {
	    retval->modified = True;
	    retval->datatree_real = strtod(value, NULL);
	}
	retval->datatree_schema = node;
    default:
	PANIC("TclQddb_StringToTreeData");
    }
    return retval;
}


void TclQddb_NullOutNode(dt, do_realloc)
    Qddb_DataTree	*dt;
    int			do_realloc;
{
    switch (dt->datatree_type) {
    case DATATREE_INT:
	dt->modified = False;
	dt->datatree_int = 0.0;
	break;
    case DATATREE_REAL:
	dt->modified = False;
	dt->datatree_real = 0.0;
	break;
    case DATATREE_DATE:
	if (dt->datatree_date != NULL)
	    Free(dt->datatree_date);
	if (do_realloc)
	    dt->datatree_date = Calloc(1);
	else
	    dt->datatree_date = NULL;
	break;
    case DATATREE_STRING:
	if (dt->datatree_string != NULL)
	    Free(dt->datatree_string);
	if (do_realloc)
	    dt->datatree_string = Calloc(1);
	else
	    dt->datatree_string = NULL;
	break;
    default:
	return;
    }
}

Qddb_Boolean TclQddb_IsEmptyDataTree(dt)
    Qddb_DataTree	**dt;
{
    Qddb_DataTree	*inst;
    int			i;

    while (*dt != NULL) {
	inst = *dt;
	for (i = 0; inst[i].datatree_type != DATATREE_END; i++) {
	    switch (inst[i].datatree_type) {
	    case DATATREE_INT:
		if (inst[i].modified == True)
		    return False;
		break;
	    case DATATREE_REAL:
		if (inst[i].modified == True)
		    return False;
		break;
	    case DATATREE_DATE:
		if (inst[i].datatree_date[0] != '\0')
		    return False;
		break;
	    case DATATREE_STRING:
		if (inst[i].datatree_string[0] != '\0')
		    return False;
		break;
	    case DATATREE_CHILDREN:
		if (TclQddb_IsEmptyDataTree(inst[i].datatree_children) == False)
		    return False;
	    default:
		break;
	    }
	}
	dt++;
    }
    return True;
}
