# $RCSfile: fx_chartfunc.tcl,v $ $Revision: 2.0.0.1 $ $Date: 2000/11/09 15:06:00 $
#
# Copyright (C) 1995-2000 Herrin Software Development, Inc.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# Generic charting (graph/barchart) routines

proc Fx:DisplayNumericSelectProc {sel} {
    global fx_display_numeric_search_listbox_retval

    set fx_display_numeric_search_listbox_retval 1
}

proc Fx:CheckNumericSearch {s varname {index ""}} {
    global fx_display_leaves fxPriv fx_rconfig $varname

    if {![info exists fx_rconfig(\$search\$,print)]} {
	return {}
    }
    set display_attrs {}
    set fx_display_attrs $fx_rconfig(\$search\$,print)
    if {[info exists fx_rconfig(\$search\$,user_defs)]} {
	foreach i [set fx_rconfig(\$search\$,user_defs)] {
	    set def [lindex $i 1]
	    set user_defs([lindex $def 0]) [lindex $def 1]
	}
    }
    foreach i $fx_display_attrs {
	if {[string compare $i [set ${varname}($index)]] == 0} {
	    if {[info exists user_defs($i)]} {
		set verb $user_defs($i)
		set type real
	    } else {
		if {[catch "qddb_schema option type $s $i" type] != 0} {
		    continue
		}
		set verb [qddb_schema option verbosename $s $i]
		if {[llength $verb] == 0} {
		    set verb [split $i .]
		    set verb [lindex $verb [expr [llength $verb] - 1]]
		}
	    }
	    return [list $verb $i $type 1]
	}
	if {[info exists fx_display_info]} {continue}
	if {[info exists user_defs($i)]} {
	    set verb $user_defs($i)
	    set type real
	} else {
	    if {[catch "qddb_schema option type $s $i" type] != 0} {
		continue
	    }
	    if {[string compare $index x] != 0} {
		if {[string compare $type integer] != 0 && \
			[string compare $type real] != 0} {
		    continue
		}
	    }
	    set verb [qddb_schema option verbosename $s $i]
	    if {[llength $verb] == 0} {
		set verb [split $i .]
		set verb [lindex $verb [expr [llength $verb] - 1]]
	    }
	}
	set fx_display_info [list $verb $i $type 0]
    }
    return $fx_display_info
}

proc Fx:DisplayNumericSearch {w b s varname {index ""} {varname2 ""} {index2 ""} {func ""}} {
    global fx_display_numeric_search_listbox_retval \
	    fx_display_leaves fxPriv fx_rconfig $varname

    if {![info exists fx_rconfig(\$search\$,print)]} {
	return
    }
    set fx_display_info {}
    set display_attrs {}
    set fx_display_attrs $fx_rconfig(\$search\$,print)
    if {[info exists fx_rconfig(\$search\$,user_defs)]} {
	foreach i [set fx_rconfig(\$search\$,user_defs)] {
	    set def [lindex $i 1]
	    set user_defs([lindex $def 0]) [lindex $def 1]
	}
    }
    foreach i $fx_display_attrs {
	if {[info exists user_defs($i)]} {
	    set verb $user_defs($i)
	} else {
	    if {[catch "qddb_schema option type $s $i" type] != 0} {
		continue
	    }
	    if {[string compare $index x] != 0} {
		if {[string compare $type integer] != 0 && \
			[string compare $type real] != 0} {
		    continue
		}
	    }
	    set verb [qddb_schema option verbosename $s $i]
	    set type [qddb_schema option type $s $i]
	    if {[llength $verb] == 0} {
		set verb [split $i .]
		set verb [lindex $verb [expr [llength $verb] - 1]]
	    }
	}
	set attr_types($i) $type
	lappend fx_display_info [list $verb $i]
	lappend display_attrs $i
    }
    if {[llength $fx_display_info] == 0} {
	fx_dialog .fx_reportgen.dialog \
		"No numeric fields specified" "You must specify numeric fields before graphing!" \
		error 0 Ok
	return
    }
    set oldGrab [grab current $w]
    set oldFocus [focus]
    set parent [winfo toplevel $b]
    set f $parent.numsearch
    toplevel $f -cursor arrow
    wm transient $f $parent
    wm overrideredirect $f 1
    wm withdraw $f

    frame $f.ft -takefocus 0 -highlightthickness 0
    pack $f.ft -side top -expand on -fill both
    set x $f.ft
    Fx_MultiColumnListBox fx_display_numeric_search_listbox -w $x.f -side top -numcols 2 \
	    -width 25 -headings {Verbosename Attribute} \
	    -align {left left} -single_select on -separators {" " " "} \
	    -initselection on \
	    -height 5 -onselect Fx:DisplayNumericSelectProc
    frame $f.b -takefocus 0 -highlightthickness 0 -relief groove -bd 3
    pack $f.b -side bottom -expand on -fill x
    button $f.b.select -text Select -command {set fx_display_numeric_search_listbox_retval 1}
    pack $f.b.select -side left -padx 4m -pady 2m
    button $f.b.cancel -text Cancel -command {set fx_display_numeric_search_listbox_retval 0}
    pack $f.b.cancel -side right -padx 4m -pady 2m
    fx_display_numeric_search_listbox AppendRows $fx_display_info
    fx_display_numeric_search_listbox BuildFormats 1
    fx_display_numeric_search_listbox Format
    fx_display_numeric_search_listbox Display    
    set rootx [winfo rootx $b]
    set rooty [expr [winfo rooty $b]+[winfo height $b]]
    set screenheight [winfo screenheight $b]
    set screenwidth [winfo screenwidth $b]
    update idletasks
    set reqheight [winfo reqheight $f]
    set reqwidth [winfo reqwidth $f]
    if {[expr $rooty + $reqheight] > $screenheight} {
	set rooty [expr $rooty + $screenheight - $rooty - $reqheight]
    }
    if {[expr $rootx + $reqwidth] > $screenwidth} {
	set rootx [expr $rootx + $screenwidth - $rootx - $reqwidth]
    }
    wm geometry $f "+$rootx+$rooty"
    bind $f <Escape> "global fx_display_numeric_search_listbox_retval; \
	    set fx_display_numeric_search_listbox_retval 0"
    bind $x.f.l2 <Escape> "global fx_display_numeric_search_listbox_retval; \
	    set fx_display_numeric_search_listbox_retval 0"
    bind $f <Enter> [list set fxPriv(popup_listbox) ""]
    bind $f <Leave> [list set fxPriv(popup_listbox) $f]
    set fxPriv(popup_listbox) $f
    bind $f <ButtonPress> {
	if {[info exists fxPriv(popup_listbox)]} {
	    if {[string compare $fxPriv(popup_listbox) ""] != 0} {
		set fx_display_numeric_search_listbox_retval 0
	    }
	}
    }
    wm deiconify $f
    tkwait visibility $f
    catch "grab -global $f"
    focus $x.f.l2
    tkwait variable fx_display_numeric_search_listbox_retval
    if {$fx_display_numeric_search_listbox_retval} {
	set sel [$x.f.l2 curselection]
	if {[string length $index] == 0} {
	    set $varname [lindex $display_attrs $sel]
	} else {
	    set ${varname}($index) [lindex $display_attrs $sel]
	}
	if {[string length $varname2] > 0} {
	    global $varname2
	    if {[string length $index2] == 0} {
		set $varname2 [lindex [lindex $fx_display_info $sel] 0]
	    } else {
		set ${varname2}($index2) [lindex [lindex $fx_display_info $sel] 0]
	    }
	}
	if {[string length $func] > 0} {
	    $func $attr_types([lindex $display_attrs $sel])
	}
    }
    itcl::delete object fx_display_numeric_search_listbox
    destroy $f
    update idletasks
    if {[info exists grabStatus] && [string compare $grabStatus global] == 0} {
	catch "grab -global $oldGrab"
	catch "focus -force $oldGrab"
    } else {
	catch "grab $oldGrab"
	catch "focus $oldFocus"
    }
    update idletasks
    catch "unset fxPriv(popup_listbox)"
}

