# $RCSfile: fx_fonts.tcl,v $ $Revision: 2.4 $ $Date: 2002/12/27 21:48:11 $
#
# Copyright (C) 1995-2001 Herrin Software Development, Inc.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


# EXPORTED:
#    SelectFont [w [default font] [fixed default font]]

namespace eval ::qddb::fx::fonts {

    proc SelectFont {{w ".fx_selectfont"} {default {helvetica 12 bold}} {fixed_default {courier 12 bold}}} {
	variable frame_padx 3
	variable frame_pady 3

	toplevel $w
	set parent [winfo parent $w]
	wm title $w "Font Selection"
	wm transient $w $parent
	wm iconname $w SelectFont
	wm group $w $parent
	wm protocol $w WM_DELETE_WINDOW { }

	if {[catch {GetCurrentFont} myfont]} {
	    set ::qddb::fx::fonts::select_variable(family) [lindex $default 0]
	    set ::qddb::fx::fonts::select_variable(size) [lindex $default 1]
	    set ::qddb::fx::fonts::select_variable(style) [lindex $default 2]
	}
	if {[catch {GetCurrentFont fixed} myfixedfont]} {
	    set ::qddb::fx::fonts::select_fixed(family) [lindex $fixed_default 0]
	    set ::qddb::fx::fonts::select_fixed(size) [lindex $fixed_default 1]
	    set ::qddb::fx::fonts::select_fixed(style) [lindex $fixed_default 2]
	}
	frame $w.f_top
	set f $w.f_top
	grid $f	-padx $frame_padx -pady $frame_pady

	# Three internal frames for {families, style, size}
	CreateFamilyFrame $f 1 1 variable
	CreateStyleFrame $f  1 2 variable
	CreateSizeFrame $f   1 3 variable

	CreateFamilyFrame $f 1 4 fixed
	CreateStyleFrame $f  1 5 fixed
	CreateSizeFrame $f   1 6 fixed

	frame $w.f_middle -relief ridge -bd 1
	set f $w.f_middle
	entry $f.entry -width 40 -relief flat -justify c
	$f.entry insert 0 "This is the font you have selected!"
	grid $f.entry -sticky nsew
	set example_textbox $f.entry
	grid $f -row 2 -sticky nsew -padx [expr $frame_padx * 2] -pady [expr $frame_pady * 2]
	grid rowconfigure $f 2 -minsize 40 -pad 10

	frame $w.f_bottom -relief ridge -bd 1
	set f $w.f_bottom
	button $f.ok -text "Ok" -command "\
	    set [namespace current]::windows_reconfigured 1; \
	    [namespace current]::BeforeReconfigure; \
	    [namespace current]::SaveFontConfig; \
	    [namespace current]::ApplyConfiguredFont; \
	    destroy $w
	"
	button $f.apply -text "Apply" -command "\
		set [namespace current]::windows_reconfigured 1; \
		[namespace current]::BeforeReconfigure; \
		[namespace current]::ApplyCurrentFont; \
		eval {Fx:CatchRun {eval [Fx_Menubar::AfterReconfigure]} {fonts reconfigure}}
	"
	button $f.quit -text "Quit" -command "
	set [namespace current]::windows_reconfigured 0
	destroy $w
	"
	grid $f.ok -column 1 -row 1
	grid $f.apply -column 2 -row 1
	grid $f.quit -column 3 -row 1
	grid $f -row 3 -sticky nsew -padx [expr $frame_padx * 2] -pady [expr $frame_pady * 2]
	update idletasks; update
	trace variable ::qddb::fx::fonts::select_variable w \
		[list [namespace current]::RenderCurrentFont $example_textbox]
	tkwait window $w
	trace vdelete ::qddb::fx::fonts::select_variable w \
		[list [namespace current]::RenderCurrentFont $example_textbox]
	return [set [namespace current]::windows_reconfigured]
    }
    
    proc CreateFamilyFrame {w main_row main_col {type variable}} {
	upvar 1 frame_padx frame_padx
	upvar 1 frame_pady frame_pady

	frame $w.family_frame_$type -relief ridge -bd 1
	set f $w.family_frame_$type
	listbox $f.family -exportsel 0 -selectmode browse \
		-xscrollcommand [list $f.family_x set] \
		-yscrollcommand [list $f.family_y set]
	scrollbar $f.family_x -orient horizontal -command [list $f.family xview]
	scrollbar $f.family_y -command [list $f.family yview]
	set family_list [lsort -ascii [font families]]
	set current_index [lsearch -exact $family_list [set ::qddb::fx::fonts::select_[set type](family)]]
	foreach i $family_list {
	    $f.family insert end $i
	}
	$f.family selection set $current_index
	$f.family see $current_index
	grid $f -row $main_row -column $main_col -rowspan 1 -columnspan 1 -sticky nsew \
		-padx $frame_padx -pady $frame_pady
	grid rowconfigure $f 0 -weight 1
	grid columnconfigure $f 0 -weight 1
	grid $f.family $f.family_y -sticky nswe -in $f
	grid $f.family_x -sticky nswe -in $f
	bind $f.family <1> "set ::qddb::fx::fonts::select_[set type](family) \[%W get \[%W nearest %y\]\]"
	bind $f.family <Key> "set ::qddb::fx::fonts::select_[set type](family) \[%W get \[%W nearest %y\]\]"
    }

    proc CreateStyleFrame {w main_row main_col {type variable}} {
	upvar 1 frame_padx frame_padx
	upvar 1 frame_pady frame_pady

	set f $w.style_frame_$type
	frame $f -relief ridge -bd 1
	foreach {style row} {
	    bold      0
	    italic    1
	    underline 2
	    overstrike 3
	} {
	    checkbutton $f.style_$style -text $style \
		    -variable ::qddb::fx::fonts::select_[set type](style,$style)
	    grid $f.style_$style -row $row -sticky w -in $f
	}
	grid $f -row $main_row -column $main_col -rowspan 1 -columnspan 1 -sticky nsew \
		-padx $frame_padx -pady $frame_pady
    }

    proc CreateSizeFrame {w main_row main_col {type variable}} {
	upvar 1 frame_padx frame_padx
	upvar 1 frame_pady frame_pady

	set f $w.size_frame_$type
	frame $f -relief ridge -bd 1
	foreach {size row col} {
	    8 0 0
	    10 1 0
	    12 2 0
	    14 3 0
	    18 0 1
	    20 1 1
	    24 2 1
	    28 3 1
	} {
	    radiobutton $f.size_$size -text $size -variable ::qddb::fx::fonts::select_[set type](size) \
		    -value $size
	    grid $f.size_$size -row $row -column $col -sticky w -in $f
	}
	grid $f -row $main_row -column $main_col -rowspan 1 -columnspan 1 -sticky nsew \
		-padx $frame_padx -pady $frame_pady
    }

    proc GetCurrentFont {{type variable}} {
	global fx_font

	if {[info exists ::qddb::fx::fonts::select_[set type](family)]} {
	    set family [set ::qddb::fx::fonts::select_[set type](family)]
	} else {
	    set ::qddb::fx::fonts::select_[set type](family) [lindex $fx_font 0]
	    set family [set ::qddb::fx::fonts::select_[set type](family)]
	}
	if {[info exists ::qddb::fx::fonts::select_[set type](size)]} {
	    set size [set ::qddb::fx::fonts::select_[set type](size)]
	} else {
	    set ::qddb::fx::fonts::select_[set type](size) [lindex $fx_font 1]
	    set size [set ::qddb::fx::fonts::select_[set type](size)]
	}
	set style {}
	foreach i {bold italic underline overstrike} {
	    if {[info exists ::qddb::fx::fonts::select_[set type](style,$i)]} {
		if {[set ::qddb::fx::fonts::select_[set type](style,$i)] != 0} {
		    lappend style $i
		}
	    } else {
		if {[lsearch [lindex $fx_font 2] $i] != -1} {
		    set ::qddb::fx::fonts::select_[set type](style,$i) 1
		    lappend style $i
		} else {
		    set ::qddb::fx::fonts::select_[set type](style,$i) 0
		}
	    }
	}
	set retval [list $family $size $style]
    }

    proc RenderCurrentFont {w name1 name2 op} {
	$w configure -font [GetCurrentFont]
    }

    proc BeforeReconfigure {} {
	update idletasks
	set reconf [Fx_Menubar::BeforeReconfigure]
	if {[llength $reconf] > 0} {
	    eval $reconf
	}
    }

    proc SaveFontConfig {} {
	global ::fx_classes ::fx_font

	set ::fx_font [GetCurrentFont]
	set ::fx_fixed_font [GetCurrentFont fixed]
	foreach i $::fx_classes {
	    if {$i == "Listbox"} {
		option add "*$i.font" $::fx_fixed_font
	    } else {
		option add "*$i.font" $::fx_font
	    }
	}
    }
    
    proc ResetAllFontsToDefault {{w .} font} {
	set class [winfo class $w]
	set font [option get $w font $class]
	if {[string compare $class Menu] == 0} {
	    catch [list $w configure -font $font]
	    set last [$w index last]
	    if {[string compare $last none] == 0} {
		return
	    }
	    for {set i 0} {$i <= $last} {incr i} {
		catch [list $w entryconfigure $i -font $font]
		if {[catch "$w entrycget $i -menu" menu] == 0} {
		    ChangeAllFonts $menu
		}
	    }
	    return
	}
	foreach i [winfo children $w] {
	    set font [option get $i font $class]
	    catch [list $i configure -font $font]
	    ChangeAllFonts $i
	}
    }


    proc ApplyCurrentFont {{w .}} {
	set class [winfo class $w]
	if {$class == "Listbox"} {
	    set font [GetCurrentFont fixed]
	} else {
	    set font [GetCurrentFont]
	}
	if {[string compare $class Menu] == 0} {
	    catch [list $w configure -font $font]
	    set last [$w index last]
	    if {[string compare $last none] == 0} {
		return
	    }
	    for {set i 0} {$i <= $last} {incr i} {
		catch [list $w entryconfigure $i -font $font]
		if {[catch "$w entrycget $i -menu" menu] == 0} {
		    ApplyCurrentFont $menu
		}
	    }
	    return
	}
	foreach i [winfo children $w] {
	    catch [list $i configure -font $font]
	    ApplyCurrentFont $i
	}
    }

    proc ApplyConfiguredFont {{w .}} {
	set class [winfo class $w]
	set font [option get $w font $class]
	if {[string compare $class Menu] == 0} {
	    catch [list $w configure -font $font]
	    set last [$w index last]
	    if {[string compare $last none] == 0} {
		return
	    }
	    for {set i 0} {$i <= $last} {incr i} {
		catch [list $w entryconfigure $i -font $font]
		if {[catch "$w entrycget $i -menu" menu] == 0} {
		    ApplyCurrentFont $menu
		}
	    }
	    return
	}
	foreach i [winfo children $w] {
	    set font [option get $i font $class]
	    catch [list $i configure -font $font]
	    ApplyCurrentFont $i
	}
    }
}

if {[info exists fx_debug] && $fx_debug == 1} {
    puts "auto-loaded fx_fonts.tcl"
}
