# $RCSfile: fx_print.tcl,v $ $Revision: 2.0.0.1 $ $Date: 2000/11/09 15:06:01 $
#
# Copyright (C) 1995-2000 Herrin Software Development, Inc.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.



if {![info exists fx_config(printer)]} {
    set fx_config(header) {}
    set fx_config(footer) {}
    set fx_config(print_heading) {}
    set fx_config(printer_name) {lp}
    set fx_config(dosfile_name) {report.txt}
    set fx_config(unixfile_name) {report.txt}
    # 2 == unix file, 1 == printer, 0 == dosfloppy, 3 == screen, 4 == e-mail
    set fx_config(printer) 1
    set fx_config(printer_format) ascii
}
set fx_config(print_cmd_default) "| lpr -h"
set fx_config(print_cmd) "| lpr -h -P"

if {![info exists fx_config(printer_paginate)]} {
    set fx_config(printer_paginate) 1
}
if {![info exists fx_config(printer_nopostscript)]} {
    set fx_config(printer_nopostscript) 0
}
if {![info exists fx_config(postscript,orient)]} {
    set fx_config(postscript,orient) portrait
}
if {![info exists fx_config(postscript,charsperinch)]} {
    set fx_config(postscript,charsperinch) 12
}
if {![info exists fx_config(postscript,linesperpage)]} {
    set fx_config(postscript,linesperpage) 66
}
if {![info exists fx_config(postscript,vs)]} {
    set fx_config(postscript,vs) 11.4
}
if {![info exists fx_config(postscript,vsunits)]} {
    set fx_config(postscript,vsunits) p
}
if {![info exists fx_config(postscript,leftmargin)]} {
    set fx_config(postscript,leftmargin) 36
}
if {![info exists fx_config(postscript,leftmarginunits)]} {
    set fx_config(postscript,leftmarginunits) p
}
if {![info exists fx_config(postscript,topmargin)]} {
    set fx_config(postscript,topmargin) 36
}
if {![info exists fx_config(postscript,topmarginunits)]} {
    set fx_config(postscript,topmarginunits) p
}
if {![info exists fx_config(postscript,botmargin)]} {
    set fx_config(postscript,botmargin) 36
}
if {![info exists fx_config(postscript,botmarginunits)]} {
    set fx_config(postscript,botmarginunits) p
}

itcl::class Fx_PrintDialog {
    constructor {args} {
	global ::fx_config ::fx_tmpconfig ::fx_perform_global_grabs ::fx_monochrome

	eval configure $args
	set p $toplevel.print_dialog
	if {[winfo exists $p]} {
	    focus $p
	}
	if {[string compare $toplevel ""] == 0} {
	    toplevel $p
	    wm title $p "Printer Selection"
	    wm withdraw $p
	    wm transient $p .
	    wm protocol $p WM_DELETE_WINDOW [list set fx_config(cancel_print) 1]
	} else {
	    frame $p -relief groove -bd 4
	    set parent [winfo toplevel [winfo parent $p]]
	    place $p -in $parent -relx 0.5 -rely 0.5 -anchor c
	}	
	set fx_config(print_title) $title
	entry $p.title -relief flat -justify center
	$p.title insert 0 $title
	$p.title configure -state disabled
	pack $p.title -side top -expand on -fill x
	label $p.m -text "Select a printer" -justify center
	if {!$fx_monochrome} {
	    $p.m configure -fg red
	}
	pack $p.m -side top -expand on -fill x
	frame $p.f0
	pack $p.f0 -side top -expand on -fill x -pady 2m
	label $p.f0.l -width 10 -text Heading: -anchor e
	pack $p.f0.l -side left
	entry $p.f0.e -textvariable fx_config(print_heading)
	pack $p.f0.e -side left -expand on -fill x
	frame $p.f10 -relief groove -bd 2
	pack $p.f10 -side top -expand on -fill x -ipadx 20m
	checkbutton $p.f10.paginate \
		-text "Paginate" -anchor w -variable fx_tmpconfig(printer_paginate)
	radiobutton $p.f10.text \
		-text "ASCII" -anchor w -variable fx_tmpconfig(printer_format) -value ascii \
		-command "
                 set fx_tmpconfig(printer_format) ascii
	         $p.f10.b configure -state disabled
	"
	radiobutton $p.f10.ps \
		-text "Postscript" -anchor w -variable fx_tmpconfig(printer_format) -value postscript \
		-command "
                 set fx_tmpconfig(printer_format) postscript
	         $p.f10.b configure -state normal
	"
	button $p.f10.b -text "Settings..." -padx 2m \
		-command "Fx:BusyExec $p \"Fx:PostscriptSettings $p\""
	if {$fx_config(printer_nopostscript) == 0} {
	    pack $p.f10.b -side right -expand off -fill none
	    pack $p.f10.ps -side right
	    pack $p.f10.text -side right
	} else {
	    set lastformat $fx_tmpconfig(printer_format)
	    set fx_tmpconfig(printer_format) ascii
	}
        pack $p.f10.paginate -side left -expand off -fill none
	frame $p.f
	pack $p.f -side top -expand on -fill x
	if {[info exists fx_config(printer)]} {
	    set fx_tmpconfig(printer) $fx_config(printer)
	}
	if {[info exists fx_config(printer_name)]} {
	    set fx_tmpconfig(printer_name) $fx_config(printer_name)
	}
	if {[info exists fx_config(dosfile_name)]} {
	    set fx_tmpconfig(dosfile_name) $fx_config(dosfile_name)
	}
	if {[info exists fx_config(unixfile_name)]} {
	    set fx_tmpconfig(unixfile_name) $fx_config(unixfile_name)
	}
	if {[info exists fx_config(printer_format)]} {
	    set fx_tmpconfig(printer_format) $fx_config(printer_format)
	} else {
	    set fx_tmpconfig(printer_format) ascii
	}
	if {[info exists fx_config(printer_paginate)]} {
	    set fx_tmpconfig(printer_paginate) $fx_config(printer_paginate)
	} else {
	    set fx_tmpconfig(printer_paginate) 1
	}
	if {$fx_tmpconfig(printer_format) == "ascii"} {
	    $p.f10.text invoke
	}
	set fx_tmpconfig(tmp) {}
	radiobutton $p.f.l -text "Printer:" -anchor w -width 10 -variable fx_tmpconfig(printer) -value 1 \
	    -command "
		 $p.f.e configure -state normal
                 if {![info exists fx_tmpconfig(printer_format)]} \"$p.f10.text invoke\"
                 if {[string compare $fx_tmpconfig(printer_format) postscript] == 0} \"$p.f10.ps invoke\"
                 $p.f1.e configure -state disabled
                 $p.f3.e configure -state disabled
                 $p.f3.b configure -state disabled
                 focus $p.f.e
	"
	pack $p.f.l -side left
	entry $p.f.e -textvariable fx_tmpconfig(printer_name)
	pack $p.f.e -side left -expand on -fill x

	frame $p.f1
	pack $p.f1 -side top -expand on -fill x
	radiobutton $p.f1.l -text "DOS (A:):" -anchor w -width 10 -variable fx_tmpconfig(printer) -value 0 \
	    -command "
                 $p.f.e configure -state disabled
		 $p.f1.e configure -state normal
                 $p.f3.e configure -state disabled
                 $p.f3.b configure -state disabled
                 focus $p.f1.e
            "
	pack $p.f1.l -side left
	entry $p.f1.e -textvariable fx_tmpconfig(dosfile_name)
	pack $p.f1.e -side left -expand on -fill x

	frame $p.f3
	pack $p.f3 -side top -expand on -fill x
	radiobutton $p.f3.l -text "File:" -anchor w -width 10 -variable fx_tmpconfig(printer) -value 2 \
	    -command "
                 $p.f.e configure -state disabled
		 $p.f1.e configure -state disabled
                 $p.f3.e configure -state normal
                 $p.f3.b configure -state normal
                 focus $p.f3.e
            "
	pack $p.f3.l -side left
	entry $p.f3.e -textvariable fx_tmpconfig(unixfile_name)
	pack $p.f3.e -side left -expand on -fill x
	button $p.f3.b -text "Browse..." -padx 2m -command "
	    set fx_tmpconfig(unixfile_name) \[Fx:BusyExec $p \"tk_getSaveFile -parent . -title Browse\"\]
	"
	pack $p.f3.b -side left -expand off -fill none
	frame $p.f4
	pack $p.f4 -side top -expand on -fill x
	radiobutton $p.f4.l -text "Preview on screen" -anchor w \
		-width 10 -variable fx_tmpconfig(printer) -value 3 \
	    -command "
                 $p.f.e configure -state disabled
		 $p.f1.e configure -state disabled
                 $p.f3.e configure -state disabled
                 $p.f3.b configure -state disabled
                 focus $p.f4.l
            "
	pack $p.f4.l -side left -expand on -fill x
	radiobutton $p.f4.l2 -text "E-mail" -anchor w \
		-variable fx_tmpconfig(printer) -value 4 \
		-command "
                 $p.f.e configure -state disabled
		 $p.f1.e configure -state disabled
                 $p.f3.e configure -state disabled
                 $p.f3.b configure -state disabled
                 focus $p.f4.l2
	"
	pack $p.f4.l2 -side left -expand on -fill x
	if {$fx_tmpconfig(printer) == 1} {
	    if {$fx_tmpconfig(printer_format) == "postscript"} {
		$p.f10.ps invoke
	    } else {
		$p.f10.text invoke
	    }
	    $p.f.l invoke
	} elseif {$fx_tmpconfig(printer) == 2} {
	    $p.f3.l invoke
	} elseif {$fx_tmpconfig(printer) == 4} {
	    $p.f4.l2 invoke
	} elseif {$fx_tmpconfig(printer) == 0} {
	    $p.f1.l invoke
	} else {
	    $p.f4.l invoke
	}
	set ok_cmd "$this SetPrinter; set fx_config(cancel_print) 0; \
		$p.f2.f.ok flash; break"
	bind $p.f.e <Return> $ok_cmd
	bind $p.f1.e <Return> $ok_cmd
	bind $p.f0.e <Return> $ok_cmd
	bind $p.f3.e <Return> $ok_cmd

	frame $p.f2
	pack $p.f2 -side top -padx 5m -pady 2m -expand on -fill x

	frame $p.f2.f -relief sunken -bd 1
	pack $p.f2.f -side left -padx 3m -pady 2m -in $p.f2

	set ok_cmd "$this SetPrinter; set fx_config(cancel_print) 0; $p.f2.f.ok flash"
	button $p.f2.f.ok -text Ok -command $ok_cmd
	pack $p.f2.f.ok -padx 2m -pady 2m -in $p.f2.f -ipadx 2m -ipady 1m

	button $p.f2.cancel -text Cancel -command "set fx_config(cancel_print) 1"
	pack $p.f2.cancel -side right -ipadx 3m -ipady 1m

	if {[string compare $toplevel ""] == 0} {
	    update; update idletasks
	    set ht [winfo reqheight $p]
	    set wid [winfo reqwidth $p]
	    wm geometry $p \
		    "+[expr [winfo screenwidth .]/2 - $wid/2]+[expr [winfo screenheight .]/2 - $ht/2]"
	    wm minsize $p $wid $ht
	    wm maxsize $p $wid $ht
	    wm deiconify $p
	} else {
	    fx_containfocus [list $p.f0.e $p.f2.cancel]
	}
	set oldgrab [grab current $p]
	set oldfocus [focus]
	switch $fx_tmpconfig(printer) {
	    0 {
		catch "focus $p.f1.e"
	    }
	    1 {
		catch "focus $p.f.e"
	    }
	    2 {
		catch "focus $p.f3.e"
	    }
	    3 {
		catch "focus $p.f4.l"
	    }
	    4 {
		catch "focus $p.f4.l2"
	    }
	    default {
		catch "focus $p.f.e"
	    }
	}
	catch "grab $p"
	tkwait variable fx_config(cancel_print)
	catch "grab release $p"
	if {[info exists lastformat]} {
	    set fx_config(printer_format) $lastformat
	}
	catch "itcl::delete object $this"
    }
    destructor {
	catch "grab release $p"
	if {[string compare $oldgrab ""] != 0} {
	    catch "grab $oldgrab"
	}
	if {[string compare $oldfocus ""] != 0} {
	    catch "focus $oldfocus"
	}
	catch "destroy $p"
    }
    method SetPrinter {} {
	global ::fx_config ::fx_tmpconfig

	set fx_config(printer) $fx_tmpconfig(printer)
	set fx_config(printer_name) $fx_tmpconfig(printer_name)
	set fx_config(unixfile_name) $fx_tmpconfig(unixfile_name)
	set fx_config(dosfile_name) $fx_tmpconfig(dosfile_name)
	set fx_config(printer_format) $fx_tmpconfig(printer_format)
	set fx_config(printer_paginate) $fx_tmpconfig(printer_paginate)
    }
    public variable title {Printer Selection}
    public variable toplevel {}
    protected variable p
    protected variable oldgrab {}
    protected variable oldfocus {}
}

proc Fx:PrinterError {{type printer}} {
    global fx_config

    switch -exact $type {
	printer {
	    fx_dialog .dialog "Printer Error" \
		    "Printer Error: Please make sure that the requested printer '$fx_config(printer_name)' exists." \
		    error 0 OK
	}
	dosfile {
	    fx_dialog .dialog "File Write Error" \
		    "File Error: Cannot create file $fx_config(dosfile_name)!" \
		    error 0 OK
	}
	file {
	    fx_dialog .dialog "File Write Error" \
		    "File Error: Cannot create file $fx_config(unixfile_name)!" \
		    error 0 OK
	}
	email {
	    fx_dialog .dialog "E-mail Error" \
		    "E-mail Error: Cannot send e-mail!" \
		    error 0 OK
	}
    }
}

proc Fx:PrintListToString {list} {
    global fx_config

    if {$fx_config(printer_nopostscript)} {
	return
    }
    set outstring ""
    if {[info exists fx_config(print_heading)]} {
	set alen [string length [lindex $list 0]]
	set hlen [string length $fx_config(print_heading)]
	set alen [expr ($alen - $hlen)/2]
	set form [format "%%%ds" $alen]
	set form [format $form " "]
	append outstring $form
	append outstring $fx_config(print_heading)
	append outstring "$form\n"
    }
    append outstring "\n"
    foreach i $list {
	append outstring "$i\n"
    }    
    return $outstring
}

proc Fx:PrintPostscript {outstring} {
    global fx_config

    return [qddb_util postscript \
	    -orient $fx_config(postscript,orient) \
	    -cpi $fx_config(postscript,charsperinch) \
	    -lines $fx_config(postscript,linesperpage) \
	    -baseline $fx_config(postscript,topmargin)$fx_config(postscript,topmarginunits) \
	    -vs $fx_config(postscript,vs)$fx_config(postscript,vsunits) \
	    -leftmargin $fx_config(postscript,leftmargin)$fx_config(postscript,leftmarginunits) \
	    $outstring]
}

proc Fx:Print {{contents ""}} {
    global fx_config
    
    update idletasks
    set outstring [Fx:PrintListToString $contents]
    set tmpname [TempNam]
    if {[catch "open $tmpname w" fd] != 0} {
	Fx:Dialog "Print error" "$fd"
	return
    }
    if {$fx_config(printer_paginate)} {
	if {[string compare $fx_config(printer_format) "postscript"] == 0} {
	    set outstring [qddb_util pr -headers on -footers off -formfeed on \
		    -linesperpage $fx_config(postscript,linesperpage) $outstring]
	} else {
	    set outstring [qddb_util pr -headers on -footers on -formfeed off \
		    -linesperpage 66 $outstring]
	}
    }
    puts -nonewline $fd $outstring
    catch {close $fd}
    if {[catch [list open $tmpname r+] fd] != 0} {
	Fx:Dialog "Print error" "$fd"
	return
    }
    set outstring [read $fd]
    if {[string compare $fx_config(printer_format) "postscript"] == 0 && \
	    $fx_config(printer) != 3} {
	set outstring [Fx:PrintPostscript $outstring]
	seek $fd 0 start
	puts -nonewline $fd $outstring
    }
    catch {close $fd}
    if {$fx_config(printer) == 1} {
	if {[string length $fx_config(printer_name)] > 0} {
            if {[catch {open "$fx_config(print_cmd)$fx_config(printer_name)" w} fd] != 0} {
		Fx:PrinterError
		return
	    }
	} else {
	    if {[catch {open "$fx_config(print_cmd_default)" w} fd] != 0} {
		Fx:PrinterError
		return
	    }
	}
	puts -nonewline $fd $outstring
	if {[catch "close $fd"]} {
	    Fx:PrinterError
	    return
	}
    } elseif {$fx_config(printer) == 2} {
	if {[string length $fx_config(unixfile_name)] == 0} {
	    return
	}
	set s [split $fx_config(unixfile_name) /]
	if {[llength $s] > 1} {
	    set dir [join [lrange $s 0 [expr [llength $s] - 2]] /]
	    set file ${dir}/[lindex $s [expr [llength $s] - 1]]
	} else {
	    set file $fx_config(unixfile_name)
	}
	if {[catch {file copy -force $tmpname $file} msg] != 0} {
	    Fx:Dialog "Print error" "$msg"
	}
    } elseif {$fx_config(printer) == 0} {
	# check out the file name first
	if {[string length $fx_config(dosfile_name)] == 0} {
	    return
	}
	update idletasks
	while {[catch "exec mcopy -tn $tmpname a:$fx_config(dosfile_name)"] != 0} {
	    set ans [fx_dialog .dialog "Floppy Error" \
		    "A floppy error has occurred.   Would you like to try formatting the floppy?" \
		    error 0 Yes No]
	    if {$ans == 0} {
		update idletasks
		fx_message .msg "Formatting..."
		update idletasks
		if {[catch "exec mformat a:"] != 0} {
		    .msg.msg configure -text "Format failed!"
		    update idletasks
		    catch "exec sleep 2"
		    destroy .msg
		    update idletasks
		    set ans 1
		} else {
		    .msg.msg configure -text "Format succeeded!"
		    update idletasks
		    catch "exec sleep 2"
		    destroy .msg
		    update idletasks
		}
	    }
	    if {$ans == 1} {
		set ans [fx_dialog .dialog "Floppy Error" \
			"Please make sure that an error-free formatted floppy is in drive A:" \
			error 0 OK Cancel]
		if {$ans != 0} {
		    break
		}
	    }
	}
    } elseif {$fx_config(printer) == 4} {
	if {[info exists fx_config(print_title)]} {
	    set title $fx_config(print_title)
	} elseif {[info exists fx_config(print_heading)]} {
	    set title $fx_config(print_heading)
	} else {
	    set title ""
	}
	Fx:Email_SimpleSend $outstring $title
    } else {
	if {[info exists fx_config(print_title)]} {
	    set title $fx_config(print_title)
	} else {
	    set title $fx_config(print_heading)
	}
	set message ""
	foreach i $contents {
	    append message "$i\n"
	}
	set tmp2name [TempNam]
	set fd [open $tmp2name w]
	puts -nonewline $fd $message
	catch {close $fd}
	Fx:FileView $tmp2name $title
	if {[info exists fx_config(file_viewer)] && [winfo exists $fx_config(file_viewer)]} {
	    tkwait window $fx_config(file_viewer)
	}
	catch {file delete -force $tmp2name}
    }
    update idletasks
    catch {file delete -force $tmpname}
}

proc Fx:PrintFile {fn} {
    global fx_config
    
    update idletasks
    if {[catch [list open $fn r] fd] != 0} {
	Fx:PrinterError
	return
    }
    set outstring [read $fd]
    if {[info exists fx_config(print_heading)] && $fx_config(printer_nopostscript) == 0} {
	set alen 80
	set hlen [string length $fx_config(print_heading)]
	set alen [expr ($alen - $hlen)/2]
	set form [format "%%%ds" $alen]
	set form [format $form " "]
	set outheader $form
	append outheader $fx_config(print_heading)
	append outheader "$form\n\n"
    } else {
	set outheader ""
    }
    set outstring "$outheader$outstring"
    set tmpname [TempNam]
    if {[catch "open $tmpname w" fd] != 0} {
	Fx:Dialog "Print error" "$fd"
	return
    }
    if {$fx_config(printer_paginate)} {
	if {[string compare $fx_config(printer_format) "postscript"] == 0} {
	    set outstring [qddb_util pr -headers on -footers off -formfeed on \
		    -linesperpage $fx_config(postscript,linesperpage) $outstring]
	} else {
	    set outstring [qddb_util pr -headers on -footers on -formfeed off \
		    -linesperpage 66 $outstring]
	}
    }
    puts -nonewline $fd $outstring
    catch {close $fd}
    if {[catch [list open $tmpname r+] fd] != 0} {
	Fx:Dialog "Print error" "$fd"
	return
    }
    set outstring [read $fd]
    if {[string compare $fx_config(printer_format) "postscript"] == 0} {
	set outstring [Fx:PrintPostscript $outstring]
	seek $fd 0 start
	puts -nonewline $fd $outstring
    }
    catch {close $fd}
    if {$fx_config(printer) == 1} {
	if {[string length $fx_config(printer_name)] > 0} {
	    if {[catch {eval exec cat $tmpname $fx_config(print_cmd)$fx_config(printer_name)}] != 0} {
		Fx:Dialog "Print error" "$fd"
	    }
	} else {
	    if {[catch {eval exec cat $tmpname $fx_config(print_cmd_default)}] != 0} {
		Fx:Dialog "Print error" "$fd"
	    }
	}
    } elseif {$fx_config(printer) == 2} {
	if {[string length $fx_config(unixfile_name)] == 0} {
	    return
	}
	set s [split $fx_config(unixfile_name) /]
	if {[llength $s] > 1} {
	    set dir [join [lrange $s 0 [expr [llength $s] - 2]] /]
	    set file ${dir}/[lindex $s [expr [llength $s] - 1]]
	} else {
	    set file $fx_config(unixfile_name)
	}
	if {[catch {file copy -force $tmpname $file} msg] != 0} {
	    Fx:Dialog "Print error" "$msg"
	}
    } elseif {$fx_config(printer) == 0} {
	# check out the file name first
	if {[string length $fx_config(dosfile_name)] == 0} {
	    return
	}
	update idletasks
	while {[catch "exec mcopy -tn $tmpname a:$fx_config(dosfile_name)"] != 0} {
	    set ans [fx_dialog .dialog "Floppy Error" \
		    "A floppy error has occurred.   Would you like to try formatting the floppy?" \
		    error 0 Yes No]
	    if {$ans == 0} {
		update idletasks
		fx_message .msg "Formatting..."
		update idletasks
		if {[catch "exec mformat a:"] != 0} {
		    .msg.msg configure -text "Format failed!"
		    update idletasks
		    catch "exec sleep 2"
		    destroy .msg
		    update idletasks
		    set ans 1
		} else {
		    .msg.msg configure -text "Format succeeded!"
		    update idletasks
		    catch "exec sleep 2"
		    destroy .msg
		    update idletasks
		}
	    }
	    if {$ans == 1} {
		set ans [fx_dialog .dialog "Floppy Error" \
			"Please make sure that an error-free formatted floppy is in drive A:" \
			error 0 OK Cancel]
		if {$ans != 0} {
		    break
		}
	    }
	}
    } elseif {$fx_config(printer) == 4} {
	if {[info exists fx_config(print_title)]} {
	    set title $fx_config(print_title)
	} elseif {[info exists fx_config(print_heading)]} {
	    set title $fx_config(print_heading)
	} else {
	    set title ""
	}
	Fx:Email_SimpleSend $outstring $title
    } else {
	if {[info exists fx_config(print_title)]} {
	    set title $fx_config(print_title)
	} elseif {[info exists fx_config(print_heading)]} {
	    set title $fx_config(print_heading)
	} else {
	    set title ""
	}
	Fx:FileView $fn $title
	if {[info exists fx_config(file_viewer)] && [winfo exists $fx_config(file_viewer)]} {
	    tkwait window $fx_config(file_viewer)
	}
    }
    update idletasks
    catch {file delete -force $tmpname}
}

proc Fx:PostscriptRecal {{val ""}} {
    global fx_tmpconfig

    if {"$val" != "NOVS" && "$val" != "NOVS2" && "$val" != "" && $fx_tmpconfig(recal_lines)} {
	set fx_tmpconfig(postscript,vsunits) p
	set fx_tmpconfig(postscript,vs) [expr (1.0 / ($val / 72.0)) * 1.9]
    } elseif {"$val" == "NOVS"} {
	set val $fx_tmpconfig(postscript,charsperinch)
	switch -exact $fx_tmpconfig(postscript,vsunits) {
	    i {
		set fx_tmpconfig(postscript,vs) [expr (1.0 / $val) * 1.9]
	    }
	    c {
		set fx_tmpconfig(postscript,vs) [expr (1.0 / $val) * 1.9 * 0.3937008]
	    }
	    m {
		set fx_tmpconfig(postscript,vs) [expr (1.0 / $val) * 1.9 * 0.03937008]
	    }
	    p -
	    default {
		set fx_tmpconfig(postscript,vs) [expr (1.0 / ($val / 72.0)) * 1.9]
	    }
	}
    }
    if {[string length $val] == 0} {
	set val $fx_tmpconfig(postscript,charsperinch)
    }
    switch -exact $fx_tmpconfig(postscript,topmarginunits) {
	i {
	    set topmargin [expr $fx_tmpconfig(postscript,topmargin) * 72.0]
	}
	c {
	    set topmargin [expr $fx_tmpconfig(postscript,topmargin) * 72.0 * 0.3937008]
	}
	m {
	    set topmargin [expr $fx_tmpconfig(postscript,topmargin) * 72.0 * 0.03937008]
	}
	p -
	default {
	    set topmargin $fx_tmpconfig(postscript,topmargin)
	}
    }
    switch -exact $fx_tmpconfig(postscript,botmarginunits) {
	i {
	    set botmargin [expr $fx_tmpconfig(postscript,botmargin) * 72.0]
	}
	c {
	    set botmargin [expr $fx_tmpconfig(postscript,botmargin) * 72.0 * 0.3937008]
	}
	m {
	    set botmargin [expr $fx_tmpconfig(postscript,botmargin) * 72.0 * 0.03937008]
	}
	p -
	default {
	    set botmargin $fx_tmpconfig(postscript,botmargin)
	}
    }
    if {[string compare $fx_tmpconfig(postscript,orient) portrait] == 0} {
	switch -exact $fx_tmpconfig(postscript,papertype) {
	    a4 {
		set points [expr 842.0 - $topmargin - $botmargin]
	    }
	    legal {
		set points [expr 1008.0 - $topmargin - $botmargin]
	    }
	    letter -
	    default {
		set points [expr 792.0 - $topmargin - $botmargin]
	    }
	}
    } else {
	switch -exact $fx_tmpconfig(postscript,papertype) {
	    a4 {
		set points [expr 596.0 - $topmargin - $botmargin]
	    }
	    legal -
	    letter -
	    default {
		set points [expr 612.0 - $topmargin - $botmargin]
	    }
	}
    }
    if {$fx_tmpconfig(recal_lines)} {
	set fx_tmpconfig(postscript,linesperpage) [expr int($points / $fx_tmpconfig(postscript,vs))]
    } else {
	after idle {set fx_tmpconfig(recal_lines) 1}
    }
}

proc Fx:PostscriptSettings {toplevel} {
    global fx_tmpconfig fx_config

    set fx_tmpconfig(recal_lines) 0
    toplevel .fx_ps
    wm withdraw .fx_ps
    wm title .fx_ps "Postscript settings"
    wm iconname .fx_ps "Postscript settings"
    wm transient .fx_ps $toplevel
    set p .fx_ps
    set oldgrab [grab current $p]
    catch {grab $p}
    foreach i [array names fx_config postscript,*] {
	set fx_tmpconfig($i) $fx_config($i)
    }
    frame $p.top
    pack $p.top -side top -expand on -fill both -ipadx 10m -ipady 5m
    set f $p.top

    set row 0
    label $f.l0 -text "Paper type:"
    if {![info exists fx_tmpconfig(postscript,papertype)]} {
	set fx_tmpconfig(postscript,papertype) letter
    }
    radiobutton $f.r0 -text A4 -value "a4" -variable fx_tmpconfig(postscript,papertype) \
	    -command {Fx:PostscriptRecal}
    radiobutton $f.r1 -text Letter -value "letter" -variable fx_tmpconfig(postscript,papertype) \
	    -command {Fx:PostscriptRecal}
    radiobutton $f.r2 -text Legal -value "legal" -variable fx_tmpconfig(postscript,papertype) \
	    -command {Fx:PostscriptRecal}
    grid $f.l0 -column 0 -row $row -sticky e
    grid $f.r0 -column 1 -row $row -sticky e
    grid $f.r1 -column 2 -columnspan 2 -row $row -sticky n
    grid $f.r2 -column 4 -row $row -sticky w

    incr row
    label $f.l00 -text "Orientation:"
    if {![info exists fx_tmpconfig(postscript,orient)]} {
	set fx_tmpconfig(postscript,orient) portrait
    }
    radiobutton $f.r00 -text Portrait -value "portrait" -variable fx_tmpconfig(postscript,orient) \
	    -command {Fx:PostscriptRecal}
    radiobutton $f.r01 -text Landscape -value "landscape" -variable fx_tmpconfig(postscript,orient) \
	    -command {Fx:PostscriptRecal}
    grid $f.l00 -column 0 -row $row -sticky e
    grid $f.r00 -column 1 -columnspan 2 -row $row -sticky e
    grid $f.r01 -column 3 -columnspan 2 -row $row -sticky e

    incr row
    label $f.l1 -text "Chars/inch:" -anchor center
    if {![info exists fx_tmpconfig(postscript,charsperinch)]} {
	set fx_tmpconfig(postscript,charsperinch) 12
    }
    entry $f.e1 -textvariable fx_tmpconfig(postscript,charsperinch) -width 4 -state disabled
    fx_bindint $f.e1
    scale $f.s1 -length 2i -showvalue off -sliderrelief raised -from 4 -to 30 \
	    -orient horiz -variable fx_tmpconfig(postscript,charsperinch)
    grid $f.l1 -column 0 -row $row -sticky e
    grid $f.e1 -column 1 -row $row -sticky w
    grid $f.s1 -row $row -column 2 -columnspan 2
    $f.s1 set $fx_tmpconfig(postscript,charsperinch)
    update; update idletasks
    $f.s1 configure -command {Fx:PostscriptRecal}
    label $f.l2 -text "Tab spacing:"
    if {![info exists fx_tmpconfig(postscript,tabspacing)]} {
	set fx_tmpconfig(postscript,tabspacing) 8
    }
    entry $f.e2 -textvariable fx_tmpconfig(postscript,tabspacing) -width 4
    fx_bindint $f.e2
    incr row
    label $f.l20 -text "Lines/page:"
    if {![info exists fx_tmpconfig(postscript,linesperpage)]} {
	set fx_tmpconfig(postscript,linesperpage) 66
    }
    entry $f.e20 -textvariable fx_tmpconfig(postscript,linesperpage) -width 4 -state disabled
    fx_bindint $f.e20
    grid $f.l2 -column 0 -row $row -sticky e
    grid $f.e2 -column 1 -row $row -sticky w
    grid $f.l20 -column 2 -columnspan 2 -row $row -sticky e
    grid $f.e20 -column 4 -row $row -sticky w

    incr row
    label $f.l3 -text "Top margin:"
    if {![info exists fx_tmpconfig(postscript,topmargin)]} {
	set fx_tmpconfig(postscript,topmargin) 36
    }
    entry $f.e3 -textvariable fx_tmpconfig(postscript,topmargin) -width 4
    fx_bindreal $f.e3
    bind $f.e3 <FocusOut> {
	if {[string length [%W get]] != 0} {
	    after idle "catch Fx:PostscriptRecal"
	} else {
	    set fx_tmpconfig(postscript,topmargin) 0.5
	    set fx_tmpconfig(postscript,topmarginunits) i
	}
    }
    if {![info exists fx_tmpconfig(postscript,topmarginunits)]} {
	set fx_tmpconfig(postscript,topmarginunits) p
    }
    radiobutton $f.r30 -text pt -variable fx_tmpconfig(postscript,topmarginunits) -value p \
	    -command {Fx:PostscriptRecal}
    radiobutton $f.r31 -text mm -variable fx_tmpconfig(postscript,topmarginunits) -value m \
	    -command {Fx:PostscriptRecal}
    radiobutton $f.r32 -text cm -variable fx_tmpconfig(postscript,topmarginunits) -value c \
	    -command {Fx:PostscriptRecal}
    radiobutton $f.r33 -text in -variable fx_tmpconfig(postscript,topmarginunits) -value i \
	    -command {Fx:PostscriptRecal}
    grid $f.l3 -column 0 -row $row -sticky e
    grid $f.e3 -column 1 -row $row -sticky w
    grid $f.r30 -column 2 -row $row -sticky w
    grid $f.r31 -column 3 -row $row -sticky w
    grid $f.r32 -column 4 -row $row -sticky w
    grid $f.r33 -column 5 -row $row -sticky w

    incr row
    label $f.l03 -text "Bottom margin:"
    if {![info exists fx_tmpconfig(postscript,botmargin)]} {
	set fx_tmpconfig(postscript,botmargin) 36
    }
    entry $f.e03 -textvariable fx_tmpconfig(postscript,botmargin) -width 4
    fx_bindreal $f.e03
    bind $f.e03 <FocusOut> {
	if {[string length [%W get]] != 0} {
	    after idle "catch Fx:PostscriptRecal"
	} else {
	    set fx_tmpconfig(postscript,botmargin) 0.5
	    set fx_tmpconfig(postscript,botmarginunits) i
	}
    }
    if {![info exists fx_tmpconfig(postscript,botmarginunits)]} {
	set fx_tmpconfig(postscript,botmarginunits) p
    }
    radiobutton $f.r030 -text pt -variable fx_tmpconfig(postscript,botmarginunits) -value p \
	    -command {Fx:PostscriptRecal}
    radiobutton $f.r031 -text mm -variable fx_tmpconfig(postscript,botmarginunits) -value m \
	    -command {Fx:PostscriptRecal}
    radiobutton $f.r032 -text cm -variable fx_tmpconfig(postscript,botmarginunits) -value c \
	    -command {Fx:PostscriptRecal}
    radiobutton $f.r033 -text in -variable fx_tmpconfig(postscript,botmarginunits) -value i \
	    -command {Fx:PostscriptRecal}
    grid $f.l03 -column 0 -row $row -sticky e
    grid $f.e03 -column 1 -row $row -sticky w
    grid $f.r030 -column 2 -row $row -sticky w
    grid $f.r031 -column 3 -row $row -sticky w
    grid $f.r032 -column 4 -row $row -sticky w
    grid $f.r033 -column 5 -row $row -sticky w

    incr row
    label $f.l4 -text "Left margin:"
    if {![info exists fx_tmpconfig(postscript,leftmargin)]} {
	set fx_tmpconfig(postscript,leftmargin) 36
    }
    entry $f.e4 -textvariable fx_tmpconfig(postscript,leftmargin) -width 4
    fx_bindreal $f.e4
    bind $f.e4 <FocusOut> {
	if {[string length [%W get]] != 0} {
	    after idle "catch Fx:PostscriptRecal"
	} else {
	    catch Fx:PostscriptRecal
	}
    }
    if {![info exists fx_tmpconfig(postscript,leftmarginunits)]} {
	set fx_tmpconfig(postscript,leftmarginunits) p
    }
    radiobutton $f.r40 -text pt -variable fx_tmpconfig(postscript,leftmarginunits) -value p \
	    -command {Fx:PostscriptRecal}
    radiobutton $f.r41 -text mm -variable fx_tmpconfig(postscript,leftmarginunits) -value m \
	    -command {Fx:PostscriptRecal}
    radiobutton $f.r42 -text cm -variable fx_tmpconfig(postscript,leftmarginunits) -value c \
	    -command {Fx:PostscriptRecal}
    radiobutton $f.r43 -text in -variable fx_tmpconfig(postscript,leftmarginunits) -value i \
	    -command {Fx:PostscriptRecal}
    grid $f.l4 -column 0 -row $row -sticky e
    grid $f.e4 -column 1 -row $row -sticky w
    grid $f.r40 -column 2 -row $row -sticky w
    grid $f.r41 -column 3 -row $row -sticky w
    grid $f.r42 -column 4 -row $row -sticky w
    grid $f.r43 -column 5 -row $row -sticky w

    incr row
    label $f.l5 -text "Vertical space:"
    if {![info exists fx_tmpconfig(postscript,vs)]} {
	Fx:PostscriptRecal $fx_tmpconfig(postscript,charsperinch)
    }
    entry $f.e5 -textvariable fx_tmpconfig(postscript,vs) -width 4
    fx_bindreal $f.e5
    bind $f.e5 <FocusOut> {
	if {[string length [%W get]] != 0} {
	    after idle [list catch "Fx:PostscriptRecal NOVS2"]
	} else {
	    catch {Fx:PostscriptRecal NOVS2}
	}
    }
    if {![info exists fx_tmpconfig(postscript,vsunits)]} {
	set fx_tmpconfig(postscript,vsunits) p
    }
    radiobutton $f.r50 -text pt -variable fx_tmpconfig(postscript,vsunits) -value p \
	    -command {Fx:PostscriptRecal NOVS}
    radiobutton $f.r51 -text mm -variable fx_tmpconfig(postscript,vsunits) -value m \
	    -command {Fx:PostscriptRecal NOVS}
    radiobutton $f.r52 -text cm -variable fx_tmpconfig(postscript,vsunits) -value c \
	    -command {Fx:PostscriptRecal NOVS}
    radiobutton $f.r53 -text in -variable fx_tmpconfig(postscript,vsunits) -value i \
	    -command {Fx:PostscriptRecal NOVS}
    grid $f.l5 -column 0 -row $row -sticky e
    grid $f.e5 -column 1 -row $row -sticky w
    grid $f.r50 -column 2 -row $row -sticky w
    grid $f.r51 -column 3 -row $row -sticky w
    grid $f.r52 -column 4 -row $row -sticky w
    grid $f.r53 -column 5 -row $row -sticky w

    frame $p.bot
    pack $p.bot -side top -fill x
    button $p.bot.ok -text Ok -command {
	global fx_tmpconfig fx_config
	foreach i [array names fx_tmpconfig postscript,*] {
	    set fx_config($i) $fx_tmpconfig($i)
	}
	catch {destroy .fx_ps}
    }
    pack $p.bot.ok -side left -padx 10m -pady 5m
    button $p.bot.cancel -text Cancel -command {catch "destroy .fx_ps"}
    pack $p.bot.cancel -side right -padx 10m -pady 5m
    update; update idletasks
    set ht [winfo reqheight $p]
    set wid [winfo reqwidth $p]
    wm geometry $p \
	    "+[expr [winfo screenwidth .]/2 - $wid/2]+[expr [winfo screenheight .]/2 - $ht/2]"
    wm minsize $p $wid $ht
    wm maxsize $p $wid $ht    
    wm deiconify $p
    focus $p
    tkwait window $p
    catch {grab release $p}
    catch {grab $oldgrab}
}

if {[info exists fx_debug] && $fx_debug == 1} {
    puts "auto-loaded fx_print.tcl"
}
