# $RCSfile: fx_script.tcl,v $ $Revision: 2.4 $ $Date: 2001/08/09 02:06:07 $
#
# Copyright (C) 1995-2000 Herrin Software Development, Inc.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package require BWidget

proc Fx:Prologue {config {s ""} {bannergif ""} {bannerxbm ""} {bannertitle ""}} {
    global auto_path blt_library fx_config fx_config_dir fx_blt \
	    fx_monochrome argv qddbscript qddb_library argv argv0

    if {[info exists blt_library]} {
	lappend auto_path $blt_library
    }
    set fx_config_dir .${config}_config
    if {"$s" == ""} {
	if {[catch [list qddb_schema open "[lindex $argv 0]"] s] != 0} {
	    puts stderr "Cannot open Schema for relation: \"[lindex $argv 0]\"\n($s)\n"
	    puts stderr "Usage: $argv0 RelationName"
	    exit 1
	}
    }
    set qddbscript(schema) $s
    Fx:Init $s
    set scrwid [winfo screenwidth .]
    set scrht [winfo screenheight .]
    
    toplevel .banner -relief raised -bd 2
    wm overrideredirect .banner 1
    frame .banner.f -bg black
    pack .banner.f -expand on -fill both
    if {[winfo depth .] < 8} {
	if {[string length $bannerxbm] == 0} {
	    label .banner.f.l -bitmap @$qddb_library/fx/pixmaps/qddb.xbm
	} else {
	    label .banner.f.l -bitmap @$bannerxbm
	}
    } else {
	if {[string length $bannergif] == 0} {
	    image create photo qddb_color -file $qddb_library/fx/pixmaps/qddb_color.gif
	    label .banner.f.l -image qddb_color -justify c -bd 0
	} else {
	    image create photo qddb_color -file $bannergif
	    label .banner.f.l -image qddb_color -justify c -bd 0
	}
    }
    pack .banner.f.l -side top -expand on -fill x -padx 0m -pady 0m
    
    set msgcolor black
    if {!$fx_monochrome} {
	.banner.f.l configure -bg $msgcolor
	.banner.f configure -bg $msgcolor
    }
    set path [qddb_schema path $s]
    set tail [file tail $path]
    if {[string compare $tail ""] == 0} {
	set tail [file tail [file dirname $path]]
    }
    if {[string length $bannertitle] == 0} {
	wm title .banner "$config ($tail)"
    } else {
	wm title .banner $bannertitle
    }
    wm iconname . "$tail"
    wm withdraw .
    if {$fx_blt} {
	::blt::busy hold .banner.f.l
    }
    wm withdraw .banner
    update idletasks
    update
    wm geometry .banner \
	    +[expr $scrwid/2 - [winfo reqwidth .banner]/2]+[expr $scrht/2 - [winfo reqheight .banner]/2]
    wm deiconify .banner
    update idletasks
    update
    wm title . "Qddb ($tail)"
    set qddbscript(trans_name) [split [qddb_schema path $s] /]
    set qddbscript(trans_name) [glob ~]/.${config}_restrict/[join $qddbscript(trans_name) .]
    return $s
}

proc Fx:Epilogue {} {
    menubar configure -instances [Fx_Entry::GetInstances] -frames [Fx_Frame::GetInstances]
    menubar SearchModeProc
    Fx:InitConfigure
    destroy .banner
}

proc Fx:InitConfigure {} {
    global fx_config fx_rootx fx_rooty

    Fx_Entry::BeforeReconfigure {wm withdraw .}
    Fx_Entry::AfterReconfigure {Fx:Reconfigure}
    Fx_Menubar::BeforeReconfigure {wm withdraw .}
    Fx_Menubar::AfterReconfigure {Fx:Reconfigure}
    Fx:Reconfigure
    bind . <Configure> {Fx:Configure}
}

proc Fx:Configure {} {
    global fx_config

    set geo [split [wm geometry .] +]
    set x [lindex $geo 1]
    set y [lindex $geo 2]
    set fx_config(geometry,\$maingeom\$) [list $x $y]
}

proc Fx:Reconfigure {} {
    after idle {
	global fx_config fx_rootx fx_rooty

	wm resizable . true true
        update
        update idletasks
	if {![info exists fx_config(geometry,\$maingeom\$)]} {
	    Fx:Configure
	    bind . <Configure> {Fx:Configure}
	}
        update
        update idletasks
	set fx_rootx [lindex $fx_config(geometry,\$maingeom\$) 0]
	set fx_rooty [lindex $fx_config(geometry,\$maingeom\$) 1]
	if {$fx_rootx == 0 && $fx_rooty == 0} {
	    set fx_rootx [expr [winfo screenwidth .]/2 - [winfo reqwidth .]/2]
	    set fx_rooty [expr [winfo screenheight .]/2 - [winfo reqheight .]/2]
	}
	wm resizable . false false
        wm geometry . +$fx_rootx+$fx_rooty
        wm deiconify .
    }
}

if {[info exists fx_debug] && $fx_debug == 1} {
    puts "auto-loaded fx_script.tcl"
}
