# $RCSfile: fx_table.tcl,v $ $Revision: 2.0.0.1 $ $Date: 2000/11/09 15:06:01 $
#
# Copyright (C) 1995-2000 Herrin Software Development, Inc.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


# Table routines (import/export)

# Export routines

proc Fx:ExportBrowse {} {
    global fx_export_file fx_export fx_config

    if {$fx_config(\$export\$,format) == "csv"} {
	set extensions { {CSV .csv} {AS .as} }
    } elseif {$fx_config(\$export\$,format) == "as"} {
	set extensions { {AS .as} {CSV .csv} }
    } else {
	set extensions { {CSV .csv} {AS .as} }
    }
    set fx_export_file [Fx:BusyExec .fx_export \
	    [list tk_getSaveFile -defaultextension ".$fx_config(\$export\$,format)" -filetypes \
	    $extensions -parent .fx_export -title Browse]]
    if {[string length $fx_export_file] > 0 && [string length [file extension $fx_export_file]] == 0} {
	append fx_export_file .$fx_config(\$export\$,format)
    }
}

proc Fx:ExportOnSelect {sel} {
    global fx_config

    set fx_config(\$export\$,dateformat) $sel
}

proc Fx:ExportExec {tables} {
    global fx_config fx_export_floppy fx_export_file fx_export_dates \
	    fx_export_trans fx_export

    if {[string compare $fx_config(\$export\$,format) csv] == 0} {
	if {[catch {eval qddb_export $fx_config(\$export\$,format) \
		-dateformat [list [lindex $fx_export_trans(csv) $fx_config(\$export\$,dateformat)]] \
		-columntitles $fx_config(\$export\$,columntitles) -concatsheets on $tables} exportval]} {
	    Fx:Dialog "Error" "$exportval" .fx_export.dialog 
	    return 1
	}
    } else {
	if {[catch {eval qddb_export $fx_config(\$export\$,format) \
		-dateformat [list $fx_config(\$export\$,dateformat)] \
		-columntitles $fx_config(\$export\$,columntitles) $tables} exportval]} {
	    Fx:Dialog "Error" "$exportval" .fx_export.dialog 
	    return 1
	}
    }
    if {[string compare $fx_config(\$export\$,placement) file] == 0} {
	if {[catch [list open $fx_export_file w] fd]} {
	    Fx:Dialog "Error" "Cannot open file \"$fx_export_file\"" .fx_export.dialog 
	    return 1
	}
	puts -nonewline $fd $exportval
	close $fd
    } else {
	set tmpname [TempNam]
	if {[catch [list open $tmpname w 0400] fd]} {
	    Fx:Dialog "Error" "Cannot open temporary file \"$tmpname\"" .fx_export.dialog 
	    return
	}
	puts -nonewline $fd $exportval
	close $fd
	update idletasks
	while {[catch "exec mcopy -tn $tmpname a:$fx_export_floppy"] != 0} {
	    set ans [fx_dialog .fx_export.dialog "Floppy Error" \
		    "A floppy error has occurred.   Would you like to try formatting the floppy?" \
		    error 0 Yes No]
	    if {$ans == 0} {
		update idletasks
		fx_message .fx_export.msg "Formatting..."
		update idletasks; update
		if {[catch "exec mformat a:"] != 0} {
		    .fx_export.msg.msg configure -text "Format failed!"
		    update idletasks; update
		    catch "exec sleep 2"
		    destroy .fx_export.msg
		    update idletasks; update
		    set ans 1
		} else {
		    .fx_export.msg.msg configure -text "Format succeeded!"
		    update idletasks; update
		    catch "exec sleep 2"
		    destroy .fx_export.msg
		    update idletasks; update
		}
	    }
	    if {$ans == 1} {
		set ans [fx_dialog .fx_export.dialog "Floppy Error" \
			"Please make sure that an error-free formatted floppy is in drive A:" \
			error 0 OK Cancel]
		if {$ans != 0} {
		    break
		}
	    }
	}
	catch "exec rm -f $tmpname"
    }
    return 0
}

proc Fx:ExportTable {tables} {
    global fx_config fx_export_var fx_export_dates fx_export_trans fx_export

    set w .fx_export
    set fx_export_var 0
    toplevel $w
    wm withdraw $w
    wm title $w "Export"
    wm iconname $w "Export"
    wm protocol $w WM_DELETE_WINDOW [list set fx_export_var 0]
    frame $w.f
    pack $w.f -side top -expand on -fill both
    set f $w.f
    set l $w.f.left
    set r $w.f.right
    frame $l -relief flat
    pack $l -side left -expand on -fill both -padx 5m -pady 5m
    frame $l.f0
    pack $l.f0 -side top -expand on -fill both
    label $l.f0.label -text "Date format" -justify left -relief ridge -bd 1
    pack $l.f0.label -side top -expand off -fill x -anchor n
    Fx_SingleColumnListBox fx_export -w $l.f0.l -single_select on \
	    -exportselection off -headings {Date format      Example} -onselect Fx:ExportOnSelect
    #pack forget $l.f0.l.l1 ;# get rid of title listbox
    pack forget $l.f0.l.hs ;# get rid of horizontal scrollbar
    pack $l.f0.l -expand on -fill both -anchor n
    pack $l.f0.l.l2 -expand on -fill both
    if {[info exists fx_config(\$export\$,dateformat)]} {
	$l.f0.l.l2 activate $fx_config(\$export\$,dateformat)
	$l.f0.l.l2 selection set $fx_config(\$export\$,dateformat)
    } else {
	set fx_config(\$export\$,dateformat) 0
	$l.f0.l.l2 activate $fx_config(\$export\$,dateformat)
	$l.f0.l.l2 selection set $fx_config(\$export\$,dateformat)
    }
    frame $r
    pack $r -side right -fill y -padx 5m -pady 5m
    label $r.l -text "Export format" -justify left -relief ridge -bd 1
    pack $r.l -side top -fill x
    radiobutton $r.r0 -text "CSV" -anchor w -justify left \
	    -variable fx_config(\$export\$,format) -value csv -width 10 \
	    -command {
	global fx_export fx_export_dates

	fx_export ClearRows
	fx_export ClearContents
	fx_export AppendRows $fx_export_dates($fx_config(\$export\$,format))
	fx_export Display
    }
    pack $r.r0 -side top
    radiobutton $r.r1 -text "AS" -anchor w -justify left \
	    -variable fx_config(\$export\$,format) -value as -width 10 \
	    -command {
	global fx_export fx_export_dates

	fx_export ClearRows
	fx_export ClearContents
	fx_export AppendRows $fx_export_dates($fx_config(\$export\$,format))
	fx_export Display
    }
    pack $r.r1 -side top    
    if {[string compare [qddb_util datestyle] "us"] == 0} {
	set fx_export_dateformat "mm/dd/yy"
	set fx_export_strftime   "%m/%d/%y"
    } else {
	set fx_export_dateformat "dd/mm/yy"
	set fx_export_strftime   "%d/%m/%y"
    }
    catch "unset fx_export_dates"
    catch "unset fx_export_trans"
    lappend fx_export_dates(as) \
	    "default          $fx_export_dateformat" \
	    "Mmmm dd, yyyy    January 06, 1997" \
	    "Mm d, yyyy       Jan 6, 1997" \
	    "dd Mm yy         06 Jan 97" \
	    "mm/dd/yy         01/06/97" \
	    "mm.dd.yy         01.06.97" \
	    "yyyy-mm-dd       97-01-06" \
	    "yy-mm-dd         97-01-06" \
	    "yyyy mm dd       1997 01 06" \
	    "yy mm dd         97 01 06" \
	    "yyyymmdd         19970106" \
	    "yymmdd           970106" \
	    "dd/mm/yy         06/01/97" \
	    "dd.mm.yyyy       06.01.1997" \
	    "Mm dd, yyyy      Jan 06, 1997" \
	    "Mmmm yyyy        January 1997"
    lappend fx_export_dates(csv) \
	    "default          $fx_export_dateformat" \
	    "dd Mm yy         06 Jan 97" \
	    "mm/dd/yy         01/06/97" \
	    "mm.dd.yy         01.06.97" \
	    "yyyy-mm-dd       1997-01-06" \
	    "yy-mm-dd         97-01-06" \
	    "yyyy mm dd       1997 01 06" \
	    "yy mm dd         97 01 06" \
	    "yyyymmdd         19970106" \
	    "yymmdd           970106" \
	    "dd/mm/yy         06/01/97" \
	    "dd.mm.yyyy       06.01.1997" \
	    "Mmmm yyyy        January 1997" \
	    "Days 1900        35436"
    lappend fx_export_trans(as) 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15
    lappend fx_export_trans(csv) $fx_export_strftime \
	    "%d %m %y" \
	    "%m/%d/%y" \
	    "%m.%d.%y" \
	    "%Y-%m-%d" \
	    "%y-%m-%d" \
	    "%Y %m %d" \
	    "%y %m %d" \
	    "%Y%m%d" \
	    "%y%m%d" \
	    "%d/%m/%y" \
	    "%d.%m.%Y" \
	    "%B %Y" \
	    "days1900"
    if {![info exists fx_config(\$export\$,format)]} {
	set fx_config(\$export\$,format) csv
    }
    if {[string length $fx_config(\$export\$,format)] == 0} {
	set fx_config(\$export\$,format) csv
    }
    fx_export AppendRows $fx_export_dates($fx_config(\$export\$,format))
    fx_export Display

    frame $w.m
    pack $w.m -side top -expand off -fill x -padx 5m -pady 5m
    if {![info exists fx_config(\$export\$,columntitles)]} {
	set fx_config(\$export\$,columntitles) 1
    }
    checkbutton $w.m.coltit -text "Column titles" -variable fx_config(\$export\$,columntitles) \
	    -anchor w
    pack $w.m.coltit -side left

    set l $w
    frame $l.f1
    pack $l.f1 -side top -fill x -padx 5m -pady 5m
    frame $l.f1.f0
    pack $l.f1.f0 -side top -fill x
    radiobutton $l.f1.f0.r0 -text "DOS Floppy" -variable fx_config(\$export\$,placement) \
	    -value "floppy" -justify left -anchor w -width 12 -command "
        $l.f1.f0.e configure -state normal
        after idle [list focus $l.f1.f0.e]
        $l.f1.f1.e configure -state disabled
        $l.f1.f1.b configure -state disabled
    "
    pack $l.f1.f0.r0 -side left
    entry $l.f1.f0.e -width 20 -textvariable fx_export_floppy
    pack $l.f1.f0.e -side right -expand on -fill x
    label $l.f1.f0.l -text "A:"
    pack $l.f1.f0.l -side right
    frame $l.f1.f1
    pack $l.f1.f1 -side top -fill x
    radiobutton $l.f1.f1.r1 -text "File" -variable fx_config(\$export\$,placement) \
	    -value "file" -justify left -anchor w -width 12 -command "
        $l.f1.f1.e configure -state normal
        $l.f1.f1.b configure -state normal
        after idle [list focus $l.f1.f1.e]
        $l.f1.f0.e configure -state disabled
    "
    pack $l.f1.f1.r1 -side left
    entry $l.f1.f1.e -width 20 -textvariable fx_export_file
    button $l.f1.f1.b -text "Browse..." -command Fx:ExportBrowse
    pack $l.f1.f1.b -side right
    pack $l.f1.f1.e -side right -expand on -fill x
    if {![info exists fx_config(\$export\$,placement)]} {
	set fx_config(\$export\$,placement) "file"
    } elseif {[string length $fx_config(\$export\$,placement)] == 0} {
	set fx_config(\$export\$,placement) "file"
    }
    if {[string compare $fx_config(\$export\$,placement) "file"] == 0} {
	$l.f1.f1.r1 invoke
    } else {
	$l.f1.f0.r0 invoke
    }

    frame $w.b 
    pack $w.b -side bottom -expand off -fill x -padx 5m -pady 5m
    button $w.b.ok -text "Ok" -width 10 -command {
	set fx_export_var 1
    }
    pack $w.b.ok -side left
    button $w.b.cancel -text "Cancel" -width 10 -command {
	set fx_export_var 0
    }
    pack $w.b.cancel -side right
    Fx:SetGeometry $w
    fx_bindgeom $w
    wm deiconify $w
    tkwait variable fx_export_var
    while {$fx_export_var} {
	set var fx_export_[set fx_config(\$export\$,placement)]
	global $var
	if {[string length [string trim [set $var]]] > 0} {
	    if {[Fx:BusyExec .fx_export [list Fx:ExportExec $tables]] == 0} {
		break
	    }
	}
	tkwait variable fx_export_var
    }
    Fx:SaveGeometry $w
    catch "itcl::delete object fx_export"
    catch "destroy .fx_export"
}

# Import routines

proc Fx:ImportBrowse {} {
    global fx_import_file fx_import fx_config

    if {$fx_config(\$import\$,format) == "csv"} {
	set extensions { {CSV .csv} {All *} }
    } else {
	set extensions { {All *} {CSV .csv} }
    }
    set fx_import_file [Fx:BusyExec .fx_import \
	    [list tk_getOpenFile -defaultextension ".$fx_config(\$import\$,format)" -filetypes \
	    $extensions -parent .fx_import -title Browse]]
}

proc Fx:DoImport {schema table} {
    global fx_config fx_import_tablevar fx_import_tuple

    set idx {$import$}
    set max [qddb_table col maxnum $table]
    set x 1
    foreach i $fx_config($idx,print) {
	qddb_table col configure $table $x -name $i
	lappend printlist $i
	incr x
    }
    if {![info exists printlist]} {return}
    set maxrow [qddb_table row maxnum $table]
    set maxcol [qddb_table col maxnum $table]
    foreach i $printlist {
	lappend attrlist [list $i fx_import_tuple($i)]
    }
    foreach i [array names fx_config *,unique] {
	if {$fx_config($i) != 0} {
	    set unique([lindex [split $i ,] 0]) ""
	}
    }
    fx_message .fx_import.msg [format "Importing %7d" 1]
    toplevel .fx_import2
    wm withdraw .fx_import2
    focus .fx_import2
    catch {grab .fx_import2}
    for {set i 1} {$i <= $maxrow} {incr i} {
	.fx_import.msg.msg configure -text [format "Importing %7d" $i]
	update; update idletasks
	catch {unset fx_import_tuple}
	set t [qddb_tuple new $schema]
	set v [qddb_view define $t $attrlist]
	set x 1
	set badtuple 0
	foreach j $printlist {
	    set val [qddb_table cell getval $table $i $x]
	    if {[info exists unique($j)]} {
		set k [qddb_search $schema -prunebyattr $j w $val]
		if {[string length $k] > 0} {
		    if {[llength [qddb_keylist get $k]] > 0} {
			.fx_import.msg.msg configure -text [format "Skipping non-unique tuple" $i]
			set badtuple 1
			break
		    }
		}
	    }
	    if {[catch {set fx_import_tuple($j) $val}] != 0} {
		catch {set fx_import_tuple($j) [expr int($val)]}
	    }
	    incr x
	}
	if {!$badtuple} {
	    qddb_tuple write $t
	}
	qddb_tuple delete $t
    }
    destroy .fx_import2
}

proc Fx:ImportCheck {table} {
    global fx_config

    set idx {$import$}
    set x 1
    foreach i $fx_config($idx,print) {
	qddb_table col configure $table $x -name $i
	incr x
    }
    set maxcol [qddb_table col maxnum $table]
    for {set i 1} {$i <= $maxcol} {incr i} {
	set val [qddb_table col cget $table $i -name]
	if {"$val" != ""} {
	    set names($val) ""
	}
    }
    foreach i [array names fx_config *,mandatory] {
	if {$fx_config($i) != 0} {
	    if {![info exists names([lindex [split $i ,] 0])]} {
		fx_dialog .fx_import.dialog "Mandatory column doesn't exist!" \
			"You must import all mandatory columns" error 0 Ok
		return 1
	    }
	}
    }
    return 0
}

proc Fx:ImportConfig {schema menubar table} {
    global fx_dialog

    while {1} {
	Fx_Import fx_importconfig -schema $schema -title "Import Configuration" \
		-idx {$import$} -attrs [qddb_schema leaves $schema] \
		-restrict [$menubar cget -restrict]/:import_config:
	tkwait window .fx_importconfig
	if {"$fx_dialog(fx_importconfig)" != "cancel"} {
	    update; update idletasks
	    if {[Fx:ImportCheck $table] == 0} {
		Fx:DoImport $schema $table
		break
	    }
	} else {
	    break
	}
    }
}

proc Fx:ImportExec {schema menubar file} {
    global fx_config fx_import_var fx_import qddb_perfmeter fx:status_variable

    update; update idletasks
    set var fx_import_[set fx_config(\$import\$,placement)]
    global $var
    if {"$var" == "floppy"} {
	set tmpname [TempNam]
	set ans 0
	while {[catch "exec mcopy -tn a:$file $tmpname"] != 0 && $ans == 0} {
	    set ans [fx_dialog .fx_import.dialog "Floppy Error" \
		    "A floppy error has occurred.   Would you like to try another floppy?" \
		    error 0 Yes No]
	    if {$ans == 0} {
		update idletasks
		fx_message .msg "Trying to find file..."
		update idletasks
	    }
	}
	catch {destroy .msg}
	if {$ans} {
	    return 1
	}
	if {[catch {open $tmpname r} fd] != 0} {
	    Fx:Dialog .fx_import.dialog "Cannot open file $tmpname: $fd"
	    return 1
	}
	set contents [read $fd]
	close $fd
    } else {
	if {[catch {open $file r} fd] != 0} {
	    Fx:Dialog .fx_import.dialog "Cannot open file $file: $fd"
	    return 1
	}
	set contents [read $fd]
	close $fd
    }
    fx_message .fx_import.msg ""
    set qddb_perfmeter 0
    .fx_import.msg.msg configure -textvariable qddb_perfmeter
    catch {grab .fx_import.msg.msg}
    set table [qddb_import csv $contents]
    unset contents
    catch {destroy .fx_import.msg}
    Fx:ImportConfig $schema $menubar $table
    set {fx:status_variable} "Imported [qddb_table row maxnum $table] records"
    catch {qddb_table delete $table}
    return 0
}

proc Fx:ImportTable {schema menubar} {
    global fx_config fx_import_var fx_import

    set w .fx_import
    set fx_import_var 0
    toplevel $w
    wm withdraw $w
    wm title $w "Import"
    wm iconname $w "Import"
    wm protocol $w WM_DELETE_WINDOW [list set fx_import_var 0]
    frame $w.f
    pack $w.f -side top -expand on -fill both
    set f $w.f
    set r $w.f
    if {![info exists fx_config(\$import\$,format)]} {
	set fx_config(\$import\$,format) csv
    }
    if {[string length $fx_config(\$import\$,format)] == 0} {
	set fx_config(\$import\$,format) csv
    }
    label $r.l -text "Import format" -justify left -relief ridge -bd 1
    pack $r.l -side top -fill x
    radiobutton $r.r0 -text "CSV" -anchor w -justify left \
	    -variable fx_config(\$import\$,format) -value csv -width 10
    pack $r.r0 -side top -fill x -padx 5m -pady 5m
    
    set l $w.f
    frame $l.f1
    pack $l.f1 -side top -fill x -padx 5m -pady 5m
    frame $l.f1.f0
    pack $l.f1.f0 -side top -fill x
    radiobutton $l.f1.f0.r0 -text "DOS Floppy" -variable fx_config(\$import\$,placement) \
	    -value "floppy" -justify left -anchor w -width 12 -command "
        $l.f1.f0.e configure -state normal
        after idle [list focus $l.f1.f0.e]
        $l.f1.f1.e configure -state disabled
        $l.f1.f1.b configure -state disabled
    "
    pack $l.f1.f0.r0 -side left
    entry $l.f1.f0.e -width 20 -textvariable fx_import_floppy
    pack $l.f1.f0.e -side right -expand on -fill x
    label $l.f1.f0.l -text "A:"
    pack $l.f1.f0.l -side right
    frame $l.f1.f1
    pack $l.f1.f1 -side top -fill x
    radiobutton $l.f1.f1.r1 -text "File" -variable fx_config(\$import\$,placement) \
	    -value "file" -justify left -anchor w -width 12 -command "
        $l.f1.f1.e configure -state normal
        $l.f1.f1.b configure -state normal
        after idle [list focus $l.f1.f1.e]
        $l.f1.f0.e configure -state disabled
    "
    pack $l.f1.f1.r1 -side left
    entry $l.f1.f1.e -width 20 -textvariable fx_import_file
    button $l.f1.f1.b -text "Browse..." -command Fx:ImportBrowse
    pack $l.f1.f1.b -side right
    pack $l.f1.f1.e -side right -expand on -fill x
    if {![info exists fx_config(\$import\$,placement)]} {
	set fx_config(\$import\$,placement) "file"
    } elseif {[string length $fx_config(\$import\$,placement)] == 0} {
	set fx_config(\$import\$,placement) "file"
    }
    if {[string compare $fx_config(\$import\$,placement) "file"] == 0} {
	$l.f1.f1.r1 invoke
    } else {
	$l.f1.f0.r0 invoke
    }

    frame $w.b 
    pack $w.b -side bottom -expand off -fill x -padx 5m -pady 5m
    button $w.b.ok -text "Ok" -width 10 -command {
	update idletasks; update
	set fx_import_var 1
    }
    pack $w.b.ok -side left
    button $w.b.cancel -text "Cancel" -width 10 -command {
	set fx_import_var 0
    }
    pack $w.b.cancel -side right
    update idletasks; update
    wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
    wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
    Fx:SetGeometry $w
    fx_bindgeom $w
    wm deiconify $w
    tkwait variable fx_import_var
    while {$fx_import_var} {
	set var fx_import_[set fx_config(\$import\$,placement)]
	global $var
	if {[string length [string trim [set $var]]] > 0} {
	    if {[Fx:BusyExec .fx_import [list Fx:ImportExec $schema $menubar "[set $var]"]] == 0} {
		break
	    }
	}
	tkwait variable fx_import_var
    }
    Fx:SavePosition $w
    catch "destroy .fx_import"
}
