
/* $RCSfile: Changes.c,v $ $Revision: 2.2 $ $Date: 2002/02/05 12:28:09 $
 *
 * Copyright (C) 1993-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include "Qddb.h"

/* EXPORTED:
 *	void Qddb_ChangeEntry
 *	void Qddb_AddEntry
 */

static int FindNextNumber _ANSI_ARGS_((char *));

void Qddb_ChangeEntry(schema, File, Locked, ThisEntry, Type)
    Qddb_Schema		*schema;
    int			File;
    Qddb_Boolean	Locked;
    Qddb_Entry		ThisEntry;
    Qddb_KeyListType	Type;
{
    char		*RelationFN = schema->RelationName;
    char		Path[MAXFILENAMELEN], file_name[BUFSIZ];
    int			EntryNumber;

    strcpy(Path, RelationFN);
    if (Type != ADDITION)
	strcat(Path, "/Changes/");
    else
	strcat(Path, "/Additions/");
    sscanf(ThisEntry[0], "%*s %*c %d", &EntryNumber);
    sprintf(file_name, "%s%d", Path, EntryNumber);
    if (File == -1 && (File = Open(file_name, O_RDWR|O_CREAT, 0666)) == -1) {
	PANIC("Cannot open changes file");
    } else {
#if defined(RECORDLOCKING)
	if (Qddb_LockSection(File, F_WRLCK, (off_t)0, (off_t)0, True) == -1)
	    perror("warning: failed to obtain lock\n");
#endif
	lseek(File, (off_t)0, 0);
#if defined(HAVE_FTRUNCATE)
	ftruncate(File, (off_t)0);
#else
	/* we lose the lock here */
	close(Open(file_name, O_RDWR|O_CREAT|O_TRUNC, 0666));
#if defined(RECORDLOCKING)
	if (Qddb_LockSection(File, F_WRLCK, (off_t)0, (off_t)0, True) == -1)
	    perror("warning: failed to obtain lock\n");
#endif
#endif
    }
    Qddb_WriteEntry(schema, File, ThisEntry);
    if (schema->UseCachedSecondarySearch == True)
	Qddb_UpdateWordList(schema, Type, EntryNumber, ThisEntry);
#if defined(RECORDLOCKING)
    if (Locked == False) {
	Qddb_UnlockSection(File, (off_t)0, (off_t)0);
	Close(File);
    } else {
	if (Qddb_LockSection(File, F_RDLCK, (off_t)0, (off_t)0, True) == -1)
	    perror("failed to revert to read lock\n");
    }
#else
    Close(File);
#endif
}

int Qddb_AddEntry(schema, NewFile, Lockit, ThisEntry)
    Qddb_Schema		*schema;
    int			*NewFile;
    Qddb_Boolean	Lockit;
    Qddb_Entry		ThisEntry;
{
    char		*RelationFN = schema->RelationName;
    char		Path[MAXFILENAMELEN], *tmpbuf, *tmp;
    int			EntryNumber;
    int			File;
    int			FindNextNumber _ANSI_ARGS_((char *));

    strcpy(Path, RelationFN);
    strcat(Path, "/Additions/");
    EntryNumber = FindNextNumber(Path);
    if (EntryNumber == -1)
	return -1;
    tmpbuf = Malloc(MAXFILENAMELEN);
    sprintf(tmpbuf, "%s%d", Path, EntryNumber);
    if ((File = Open(tmpbuf, O_RDWR|O_CREAT|O_EXCL, 0666)) == -1) {
	char		errbuf[BUFSIZ];

	sprintf(errbuf, "Cannot open Additions file: %d\n", EntryNumber);
	PANIC(errbuf);
    }
#if defined(RECORDLOCKING)
    Qddb_LockSection(File, F_WRLCK, (off_t)0, (off_t)0, True);
#endif
    sprintf(tmpbuf, "%%0 V %d", EntryNumber);
    tmp = ThisEntry[0];
    ThisEntry[0] = tmpbuf;
    Qddb_WriteEntry(schema, File, ThisEntry);
    if (schema->UseCachedSecondarySearch == True)
	Qddb_UpdateWordList(schema, ADDITION, EntryNumber, ThisEntry);
    Free(tmp);
#if defined(RECORDLOCKING)
    if (Lockit == False) {
	Qddb_UnlockSection(File, (off_t)0, (off_t)0);
	Close(File);
    } else {
	Qddb_LockSection(File, F_RDLCK, (off_t)0, (off_t)0, True);
	if (NewFile != NULL)
	    *NewFile = File;
    }
#else
    Close(File);
#endif
    return EntryNumber;
}

static int FindNextNumber(Path)
    char		*Path;
{
    char		tmpPath[MAXPATHLEN], buf[MAXPATHLEN];
    size_t              filesize;
    int			NextNumber;
    int			File;

    strcpy(tmpPath, Path);
    strcat(tmpPath, "/NextEntry");
    if ((File = Open(tmpPath, O_RDWR|O_CREAT, 0666)) == -1)
	PANIC("Cannot open Additions/NextEntry file");
#if defined(RECORDLOCKING)
    if (Qddb_LockSection(File, F_WRLCK, (off_t)0, (off_t)0, True) == -1)
	return -1;
#endif
    filesize = SizeOfFile(File);
    read(File, buf, filesize);
    buf[filesize] = '\0';
    sscanf(buf, "%d", &NextNumber);
    NextNumber++;
    sprintf(buf, "%d\n", NextNumber);
    lseek(File, (off_t)0, 0);
    write(File, buf, strlen(buf));
#if defined(RECORDLOCKING)
    Qddb_UnlockSection(File, (off_t)0, (off_t)0);
#endif
    close(File);
    return NextNumber;
}

