
/* $RCSfile: Date.c,v $ $Revision: 2.2 $ $Date: 2004/05/27 16:33:01 $
 *
 * Copyright (C) 1993-2004 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * Copyright 1991-1994 Karl Lehenbauer and Mark Diekhans.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies.  Karl Lehenbauer and
 * Mark Diekhans make no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */

#include "Qddb.h"

/*
 *-----------------------------------------------------------------------------
 *
 * Qddb_GetTimeZone --
 *   Determines the current timezone.  The method varies wildly between
 * different Unix implementations, so its hidden in this function.
 *
 * Parameters:
 *   o currentTime (I) - The clock value that is to be used for the current
 *     time.  This is really a time_t, but is long so it can be in tclExtend.h
 *     without requiring time_t.
 *
 * Returns:
 *    Minutes east of GMT.
 *-----------------------------------------------------------------------------
 */
int
Qddb_GetTimeZone (currentTime)
    time_t currentTime;
{
#if defined(HAVE_TM_TZADJ) && HAVE_TM_TZADJ == 1
    static time_t    curTime;
    static struct tm *timeDataPtr;
    static int       timeZone, tz_init = False;

    if (tz_init == False) {
	curTime = (time_t) currentTime;
	timeDataPtr = localtime (&curTime);
	timeZone = timeDataPtr->tm_tzadj  / 60;
	if (timeDataPtr->tm_isdst)
	    timeZone += 60;
	tz_init = True;
    }
    return timeZone;
#elif defined(HAVE_TIMEZONE_VAR) && HAVE_TIMEZONE_VAR == 1 && defined(HAVE_TZSET) && HAVE_TZSET == 1
    /* Linux doesn't support the gettimeofday timezone structure,
     * so define this before HAVE_GETTIMEOFDAY.
     */
    static int setTZ = False;
    int        timeZone;

    if (!setTZ) {
        tzset ();
        setTZ = True;
    }
    timeZone = timezone / 60;
    return timeZone;
#elif defined(HAVE_TM_GMTOFF) && HAVE_TM_GMTOFF == 1
    static struct tm *timeDataPtr;
    static int        timeZone, tz_init = False;

    if (tz_init == False) {
	timeDataPtr = localtime (&currentTime);
	timeZone = -(timeDataPtr->tm_gmtoff / 60);
	if (timeDataPtr->tm_isdst)
	    timeZone += 60;
	tz_init = True;
    }
    return timeZone;
#elif defined(HAVE_GETTIMEOFDAY) && HAVE_GETTIMEOFDAY == 1
    static struct timeval  tv;
    static struct timezone tz;
    static int             timeZone, tz_init = False;

    if (tz_init == False) {
	gettimeofday (&tv, &tz);
	timeZone = tz.tz_minuteswest;
	tz_init = True;
    }
    return timeZone;
#else
    error: autoconf did not figure out how to determine the timezone.
#endif
}

