


/* $RCSfile: EntryParse.y,v $ $Revision: 2.1 $ $Date: 2001/03/25 12:23:35 $
 *
 * Copyright (C) 1993-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

%{

#include "Qddb.h"

/* Define yyparse() to be Qddb_EntryParse() so that multiple parsers can
 * occupy the same executable.
 */
#define yyparse	Qddb_EntryParse
#define yylex   Qddblex

int yylex();

static Qddb_Schema	*CurrentSchema;
static Qddb_InCoreEntry	*CurrentEntries = NULL;
static unsigned int	CurrentEntriesLength;

void Qddb_EntryParse_Set(schema)
    Qddb_Schema		*schema;
{
    CurrentSchema = schema;
}

Qddb_InCoreEntry *Qddb_EntryParse_Get()
{
    return CurrentEntries;
}

static unsigned int	CurrentAttribute = 0;
static unsigned int	CurrentAncestors[MAXIMUM_NUMBER_OF_SCHEMA_LEVELS];
static unsigned int	CurrentInstances[MAXIMUM_NUMBER_OF_SCHEMA_LEVELS];
static unsigned int	CurrentLevel = 1;
static int              Qddb_FindAttribute _ANSI_ARGS_((char *));
%}

%pure_parser
%expect 1

%token STRING CSTRING
%token ALIAS VERBOSENAME
%token LPAREN RPAREN ASTERIK COMMA DOUBLEQUOTE SINGLEQUOTE EQUALS
%token BLANKLINE NUMBER TYPE DEFAULTVALUE

%type <String>		STRING CSTRING
%type <Integer>		UniqueSequenceNumber Identifier IdentifierLparen

%union {
    char	*String;
    double	Real;
    int		Integer;
};


%%

FileOfEntries:
    WholeEntry {
    };

WholeEntry:
    UniqueSequenceNumber Entry {
	CurrentEntriesLength = CurrentAttribute;
        CurrentEntries = (Qddb_InCoreEntry *)Realloc(CurrentEntries, 
						     sizeof(Qddb_InCoreEntry)*(CurrentAttribute+2));
	CurrentEntries[CurrentAttribute].SequenceNumber = 0;
	CurrentEntries[CurrentAttribute+1].SequenceNumber = 0;
	CurrentAttribute = 0;
    };

Entry:
    /* return a TEMPLATE */
    Record {
    } |
    Record Entry {
    };

Record:
    /* return an INSTANCE */
    IndividualRecord {
    } |
    IndividualRecord Record {
    };

IndividualRecord:
    /* return a TEMPLATE */
    Identifier {
    } |
    IdentifierLparen Record RPAREN {
        CurrentInstances[CurrentLevel] = 0;
	CurrentLevel--;
    };

IdentifierLparen:
    Identifier LPAREN {
	CurrentLevel++;
        CurrentInstances[CurrentLevel] = 0;
	CurrentAttribute--;
    };

Identifier:
    STRING {
	Qddb_SchemaNode	*ThisAttribute;
	int		AttrNo;
	int		i;

    	/* return an INSTANCE */
	AttrNo = Qddb_FindAttribute($1);
	Free($1);
	if (AttrNo < 0) {
	    qddb_errmsg = Malloc(BUFSIZ);
	    qddb_errno = QDDB_ERRNO_PARSE_ERROR;
	    sprintf(qddb_errmsg, "Attribute named '%s' is either misplaced or it is not \
an attribute name in Schema\n", $1);
	    YYERROR;
	}

	ThisAttribute = CurrentSchema->Entries+AttrNo;
        CurrentEntries = 
            (Qddb_InCoreEntry *)Realloc(CurrentEntries,
					sizeof(Qddb_InCoreEntry)*(CurrentAttribute+2));
	CurrentEntries[CurrentAttribute].BytesOfData = 0;
	CurrentEntries[CurrentAttribute].AttributeNumber = AttrNo;
	$$ = ThisAttribute->Number;
        if (CurrentAncestors[CurrentLevel] != ThisAttribute->Number ||
	    CurrentInstances[CurrentLevel] == 0) {
	    CurrentInstances[CurrentLevel] = 1;
	    CurrentAncestors[CurrentLevel] = ThisAttribute->Number;
	} else {
            CurrentInstances[CurrentLevel]++;
	}
	for (i = 0; i < CurrentLevel; i++) {
	    CurrentEntries[CurrentAttribute].Instance[i] = CurrentInstances[i+1];
	}
	CurrentAttribute++;
    } |
    STRING EQUALS CSTRING {
	Qddb_SchemaNode	*ThisAttribute;
	int		AttrNo;
	int		i;

    	/* The grammar contains an anomaly here.  Constructs of
    	 * the form:
    	 *	string (
    	 *		string1 = "hello" (
    	 *			string2 = "hi"
    	 *		)
    	 *	);
    	 * do not make any sense, but the grammar is so much 
    	 * simpler due to the allowance of this construct that
    	 * I deem it easier to check for correctness in the
    	 * action of the grammar, rather than the grammar 
    	 * itself.  The error message will be the same, it will
    	 * be produced by the C code associated with the action
    	 * instead of the parser itself.
    	 * 
    	 * Correct syntax is as follows:
    	 * 	string (
    	 *		string1 (
    	 *			string2 = "something"
    	 *			string3 = "something"
    	 *		)
    	 *		string4 = "something else"
    	 *	);
    	 */
	 AttrNo = Qddb_FindAttribute($1);
	 if (AttrNo < 0) {
	     qddb_errmsg = Malloc(BUFSIZ);
	     qddb_errno = QDDB_ERRNO_PARSE_ERROR;
	     sprintf(qddb_errmsg, "Attribute named '%s' is either misplaced or it is not \
an attribute name in Schema", $1);
	     YYERROR;
	 }
	 Free($1);
         ThisAttribute = CurrentSchema->Entries+AttrNo;
         CurrentEntries = 
            (Qddb_InCoreEntry *)Realloc(CurrentEntries,
					sizeof(Qddb_InCoreEntry)*(CurrentAttribute+2));
	 CurrentEntries[CurrentAttribute].AttributeNumber = AttrNo;
	 CurrentEntries[CurrentAttribute].Data = $3;
	 CurrentEntries[CurrentAttribute].Marked = Inapplicable;
	 CurrentEntries[CurrentAttribute].BytesOfData = strlen($3);
	 CurrentEntries[CurrentAttribute].SequenceNumber = CurrentEntries[0].SequenceNumber;
	 $$ = ThisAttribute->Number;
	 if (CurrentAncestors[CurrentLevel] != ThisAttribute->Number ||
	     CurrentInstances[CurrentLevel] == 0) {
	     CurrentInstances[CurrentLevel] = 1;
	     CurrentAncestors[CurrentLevel] = ThisAttribute->Number;
	 } else {
             CurrentInstances[CurrentLevel]++;
	 }
         for (i = 0; i < CurrentLevel; i++) {
	     CurrentEntries[CurrentAttribute].Instance[i] = CurrentInstances[i+1];
	 }
         CurrentAttribute++;
    };

UniqueSequenceNumber:
    NUMBER EQUALS CSTRING {
	CurrentEntries = (Qddb_InCoreEntry *)Calloc(sizeof(Qddb_InCoreEntry)*2);
	CurrentEntriesLength = 0;
	CurrentAttribute = 0;
	CurrentEntries[0].SequenceNumber = atoi($3);
	CurrentEntries[0].BytesOfData = 0;
	CurrentEntries[0].AttributeNumber = 0;
        CurrentEntries[1].SequenceNumber = 0;
	Free($3);
    };

%%

static int Qddb_FindAttribute(string)
    char		*string;
{
    unsigned int	i, j;

    for (i = 1; i <= CurrentSchema->NumberOfAttributes; i++) {
	if (strcmp(string, CurrentSchema->Entries[i].Name) == 0) {
	    /* this is a possibility.. */
	    if (CurrentLevel != CurrentSchema->Entries[i].Level) {
		continue;
	    }
	    /* Check out the ancestors */
	    if (CurrentSchema->Entries[i].Level != CurrentLevel)
		continue;
	    for (j = 1; j < CurrentLevel; j++) {
		if (CurrentSchema->Entries[i].AncestorNumber[j] != CurrentAncestors[j]) {
		    break;
		}

	    }
	    if (j != CurrentLevel) {
		continue;
	    }
	    return i;
	}
    }
    return -1;
}

void Qddb_ResetEntryParser()
{
    CurrentEntries = NULL;
    CurrentAttribute = 0;
    CurrentLevel = 1;
    CurrentInstances[0] = 0;
    CurrentInstances[1] = 0;
    CurrentAncestors[0] = 0;
    CurrentAncestors[1] = 0;
    Qddb_ResetSchemaLex();
}
