
/* $RCSfile: File.c,v $ $Revision: 2.0.0.1 $ $Date: 2000/11/09 15:06:03 $
 *
 * Copyright (C) 1993-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */



#include "Qddb.h"

int Open(Name, Flags, Mode)
    char		*Name;
    int			Flags;
    int			Mode;
{
    int			FileDescriptor;

    FileDescriptor = open(Name, Flags, 0666);
    return FileDescriptor;
}

int OpenDatabase(RelationFN, Mode)
    char		*RelationFN;
    int			Mode;
{
    char		TmpRelationFN[MAXPATHLEN];
    int			retval;

    strcpy(TmpRelationFN, RelationFN);
    strcat(TmpRelationFN, "/Database");
    retval = Open(TmpRelationFN, O_RDWR, Mode);
    if (retval == -1)
	return Open(TmpRelationFN, O_RDONLY, Mode);
    else
	return retval;
}

int OpenChangeFile(RelationFN, Number, Mode)
    char		*RelationFN;
    size_t		Number;
    int			Mode;
{
    char		TmpRelationFN[MAXPATHLEN];
    int			retval;

    sprintf(TmpRelationFN, "%s/Changes/%d", RelationFN, (int)Number);
    retval = Open(TmpRelationFN, O_RDWR, Mode);
    if (retval == -1)
	return Open(TmpRelationFN, O_RDONLY, Mode);
    else
	return retval;
}

int OpenAdditionFile(RelationFN, Number, Mode)
    char		*RelationFN;
    size_t		Number;
    int			Mode;
{
    char		TmpRelationFN[MAXPATHLEN];
    int			retval;

    sprintf(TmpRelationFN, "%s/Additions/%d", RelationFN, (int)Number);
    retval = Open(TmpRelationFN, O_RDWR, Mode);
    if (retval == -1)
	return Open(TmpRelationFN, O_RDONLY, Mode);
    else
	return retval;
}

int OpenHashTable(RelationFN, Mode)
    char		*RelationFN;
    int			Mode;
{
    char		TmpRelationFN[MAXPATHLEN];

    strcpy(TmpRelationFN, RelationFN);
    strcat(TmpRelationFN, "/HashTable");
    return Open(TmpRelationFN, O_RDONLY, Mode);
}

void Close(FileDesc)
    int		FileDesc;
{
    if (close(FileDesc) == -1)
    	perror("Close");
}

void Read(File, Buffer, SizeOfBuffer)
    int			File;
    char		*Buffer;
    size_t		SizeOfBuffer;
{
    int			NumberRead;

    if ((NumberRead=read(File, Buffer, SizeOfBuffer)) != SizeOfBuffer) {
    	Buffer[NumberRead] = 0;
    	fprintf(stderr, "Attempted %d, actually read %d of filedesc %d\n", 
    		(int)SizeOfBuffer, NumberRead, File);
	fflush(stderr);
    	perror("Read()");
    	exit(FILE_ERROR);
    }
}

void Write(File, Buffer, SizeOfBuffer)
    int			File;
    char		*Buffer;
    size_t		SizeOfBuffer;
{
    if (write(File, Buffer, SizeOfBuffer) != SizeOfBuffer) {
    	perror("Write():");
    	exit(FILE_ERROR);
    }
}

size_t SizeOfFile(File)
    int			File;
{
    struct stat		FileStatus;

    if (fstat(File, &FileStatus) == -1) {
    	perror("SizeOfFile()");
    	exit(FILE_ERROR);
    }
    return (size_t)FileStatus.st_size;
}

void PANIC(string)
    char		*string;
{
    fprintf(stderr, "%s, errno = %d\n", string, errno);
    fprintf(stderr, "Dumping core for analysis....\n");
    fflush(stderr);
    abort();
    exit(2);
}


#if defined(RECORDLOCKING)
int Qddb_LockSection(File, LockType, Start, Length, Wait)
    int			File;
    int			LockType;
    off_t		Start, Length;
    Qddb_Boolean	Wait;
{
    struct flock	Lock;

    Lock.l_type = LockType;
    Lock.l_whence = 0;
    Lock.l_start = Start;
    Lock.l_len = Length;
    if (Wait == True)
    	return fcntl(File, F_SETLKW, &Lock);
    else
    	return fcntl(File, F_SETLK, &Lock);
}

int Qddb_UnlockSection(File, Start, Length)
    int			File;
    off_t		Start, Length;
{
    struct flock	Lock;

    Lock.l_type = F_UNLCK;
    Lock.l_whence = 0;
    Lock.l_start = Start;
    Lock.l_len = Length;
    return fcntl(File, F_SETLKW, &Lock);
}
#else
int Qddb_LockSection(File, LockType, Start, Length, Wait)
    int			File;
    int			LockType;
    off_t		Start, Length;
    Qddb_Boolean	Wait;
{
    return 0;
}

int Qddb_UnlockSection(File, Start, Length)
    int			File;
    off_t		Start, Length;
{
    return 0;
}
#endif
