
/* $RCSfile: Index.c,v $ $Revision: 2.0.0.1 $ $Date: 2000/11/09 15:06:03 $
 *
 * Copyright (C) 1993-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */



#include "Qddb.h"

/* EXPORTED:
 *	void Qddb_MakeIndex(char *)
 */

static size_t		   NumberOfKeys;
static Qddb_KeyIndexIn	   *KeyIndex;
static size_t		   KeyIndexTop;
static size_t		   NumberOfNumericKeys;
static Qddb_NumericIndexIn *NumericIndex;
static size_t	 	   NumericIndexTop;

static Qddb_KeyNode *HashEntryToKeyNodes _ANSI_ARGS_((Qddb_Schema *, FILE *, Qddb_Hash *));
static void PurgeHashList _ANSI_ARGS_((Qddb_Bucket *, size_t));
static void PurgeHashBucket _ANSI_ARGS_((Qddb_KeyNode *));
static void PurgePosNode _ANSI_ARGS_((Qddb_PosNode *));
static void AddToHashList _ANSI_ARGS_((Qddb_Bucket *, Qddb_KeyNode *, off_t, size_t));
static void WriteHashList _ANSI_ARGS_((Qddb_Schema *, FILE *, FILE *, Qddb_Bucket *, \
				       Qddb_Boolean, size_t, size_t));
static void WriteKeyIndex _ANSI_ARGS_((Qddb_Schema *, FILE *, Qddb_KeyIndexIn *));
static void SortKeyIndex _ANSI_ARGS_((size_t, size_t));
static void WriteNumericIndex _ANSI_ARGS_((Qddb_Schema *, FILE *, Qddb_NumericIndexIn *));
static void SortNumericIndex _ANSI_ARGS_((size_t, size_t));
static void PurgeKeyIndex _ANSI_ARGS_((Qddb_Schema *, Qddb_KeyIndexIn *));
static void PurgeNumericIndex _ANSI_ARGS_((Qddb_Schema *, Qddb_NumericIndexIn *));
static Qddb_KeyNode *ConvertKeyListToKeyNode _ANSI_ARGS_((char *, Qddb_KeyList *));

static int ptr_strcmp1 _ANSI_ARGS_((char *, char **));
static int ptr_strcmp2 _ANSI_ARGS_((char **, char **));

static int ptr_strcmp1(str1, str2)
    char                *str1, **str2;
{
    return strcmp(str1, *str2);
}

static int ptr_strcmp2(str1, str2)
    char                **str1, **str2;
{
    return strcmp(*str1, *str2);
}

void Qddb_MakeIndex(schema, verbose)
    Qddb_Schema		*schema;
    Qddb_Boolean        verbose;
{
    Qddb_Bucket		*HashList;
    Qddb_Entry		ThisEntry = NULL;
    Qddb_Entry          ExcludeEntry = NULL;
    Qddb_KeyNode	*List;
    char		RelationFN[MAXFILENAMELEN];
    char		IndexFN[MAXFILENAMELEN];
    char		KeyFN[MAXFILENAMELEN], HashFN[MAXFILENAMELEN];
    int			IndexFile, DatabaseFile, KeyFile;
    FILE		*KeyFilePtr, *IndexFilePtr;
    FILE		**IndexFilePtrs;
    char		KeyIndexFN[MAXFILENAMELEN], NumericIndexFN[MAXFILENAMELEN];
    int			KeyIndexFile, NumericIndexFile;
    FILE		*KeyIndexFilePtr, *NumericIndexFilePtr;
    FILE		**HashFilePtrs;
    FILE		*HashFilePtr;
    off_t		Start;
    size_t              Length;
    char                start_ch[BUFSIZ], length_ch[BUFSIZ];
    int			i, j, Error;
    size_t              nmemb = 0;
    int                 entrysize, hashlinesize, base;
    int			idxnum = 0;
    size_t		LastHash;
    size_t		starting_bytes;
#if defined(USE_MALLOC_STATS)
#if defined(QDDB_STAB_STATS)
    int			recnum = 0;
#endif
#endif
#if defined(USE_GNU_MALLOC)
    struct mstats 	m;
#endif

    NumberOfKeys = 0;
    NumberOfNumericKeys = 0;
    NumericIndexTop = 0;
    KeyIndex = NULL;
    KeyIndexTop = 0;
    strcpy(RelationFN, schema->RelationName);
    strcpy(IndexFN, RelationFN);
    strcat(IndexFN, "/Index");
    strcpy(HashFN, RelationFN);
    strcat(HashFN, "/HashTable");
    strcpy(KeyFN, RelationFN);
    strcat(KeyFN, "/Database.key");
    strcpy(KeyIndexFN, RelationFN);
    strcat(KeyIndexFN, "/KeyIndex");
    strcpy(NumericIndexFN, RelationFN);
    strcat(NumericIndexFN, "/NumericIndex");
    IndexFile = Open(IndexFN, O_WRONLY|O_CREAT|O_TRUNC, 0666);
    IndexFilePtr = fdopen(IndexFile, "w");
    KeyIndexFile = Open(KeyIndexFN, O_RDWR|O_CREAT|O_TRUNC, 0666);
    KeyIndexFilePtr = fdopen(KeyIndexFile, "w");
    NumericIndexFile = Open(NumericIndexFN, O_RDWR|O_CREAT|O_TRUNC, 0666);
    NumericIndexFilePtr = fdopen(NumericIndexFile, "w");
    if (schema->UseCondensedIndexing == False) {
	base = 10;
	hashlinesize = BASE10_HASHLINESIZE;
    } else {
	base = 62;
	hashlinesize = BASE62_HASHLINESIZE;
    }
    entrysize = hashlinesize / 4;
    if (schema->UseExcludeWords == True) {
	ExcludeEntry = Qddb_ReadExcludeWords(schema, &nmemb);
	qsort(ExcludeEntry, nmemb, sizeof(char *),
	      (int (*) _ANSI_ARGS_((const void *, const void *)))ptr_strcmp2);
    }
    KeyFile   = Open(KeyFN, O_RDONLY, 0);
    KeyFilePtr = fdopen(KeyFile, "r");
    DatabaseFile = schema->database_fd;
    lseek(DatabaseFile, (off_t)0, 0);
    HashList = (Qddb_Bucket *)Malloc(schema->HashSize*sizeof(Qddb_Bucket));
#if defined(USE_GNU_MALLOC)
    m = mstats();
    starting_bytes = m.bytes_used;
#else
#if defined(__linux__)
    starting_bytes = qddb_malloc_bytes_used = (size_t)sbrk(0);
#else
    starting_bytes = 0;
    qddb_malloc_bytes_used = 0;
#endif
#endif
    lseek(schema->hashtable_fd, (off_t)0, 0);
#if HAVE_FTRUNCATE
    ftruncate(schema->hashtable_fd, (off_t)0);
#else
    close(open(HashFN, O_RDWR|O_TRUNC));
#endif
    HashFilePtr = fdopen(schema->hashtable_fd, "w");
    if (HashFilePtr == NULL) {
    	fprintf(stderr, "fdopen of %s/HashTable failed\n", schema->RelationName);
    	exit(1);
    }
    for (i = 0; i < schema->HashSize; i++)
    	HashList[i] = NULL;
    while (fscanf(KeyFilePtr, "%s %s\n", start_ch, length_ch) != EOF) {
	Start = Qddb_BaseToOffset(start_ch, base);
	Length = Qddb_BaseToSize(length_ch, base);
#if defined(USE_MALLOC_STATS)
	if ((++recnum % 100) == 0 && verbose) {
#if defined(QDDB_STAB_STATS)
	    printf("%d Records Processed\n", recnum);
#endif
	    m = mstats();
	    printf("Malloc stats: bytes_total(%u) chunks_used(%u) bytes_used(%u) \
chunks_free(%u) bytes_free(%u)\n", 
		   m.bytes_total, m.chunks_used, m.bytes_used, m.chunks_free, m.bytes_free);
	    fflush(stdout);
	}
#endif
#if defined(MAXIMUM_MEMORY_USAGE)
#if defined(USE_GNU_MALLOC)
	m = mstats();
	if (m.bytes_used > (size_t)(starting_bytes+qddb_maxmem*1024)) {
#else
	if (qddb_malloc_bytes_used > (size_t)(qddb_maxmem*1024 + starting_bytes)) {
#endif
	    FILE		*NewIndexFilePtr, *NewHashFilePtr;
	    char		buf[BUFSIZ];

	    sprintf(buf, "%s-%d", IndexFN, idxnum);
	    if ((NewIndexFilePtr = fopen(buf, "w")) == NULL)
		PANIC(buf);
	    sprintf(buf, "%s-%d", HashFN, idxnum);
	    if ((NewHashFilePtr = fopen(buf, "w")) == NULL)
		PANIC(buf);
	    idxnum++;
	    WriteHashList(schema, NewIndexFilePtr, NewHashFilePtr, 
			  HashList, False, (size_t)0, (size_t)schema->HashSize);
	    PurgeHashList(HashList, schema->HashSize);
#if !defined(USE_GNU_MALLOC)
#if defined(__linux__)
	    qddb_malloc_bytes_used = (size_t)sbrk(0);
#else
	    qddb_malloc_bytes_used = 0;
#endif
#endif
	    fclose(NewIndexFilePtr);
	    fclose(NewHashFilePtr);
#if defined(DIAGNOSTIC) && defined(USE_GNU_MALLOC) && defined(USE_MALLOC_STATS)
	    m = mstats();
	    if (m.bytes_used != starting_bytes && verbose)
		fprintf(stdout, "After purge, starting bytes (%d) != bytes used (%d)\n",
			(int)starting_bytes, (int)m.bytes_used);
#endif
	}
#endif
	Qddb_ReadEntry(DatabaseFile, &ThisEntry, (off_t)Start, (size_t)Length, False);
    	List = Qddb_IndexParse(schema, ThisEntry+1, &Error);
	if (Error) {
	    fprintf(stdout, "Cannot index the following tuple:\n%s\n", ThisEntry[0]);
	    fflush(stdout);
	    List = NULL;
	}
    	while (List != NULL) {
	    int		        Value;
	    Qddb_KeyNode	*next;

	    if (schema->UseExcludeWords == True && nmemb > 0) {
		if (bsearch(List->s, ExcludeEntry, nmemb, sizeof(char *), 
			    (int (*) _ANSI_ARGS_((const void *, const void *)))ptr_strcmp1) != NULL) {
		    next = List->next;
		    if (List->inst != NULL)
			Free(List->inst);
		    Free(List->s);
		    Free(List);
		    List = next;
		    continue;
		}
	    }
	    Value = Qddb_HashValue(schema, List->s);
	    next = List->next;
	    AddToHashList(&HashList[Value], List, (off_t)Start, (size_t)Length); /* Frees up List */
	    List = next;
    	}
#if !defined(USE_GNU_MALLOC)
#if defined(__linux__)
	    qddb_malloc_bytes_used = (size_t)sbrk(0);
#else
	if (qddb_malloc_bytes_used >= (Length * 12)) {
	    qddb_malloc_bytes_used -= Length * 12; /* low estimate, but it works well. */
	} else {
	    qddb_malloc_bytes_used = 0;
	}
#endif
#endif
    }
    if (ThisEntry != NULL)
	Qddb_Free(QDDB_TYPE_ENTRY, ThisEntry);
    if (ExcludeEntry != NULL)
	Qddb_Free(QDDB_TYPE_ENTRY, ExcludeEntry);
    if (verbose)
	fprintf(stdout, "Allocating %d alphanumeric keys\n", (int)NumberOfKeys);
    if (NumberOfKeys == 0)
	return;
    KeyIndex = (Qddb_KeyIndexIn *)Malloc(sizeof(Qddb_KeyIndexIn)*NumberOfKeys);
    if (verbose)
	fprintf(stdout, "Allocating %d numeric keys\n", (int)NumberOfNumericKeys);
    NumericIndex = (Qddb_NumericIndexIn *)Malloc(sizeof(Qddb_NumericIndexIn)*NumberOfNumericKeys);
    if (idxnum == 0) {
	WriteHashList(schema, IndexFilePtr, HashFilePtr, HashList, 
		      True, (size_t)0, (size_t)schema->HashSize);
	PurgeHashList(HashList, schema->HashSize);
	WriteKeyIndex(schema, KeyIndexFilePtr, KeyIndex);
	PurgeKeyIndex(schema, KeyIndex);
	WriteNumericIndex(schema, NumericIndexFilePtr, NumericIndex);
	PurgeNumericIndex(schema, NumericIndex);
    } else {
	FILE		*NewIndexFilePtr, *NewHashFilePtr;
	char		buf[BUFSIZ];

	sprintf(buf, "%s-%d", IndexFN, idxnum);
	if ((NewIndexFilePtr = fopen(buf, "w")) == NULL) {
	    fprintf(stderr, "panic on write:");
	    PANIC(buf);
	}
	sprintf(buf, "%s-%d", HashFN, idxnum);
	if ((NewHashFilePtr = fopen(buf, "w")) == NULL) {
	    fprintf(stderr, "panic on write:");
	    PANIC(buf);
	}
	idxnum++;
	WriteHashList(schema, NewIndexFilePtr, NewHashFilePtr, HashList, 
		      False, (size_t)0, (size_t)schema->HashSize);
	PurgeHashList(HashList, schema->HashSize);
	fclose(NewIndexFilePtr);
	fclose(NewHashFilePtr);
	/* Merge all (Index-*,HashTable-*) into single (Index,HashTable)
	 *
	 * Open all HashTable-* and Index-*
	 * foreach HashEntry across all HashTable-*
	 *     Read the HashEntries
	 *     Read the indices they point to
	 *     Parse indices into KeyNodes
	 *     Add them to the HashList
	 *     if too much memory is used then
	 *         Write the (Index,HashTable), saving Key/Numeric info.
	 *         Purge HashList
	 *     endif
	 * endforeach
	 * Write the (Index,HashTable), saving Key/Numeric info.
	 * Purge HashList
	 * Close all HashTable-* and Index-*
	 * Remove old (Index-*,HashTable-*)
	 */
	HashFilePtrs = (FILE **)Malloc(sizeof(FILE *)*idxnum);
	IndexFilePtrs = (FILE **)Malloc(sizeof(FILE *)*idxnum);
	for (i = 0; i < idxnum; i++) {
	    char		buf[BUFSIZ];
	    
	    sprintf(buf, "%s-%d", HashFN, i);
	    if ((HashFilePtrs[i] = fopen(buf, "r")) == NULL) {
		fprintf(stderr, "panic on read:");
		PANIC(buf);
	    }
	    sprintf(buf, "%s-%d", IndexFN, i);
	    if ((IndexFilePtrs[i] = fopen(buf, "r")) == NULL) {
		fprintf(stderr, "panic on read:");
		PANIC(buf);
	    }
	}
	LastHash = 0;
	for (i = 0; i < schema->HashSize; i++) { /* each HashEntry */
	    for (j = 0; j < idxnum; j++) { /* each (HashTable-*,Index-*) */
		Qddb_Hash		hash;
		Qddb_KeyNode		*key;
		char			buf[BASE10_HASHLINESIZE+1];

		/* 1. Read HashEntry
		 * 2. Read Index pointed to by HashEntry
		 * 3. Parse Index into KeyNodes
		 * 4. Add them to HashList
		 */
		fgets(buf, sizeof(buf), HashFilePtrs[j]);
		buf[hashlinesize] = '\0';
		buf[entrysize-1]  = '\0';
		buf[(entrysize*2)-1]  = '\0';
		buf[(entrysize*3)-1]  = '\0';
		buf[(entrysize*4)-1]  = '\0';
		hash.num = Qddb_BaseToSize(buf, base);
		hash.index = Qddb_BaseToSize(buf+entrysize, base);
		hash.pos = Qddb_BaseToOffset(buf+(2*entrysize), base);
		hash.len = Qddb_BaseToSize(buf+(3*entrysize), base);
		key = HashEntryToKeyNodes(schema, IndexFilePtrs[j], &hash);
		while (key != NULL) {
		    Qddb_KeyNode	*next;
		    Qddb_PosNode	*posnode;

		    posnode = key->pos;
#if defined(DIAGNOSTIC)
		    if (posnode->next != NULL)
			PANIC("MakeIndex: posnode");
#endif
		    key->pos = NULL;
		    next = key->next;
		    AddToHashList(&HashList[i], key, posnode->Start, posnode->Length); /* Frees up key */
		    Free(posnode);
		    key = next;
		}
	    }
#if defined(USE_GNU_MALLOC)
	    m = mstats();
#if defined(USE_MALLOC_STATS)
	    if (i % 100 == 0 && verbose) {
		printf("Hash #%d: bytes_total(%u) chunks_used(%u) bytes_used(%u) \
chunks_free(%u) bytes_free(%u)\n", i,
		   m.bytes_total, m.chunks_used, m.bytes_used, m.chunks_free, m.bytes_free);
		fflush(stdout);
	    }
#endif
	    if (m.bytes_used > (size_t)(starting_bytes+qddb_maxmem*1024)) {
#else
	    if (qddb_malloc_bytes_used > (size_t)(starting_bytes+qddb_maxmem*1024)) {
#endif
		WriteHashList(schema, IndexFilePtr, HashFilePtr, HashList, 
			      True, LastHash, (size_t)i+1);
		LastHash = i+1;
		PurgeHashList(HashList, schema->HashSize);
#if !defined(USE_GNU_MALLOC)
#    if defined(__linux__)
	    qddb_malloc_bytes_used = (size_t)sbrk(0);
#    else
		qddb_malloc_bytes_used = 0;
#    endif
#endif
	    }
	}
	for (i = 0; i < idxnum; i++) {
	    char		buf[BUFSIZ];
	    
	    fclose(HashFilePtrs[i]);
	    fclose(IndexFilePtrs[i]);
	    sprintf(buf, "%s-%d", HashFN, i);
	    if (unlink(buf) == -1) {
		fprintf(stderr, "Warning: failure to unlink %s\n", buf);
	    }
	    sprintf(buf, "%s-%d", IndexFN, i);
	    if (unlink(buf) == -1) {
		fprintf(stderr, "Warning: failure to unlink %s\n", buf);
	    }
	}
	Free(HashFilePtrs);
	Free(IndexFilePtrs);
	WriteHashList(schema, IndexFilePtr, HashFilePtr, HashList, 
		      True, LastHash, (size_t)schema->HashSize);
	PurgeHashList(HashList, schema->HashSize);
	WriteKeyIndex(schema, KeyIndexFilePtr, KeyIndex);
	PurgeKeyIndex(schema, KeyIndex);
	WriteNumericIndex(schema, NumericIndexFilePtr, NumericIndex);
	PurgeNumericIndex(schema, NumericIndex);
    }
    Free(HashList);
}

static void PurgeKeyIndex(schema, KeyIndex)
    Qddb_Schema         *schema;
    Qddb_KeyIndexIn	*KeyIndex;
{
    int                 i;

    for (i = 0; i < KeyIndexTop; i++) {
	if (KeyIndex[i].s != NULL)
	    Free(KeyIndex[i].s);
    }
    Free(KeyIndex);
}

static void PurgeNumericIndex(schema, NumericIndex)
    Qddb_Schema         *schema;
    Qddb_NumericIndexIn *NumericIndex;
{
    Free(NumericIndex);
}

static void PurgePosNode(List)
    Qddb_PosNode	*List;
{
    Qddb_PosNode	*newlist;

    while (List != NULL) {
	newlist = List->next;
	if (List->inst != NULL)
	    Free(List->inst);
	Free(List);
	List = newlist;
    }
}

static void PurgeHashBucket(List) 
    Qddb_KeyNode	*List;
{
    Qddb_KeyNode	*newlist;

    while (List != NULL) {
	newlist = List->next;
	PurgePosNode(List->pos);
	if (List->s != NULL)
	    Free(List->s);
/* notdef */
	if (List->inst != NULL)
	    Free(List->inst);
	Free(List);
	List = newlist;
    }
}

static void PurgeHashList(List, Size)
    Qddb_Bucket		*List;
    size_t		Size;
{
    int			i;

    for (i = 0; i < Size; i++) {
	PurgeHashBucket(List[i]);
	List[i] = NULL;
    }
}

Qddb_KeyNode *Qddb_IndexParse(schema, ThisEntry, ErrorPtr)
    Qddb_Schema		*schema;
    Qddb_Entry		ThisEntry;
    int			*ErrorPtr;
{
    Qddb_KeyNode	*ptr, *List;
    int			len;
    char		*inst, *token, *nextspot, *ch, *full_attr, *tmp;
    size_t		attr;

    List = NULL;
    *ErrorPtr = 0;
    while (*ThisEntry != NULL) {
	token = strtok(*ThisEntry, " "); /* Skip id */
#if defined(DIAGNOSTIC)
	if (token == NULL) {
	    PANIC("IndexParse: token is NULL\n");
	}
#endif
	nextspot = *ThisEntry + strlen(token) + 1;
	if (schema->UseReducedAttributeIdentifiers == True) {
	    inst = Malloc(strlen(token)+1);
	    strcpy(inst, token+1);
	    attr = Qddb_ReducedAttrToIndex(inst);
	    full_attr = Malloc(strlen(schema->ReducedAttributeTable[attr])+2);
	    full_attr[0] = '%', full_attr[1] = '\0';
	    strcat(full_attr, schema->ReducedAttributeTable[attr]);
	    if (Qddb_TranslateExternalID(schema, full_attr, &attr, &tmp) != 0) {
		fprintf(stderr, "Warning: attribute %s malformed, skipping attr %d..\n", 
			schema->ReducedAttributeTable[attr], (int)attr);
		Free(inst);
		Free(full_attr);
		ThisEntry++;
		continue;
	    }
	    Free(full_attr);
	    Free(tmp);
	} else if (Qddb_TranslateExternalID(schema, token, &attr, &inst) != 0) {
	    fprintf(stderr, "Warning: attribute %s malformed, skipping..\n", token);
	    ThisEntry++;
	    continue;
	}
	if (schema->Entries[attr].Exclude == True) {
	    ThisEntry++;
	    Free(inst);
	    continue;
	}
	while ((token=strtok(nextspot, schema->Entries[attr].Separators)) != NULL) {
	    nextspot = NULL;
	    len = 0;
	    ch = token;
	    while (*ch) {
		if (_isupper(*ch))
		    *ch = _tolower(*ch);
		ch++;
		len++;
	    }
	    ptr = List;
	    List = (Qddb_KeyNode *)Malloc(sizeof(Qddb_KeyNode));
	    List->next = ptr;
	    List->s = Malloc((size_t)len+1);
	    strcpy(List->s, token);
	    Qddb_StripBackslashes(List->s);
	    List->pos = NULL;
	    List->len = len;
	    List->attr = attr;
	    List->inst = (char *)Malloc(strlen(inst)+1);
	    strcpy(List->inst, inst);
	}
	ThisEntry++;
	Free(inst);
    }
    return List;
}

static void AddToHashList(List, Node, Start, Length)
    Qddb_Bucket		*List;
    Qddb_KeyNode	*Node;
    off_t		Start;
    size_t		Length;
{
    Qddb_KeyNode	*ptr;

    if (*List == NULL) {
	NumberOfKeys++;
	if (Qddb_IsNumericString(Node->s))
	    NumberOfNumericKeys++;
	*List = (Qddb_KeyNode *)Malloc(sizeof(Qddb_KeyNode));
	**List = *Node;
	(*List)->inst = NULL;
	(*List)->next = NULL;
	(*List)->pos  = (Qddb_PosNode *)Malloc(sizeof(Qddb_PosNode));
	(*List)->pos->Start = Start;
	(*List)->pos->Length = Length;
	(*List)->pos->next = NULL;
	(*List)->pos->attr = Node->attr;
	(*List)->pos->inst = Node->inst;
	Free(Node);
	return;
    }
    ptr = *List;
    while (ptr != NULL) {
	Qddb_PosNode	*PosPtr;
	
	if (ptr->len == Node->len && strcmp(ptr->s, Node->s) == 0) {
	    PosPtr = (Qddb_PosNode *)Malloc(sizeof(Qddb_PosNode));
	    PosPtr->Start = Start;
	    PosPtr->Length = Length;
	    PosPtr->attr = Node->attr;
	    PosPtr->inst = Node->inst;
	    PosPtr->next = ptr->pos;
	    ptr->pos = PosPtr;
	    /* Free up the word and node. */
	    Free(Node->s);
	    Free(Node);
	    return;
	}
	ptr = ptr->next;
    }
    NumberOfKeys++;
    if (Qddb_IsNumericString(Node->s))
	NumberOfNumericKeys++;
    ptr = *List;
    *List = (Qddb_KeyNode *)Malloc(sizeof(Qddb_KeyNode));
    **List = *Node;
    (*List)->inst = NULL;
    (*List)->next = ptr;
    (*List)->pos  = (Qddb_PosNode *)Malloc(sizeof(Qddb_PosNode));
    (*List)->pos->Start = Start;
    (*List)->pos->Length = Length;
    (*List)->pos->attr = Node->attr;
    (*List)->pos->inst = Node->inst;
    (*List)->pos->next = NULL;
    Free(Node);
}

static void WriteHashList(schema, IndexFilePtr, HashFilePtr, HashList, PassTwo, Begin, End)
    Qddb_Schema		*schema;
    FILE		*IndexFilePtr;
    FILE		*HashFilePtr;
    Qddb_Bucket		*HashList;
    Qddb_Boolean	PassTwo; /* True iff on the final pass */
    size_t		Begin, End;
{
    Qddb_KeyNode	*Key;
    Qddb_Boolean	IsALine = False;
    off_t		Start;
    size_t              Length;
    size_t		i, NumberHashed; 
    int			file_number;
    int                 base, entrysize, hashlinesize;
    char                num[BASE10_HASHLINESIZE], index[BASE10_HASHLINESIZE];
    char                len[BASE10_HASHLINESIZE], pos[BASE10_HASHLINESIZE];
    off_t		WhereAmI, WhereAmINow;
    
    if (schema->UseCondensedIndexing == False) {
	base = 10;
	hashlinesize = BASE10_HASHLINESIZE;
    } else {
	base = 62;
	hashlinesize = BASE62_HASHLINESIZE;
    }
    entrysize = (hashlinesize / 4) - 1;
    for (i = Begin; i < End; i++) {
	Key = HashList[i];
	file_number = (int)((unsigned int)fileno(IndexFilePtr));
	Start = lseek(file_number, (off_t)0, 1);
	if (Start == -1)
	    PANIC("Cannot seek to Start");
	NumberHashed = 0;
	while (Key != NULL) {
	    Qddb_PosNode		*ptr;
	    char		*outchar;
	    
	    NumberHashed++;
	    IsALine = True;
	    /* Caution: This fflush(3) is needed for the lseek(2) to
	     *	find where we are now.
	     */
	    fflush(IndexFilePtr);
	    /* For KeyIndex:
	     * 	Need to keep each key and its locator (start,length)
	     *	to the Index file.
	     */
	    file_number = (int)((unsigned int)fileno(IndexFilePtr));
	    WhereAmI = lseek(file_number, (off_t)0, 1);
	    if (Start == -1)
		PANIC("Lib/LibQddb/Index.c:WriteHashList(): lseek failed");
	    outchar = Key->s;
	    while (*outchar != '\0') {
		if (*outchar == ' ') {
		    putc('\\', IndexFilePtr);
		    putc(' ', IndexFilePtr);
		} else if (*outchar == '\\') {
		    putc('\\', IndexFilePtr);
		    putc('\\', IndexFilePtr);
		} else {
		    putc(*outchar, IndexFilePtr);
		}
		outchar++;
	    }
	    putc(' ', IndexFilePtr);
	    ptr = Key->pos;
	    while (ptr != NULL) {
		char                  *start, *length;

		if (schema->UseReducedAttributeIdentifiers == True) {
		    if (schema->UseCondensedIndexing == False) {
			start = Qddb_OffsetToBase(ptr->Start, 10);
			length = Qddb_SizeToBase(ptr->Length, 10);
			fprintf(IndexFilePtr, "%s %s %s ", start, length, ptr->inst);
		    } else {
			start = Qddb_OffsetToBase(ptr->Start, 62);
			length = Qddb_SizeToBase(ptr->Length, 62);
			fprintf(IndexFilePtr, "%s %s %s ", start, length, ptr->inst);
		    }
		} else {
		    if (schema->UseCondensedIndexing == False) {
			start = Qddb_OffsetToBase(ptr->Start, 10);
			length = Qddb_SizeToBase(ptr->Length, 10);
			fprintf(IndexFilePtr, "%s %s %d %s ", start, length, (int)ptr->attr, ptr->inst);
		    } else {
			start = Qddb_OffsetToBase(ptr->Start, 62);
			length = Qddb_SizeToBase(ptr->Length, 62);
			fprintf(IndexFilePtr, "%s %s %d %s ", start, length, (int)ptr->attr, ptr->inst);
		    }
		}
		ptr = ptr->next;
	    }
	    /* Caution: This fflush(3) is needed for the lseek(2) to
	     *	find where we are now.
	     */
	    fflush(IndexFilePtr);
	    file_number = (int)((unsigned int)fileno(IndexFilePtr));
	    WhereAmINow = lseek(file_number, (off_t)0, 1);
	    if (PassTwo == True) {
		/* For KeyIndex:
		 *	Need the length now.... WhereAmINow-WhereAmI is a good
		 *	thing.
		 */
		KeyIndex[KeyIndexTop].s = Malloc((size_t)Key->len+1);
		strcpy(KeyIndex[KeyIndexTop].s, Key->s);
		KeyIndex[KeyIndexTop].Start = WhereAmI;
		KeyIndex[KeyIndexTop].Length = WhereAmINow - WhereAmI;
		KeyIndexTop++;
		/* For NumericIndex:
		 *	If the key is numeric, convert to floating point and
		 *	save it away for NumericIndex.
		 */
		if (Qddb_IsNumericString(Key->s) == True) {
		    NumericIndex[NumericIndexTop].s = Qddb_GetNum(Key->s);
		    NumericIndex[NumericIndexTop].Start = WhereAmI;
		    NumericIndex[NumericIndexTop].Length = WhereAmINow - WhereAmI;
		    NumericIndexTop++;
		}
	    }
	    fprintf(IndexFilePtr, "\n");
	    fflush(IndexFilePtr);
	    Key = Key->next;
	}
	if (IsALine == True) {
	    file_number = (int)((unsigned int)fileno(IndexFilePtr));
	    Length = lseek(file_number, (off_t)0, 1); 
	    if (Length == -1)
		PANIC("Cannot seek to Length");
	    Length -= Start;
	    if (Length != 0) {
		strcpy(num, Qddb_SizeToBase(NumberHashed, base));
		strcpy(index, Qddb_SizeToBase(i, base));
		strcpy(pos, Qddb_OffsetToBase(Start, base));
		strcpy(len, Qddb_SizeToBase(Length, base));
		if (PassTwo == False || schema->UseCachedHashing == True) {
		    fprintf(HashFilePtr, "%*s %*s %*s %*s\n", entrysize, num, entrysize, index, 
			    entrysize, pos, entrysize, len);
		} else {
		    fprintf(HashFilePtr, "%s %s %s %s\n", num, index, pos, len);
		}
	    }
	    IsALine = False;
	} else if (PassTwo == False || schema->UseCachedHashing == True) {
	    fprintf(HashFilePtr, "%*d %*s %*d %*d\n", entrysize, 0, 
		    entrysize, Qddb_SizeToBase(i, base), entrysize, 0, entrysize, 0);
	}
    }
}

static void WriteKeyIndex(schema, KeyIndexFilePtr, KeyIndex)
    Qddb_Schema         *schema;
    FILE		*KeyIndexFilePtr;
    Qddb_KeyIndexIn	*KeyIndex;
{
    int			i;
    char		*outchar, *start, *length;

    /* Sort the array by name (method: recursive quicksort).
     */
    if (KeyIndexTop > 1)
	SortKeyIndex(1, KeyIndexTop);
    fprintf(KeyIndexFilePtr, "\n");
    for (i = 0; i < KeyIndexTop; i++) {
	outchar = KeyIndex[i].s;
	while (*outchar != '\0') {
	    if (*outchar == ' ') {
		putc('\\', KeyIndexFilePtr);
		putc(' ', KeyIndexFilePtr);
	    } else if (*outchar == '\\') {
		putc('\\', KeyIndexFilePtr);
		putc('\\', KeyIndexFilePtr);
	    } else {
		putc(*outchar, KeyIndexFilePtr);
	    }
	    outchar++;
	}
	putc(' ', KeyIndexFilePtr);
	if (schema->UseCondensedIndexing == False) {
	    start = Qddb_OffsetToBase(KeyIndex[i].Start, 10);
	    length = Qddb_SizeToBase(KeyIndex[i].Length, 10);
	} else {
	    start = Qddb_OffsetToBase(KeyIndex[i].Start, 62);
	    length = Qddb_SizeToBase(KeyIndex[i].Length, 62);
	}
	fprintf(KeyIndexFilePtr, "%s %s\n", start, length);
    }
}

static void SortKeyIndex(l, r)
    size_t		l, r;
{
    Qddb_KeyIndexIn	x, w;
    size_t		i, j;

    i = l; j = r;
    x = KeyIndex[((l+r)/2)-1];
    do {
	while (strcmp(KeyIndex[i-1].s, x.s) < 0)
	    i++;
	while (strcmp(x.s, KeyIndex[j-1].s) < 0)
	    j--;
	if (i <= j) {
	    w = KeyIndex[i-1];
	    KeyIndex[i-1] = KeyIndex[j-1];
	    KeyIndex[j-1] = w;
	    i++, j--;
	}
    } while (j >= i);
    if (l < j)
	SortKeyIndex(l, j);
    if (i < r)
	SortKeyIndex(i, r);
}


static void WriteNumericIndex(schema, NumericIndexFilePtr, NumericIndex)
    Qddb_Schema         *schema;
    FILE		*NumericIndexFilePtr;
    Qddb_NumericIndexIn	*NumericIndex;
{
    int			i;
    char                *start, *length;

    /* Sort the array by numeric key (method: recursive quicksort).
     */
    if (NumericIndexTop > 1)
	SortNumericIndex(1, NumericIndexTop);
    fprintf(NumericIndexFilePtr, "\n");
    for (i = 0; i < NumericIndexTop; i++) {
	if (schema->UseCondensedIndexing == False) {
	    start = Qddb_OffsetToBase(NumericIndex[i].Start, 10);
	    length = Qddb_SizeToBase(NumericIndex[i].Length, 10);
	} else {
	    start = Qddb_OffsetToBase(NumericIndex[i].Start, 62);
	    length = Qddb_SizeToBase(NumericIndex[i].Length, 62);
	}
	fprintf(NumericIndexFilePtr, "%.17g %s %s\n", NumericIndex[i].s, start, length);
    }
}

static void SortNumericIndex(l, r)
    size_t		l, r;
{
    size_t		i, j;
    Qddb_NumericIndexIn	x, w;

    i = l; j = r;
    x = NumericIndex[((l+r)/2)-1];
    do {
	while (NumericIndex[i-1].s < x.s)
	    i++;
	while (x.s < NumericIndex[j-1].s)
	    j--;
	if (i <= j) {
	    w = NumericIndex[i-1];
	    NumericIndex[i-1] = NumericIndex[j-1];
	    NumericIndex[j-1] = w;
	    i++, j--;
	}
    } while (j >= i);
    if (l < j)
	SortNumericIndex(l, j);
    if (i < r)
	SortNumericIndex(i, r);
}


static Qddb_KeyNode *ConvertKeyListToKeyNode(string, keylist)
    char		*string;
    Qddb_KeyList	*keylist;
{
    Qddb_KeyList	*tmp;
    Qddb_KeyNode	*retval = NULL, *newnode;
    size_t		len;

    len = strlen(string);
    while (keylist != NULL) {
	newnode = (Qddb_KeyNode *)Malloc(sizeof(Qddb_KeyNode));
	newnode->pos = (Qddb_PosNode *)Malloc(sizeof(Qddb_PosNode));
	newnode->s = Malloc(len+1);
	strcpy(newnode->s, string);
	newnode->len = len;
	newnode->attr = newnode->pos->attr = keylist->Attribute;
	newnode->inst = newnode->pos->inst = keylist->Instance;
	newnode->pos->Start = keylist->Start;
	newnode->pos->Length = keylist->Length;
	newnode->pos->next = NULL;
	newnode->next = retval;
	retval = newnode;
	tmp = keylist;
	keylist = keylist->next;
	Free(tmp);
    }
    return retval;
}

static Qddb_KeyNode *HashEntryToKeyNodes(schema, IndexFilePtr, HashPtr)
    Qddb_Schema		*schema;
    FILE		*IndexFilePtr;
    Qddb_Hash		*HashPtr;
{
    Qddb_KeyList	*keylist;
    Qddb_KeyNode	*keynode, *retval = NULL, *tmp;
    char		**KeyEntry, *string, *ch, *Buffer;
    int			i, EntryNumber, KeyFile;
    size_t		NumberRead;

    if (HashPtr->len == 0)
	return NULL;
    KeyFile = (int)((unsigned int)fileno(IndexFilePtr));
    KeyEntry = (char **)Malloc(sizeof(char *)*(HashPtr->num+1));
    Buffer = Malloc(HashPtr->len+1);
    Buffer[HashPtr->len] = '\0';
    lseek(KeyFile, HashPtr->pos, 0);
    Read(KeyFile, Buffer, HashPtr->len);
    KeyEntry[0] = Buffer;
    EntryNumber = 1;
    /* Parse the hash bucket into newline terminated lines.
     */
    for (i = 0; i < HashPtr->len; i++) {
        if (Buffer[i] == '\n') {
            Buffer[i] = '\0';
            if (i == HashPtr->len-1)
                break;
            KeyEntry[EntryNumber++] = Buffer+i+1;
        }
    }
#if defined(DIAGNOSTIC)
    if (HashPtr->num != EntryNumber)
	fprintf(stderr, "HashEntry number (%d) != number found (%d)\n", (int)HashPtr->num,
		(int)EntryNumber);
#endif
    for (i = 0; i < EntryNumber; i++) {
	size_t			key_len;

	ch = Qddb_FindKey(KeyEntry[i], &key_len);
#if defined(DIAGNOSTIC)
	if (ch == NULL) /* EEEEK */
	    PANIC("EEEK");
#endif
	*ch = '\0';
	string = Qddb_StripKey(KeyEntry[i], key_len);
	*ch = ' ';
	keylist = Qddb_ParseKey(schema, KeyEntry[i], &NumberRead, True, -1);
	keynode = ConvertKeyListToKeyNode(string, keylist);
	Free(string);
	if (retval == NULL) {
	    retval = keynode;
	} else {
	    if (keynode != NULL) {
		tmp = keynode;
		while (tmp->next != NULL)
		    tmp = tmp->next;
		tmp->next = retval;
		retval = keynode;
	    }
	}
    }
    Free(Buffer);
    Free(KeyEntry);
    return retval;
}
