
/* $RCSfile: Rows.c,v $ $Revision: 2.1 $ $Date: 2003/03/23 17:16:54 $
 *
 * Copyright (C) 1993-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "Qddb.h"

/* Qddb_RowListCompareRows:
 *
 * This routine takes advantage of the fact that all rows are the same
 * length and they are sorted.
 */
Qddb_Boolean Qddb_RowListCompareRows(TheEntry, Row1, Row2, Nums, Top)
    Qddb_InCoreEntry	*TheEntry;
    Qddb_RowList	*Row1, *Row2;
    size_t		*Nums;
    size_t		Top;
{
    Qddb_RowList	*r1, *r2;
    Qddb_Boolean	Duplicate;
    int			i;

    /* FIXME: This routine needs to be cleaned up for efficiency */
    Duplicate = True;
    for (i = 0; i < Top && Duplicate == True; i++) {
	r1 = Row1;
	while (r1 != NULL) {
	    if (Nums[i] == TheEntry[r1->Attribute].AttributeNumber) {
		r2 = Row2;
		while (r2 != NULL) {
		    if (Nums[i] == TheEntry[r2->Attribute].AttributeNumber) {
			if (r1->Attribute != r2->Attribute) {
			    return False;
			}
			break;
		    }
		    r2 = r2->next;
		}
		if (r2 != NULL)
		    break;
	    }
	    r1 = r1->next;
	}
    }
    return Duplicate;
}

int Qddb_RowListShouldSkipRow(TheEntry, Row, Nums, Top)
    Qddb_InCoreEntry	*TheEntry;
    Qddb_RowList	*Row;
    size_t		*Nums;
    size_t		Top;
{
    int			*Match, i;

    Match = (int *)Malloc(sizeof(int)*Top);
    for (i = 0; i < Top; i++)
	Match[i] = -1;
    while (Row != NULL) {
	Qddb_Boolean	mark;

	mark = TheEntry[Row->Attribute].Marked;
	if (mark != False) {
	    for (i = 0; i < Top; i++) {
		if (Nums[i] == TheEntry[Row->Attribute].AttributeNumber) {
		    Match[i] = Nums[i];
		    break;
		}
	    }
	}
	Row = Row->next;
    }
    for (i = 0; i < Top; i++) {
	if ((int)Nums[i] != Match[i]) {
#if defined(DEBUG)
	    fprintf(stderr, "skipping row\n");
#endif
	    Free(Match);
	    return True;
	}
    }
    Free(Match);
    return False;
}


void Qddb_RowListDeleteDuplicateRows(TheEntry, Rows, Nums, Top)
    Qddb_InCoreEntry	*TheEntry;
    Qddb_RowList	*Rows;
    size_t		*Nums;
    size_t		Top;
{
    Qddb_RowList	*r1, *rp1, *back;

    while (Rows != NULL && Rows->down != NULL) {
	back = Rows;
	rp1 = Rows->down;
	while (rp1 != NULL) {
	    if (Qddb_RowListCompareRows(TheEntry, Rows, rp1, Nums, Top) == True) {
		/* Same row, keep the first one */
		r1 = rp1;
		back->down = rp1->down;
		rp1 = r1->down;
		r1->down = NULL;
		Qddb_Free(QDDB_TYPE_ROWLIST, r1);
	    } else {
		rp1 = rp1->down;
		back = back->down;
	    }
	}
	Rows = Rows->down;
    }
}

/* WARNING: RowList processing depends on the list being sorted.  Disaster may
 * occur otherwise.
 */

Qddb_RowList *Qddb_RowListProcessEntry(schema, ThisEntry)
    Qddb_Schema		*schema;
    Qddb_InCoreEntry	*ThisEntry;
{
    Qddb_InCoreEntry	*OldThisEntry;
    Qddb_RowList	*List = NULL;
    Qddb_SchemaNode	*ThisNode;
    size_t		i, j;
    size_t		Attrs[2];


    OldThisEntry = ThisEntry;
    i = 0;
    while (OldThisEntry->SequenceNumber != 0) {
	OldThisEntry->Index = i++;
	OldThisEntry++;
    }
    for (i = 1; i <= schema->NumberOfAttributes; i++) {
	ThisNode = schema->Entries+i;
	if (ThisEntry->SequenceNumber == 0)
	    break;
	if (ThisNode->Level == 1) {
	    Attrs[1] = ThisNode->Number;
	    j = 0;
	    OldThisEntry = ThisEntry;
	    if (schema->Entries[ThisEntry->AttributeNumber].Level == 1) {
		while (ThisEntry->SequenceNumber != 0 &&
		       ThisEntry->AttributeNumber == i)
		    ThisEntry++, j++;		
	    } else {
		while (ThisEntry->SequenceNumber != 0 &&
		       schema->Entries[ThisEntry->AttributeNumber].AncestorNumber[1] == 
		       ThisNode->Number)
		    ThisEntry++, j++;
	    }
	    if (j == 0)
		continue;
	    if (List == NULL) {
		List = Qddb_RowListProcessAttr(schema, OldThisEntry, j, Attrs, 1);
	    } else {
		List = Qddb_RowListCartesianProduct(List,
			    Qddb_RowListProcessAttr(schema, OldThisEntry, j, Attrs, 1));
	    }
	}
    }
    Qddb_RowListPruneUntrueRows(&List);
    return List;
}

Qddb_RowList *Qddb_RowListProcessAttr(schema, ThisEntry, EntryLength, Attrs, AttrsLength)
    Qddb_Schema		*schema;
    Qddb_InCoreEntry	*ThisEntry;
    size_t		EntryLength, Attrs[], AttrsLength;
{
    Qddb_InCoreEntry	*OldThisEntry;
    Qddb_RowList	*List = NULL;
    size_t		Inst, i, j;

    /* For each instance in this attribute */
    for (i = 0; i < EntryLength; i+=j) {
	j = 0;
	OldThisEntry = ThisEntry;
	Inst = ThisEntry->Instance[AttrsLength-1];
	while (i+j < EntryLength && ThisEntry->Instance[AttrsLength-1] == Inst)
	    ThisEntry++, j++;
	if (List == NULL)
	    List = Qddb_RowListProcessInst(schema, OldThisEntry, j, Attrs, AttrsLength);
	else
	    List = Qddb_RowListUnion(List, 
			Qddb_RowListProcessInst(schema, OldThisEntry, j, Attrs, AttrsLength));
    }
    return List;
}

Qddb_RowList *Qddb_RowListProcessInst(schema, ThisEntry, EntryLength, Attrs, AttrsLength)
    Qddb_Schema		*schema;
    Qddb_InCoreEntry	*ThisEntry;
    size_t		EntryLength, Attrs[], AttrsLength;
{
    Qddb_InCoreEntry	*OldThisEntry;
    Qddb_RowList 	*List;
    size_t		Inst, i, j;
    size_t		*NewAttrs, NewAttrsLength, First;

    if (schema->Entries[ThisEntry->AttributeNumber].Level == AttrsLength) { /* at a leaf */
	if (ThisEntry->Marked != False) {
	    List = (Qddb_RowList *)Malloc(sizeof(Qddb_RowList));
	    List->next = List->down = NULL;
	    if (ThisEntry->Marked == True) {
		List->Flags = QDDB_ROWLIST_CONTAINS_TRUE;
	    } else
		List->Flags = 0;
	    List->Attribute = ThisEntry->Index;
	} else {
	    List = NULL;
	}
	return List;
    }
    List = NULL;
    First = True;
    NewAttrsLength = AttrsLength+1;
    NewAttrs = (size_t *)alloca(sizeof(size_t)*(NewAttrsLength+1));
    for (i = 1; i <= AttrsLength; i++)
	NewAttrs[i] = Attrs[i];
    Inst = ThisEntry->Instance[AttrsLength-1];
    /* For each sub-attribute in this instance */
    for (i = 0; i < EntryLength; i+=j) {
	j = 0;
	OldThisEntry = ThisEntry;
	if (NewAttrsLength == schema->Entries[ThisEntry->AttributeNumber].Level) {
	    NewAttrs[NewAttrsLength] = 
		schema->Entries[ThisEntry->AttributeNumber].Number;
	    while (i+j < EntryLength && 
		   schema->Entries[ThisEntry->AttributeNumber].Level == NewAttrsLength &&
		   schema->Entries[ThisEntry->AttributeNumber].Number == NewAttrs[NewAttrsLength])
		ThisEntry++, j++;
	} else {
	    NewAttrs[NewAttrsLength] = 
		schema->Entries[ThisEntry->AttributeNumber].AncestorNumber[NewAttrsLength];
	    while (i+j < EntryLength &&
		   schema->Entries[ThisEntry->AttributeNumber].AncestorNumber[NewAttrsLength] == 
		   NewAttrs[NewAttrsLength])
		ThisEntry++, j++;
	}
	if (First == True)
	    List = Qddb_RowListProcessAttr(schema, OldThisEntry, j, NewAttrs, NewAttrsLength);
	else
	    List = Qddb_RowListCartesianProduct(List,
			Qddb_RowListProcessAttr(schema, OldThisEntry, j, NewAttrs, NewAttrsLength));
	First = False;
    }
#if defined(ROWDEBUG)
    PrintRowDebug(List);
#endif 
    return List;
}

#if defined(ROWDEBUG)
void PrintRowDebug(List)
    Qddb_RowList	*List;
{
    Qddb_RowList	*tmp;

    while (List != NULL) {
	tmp = List;
	printf("{ ");
	while (tmp != NULL) {
	    printf("%d ", tmp->Attribute);
	    tmp = tmp->next;
	}
	printf("}\n");
	List = List->down;
    }
}
#endif

Qddb_RowList *Qddb_RowListCopy(List)
    Qddb_RowList	*List;
{
    Qddb_RowList	*NewList, *tmp;

    if (List == NULL)
	return NULL;
    tmp = NewList = NULL;
    while (List != NULL) {
	if (NewList == NULL)
	    NewList = tmp = (Qddb_RowList *)Malloc(sizeof(Qddb_RowList));
	else {
	    tmp->next = (Qddb_RowList *)Malloc(sizeof(Qddb_RowList));
	    tmp = tmp->next;
	}
	tmp->Attribute = List->Attribute;
	tmp->Flags = List->Flags;
	tmp->down = tmp->next = NULL;
	List = List->next;
    }
    return NewList;
}

Qddb_RowList *Qddb_RowListCartesianProduct(List1, List2)
    Qddb_RowList	*List1, *List2;
{
    Qddb_RowList    	*tmp1, *tmp2, *NewList = NULL;
    Qddb_RowList	*OrigList1 = List1;

    if (List1 == NULL || List2 == NULL) {
	return NULL;
    }
    tmp1 = NULL;
    while (List1 != NULL) {
	tmp2 = List2;
	while (tmp2 != NULL) {
	    Qddb_RowList	*tmplist;

	    if (NewList == NULL) {
		NewList = tmp1 = Qddb_RowListCopy(List1);
	    } else {
		tmp1->down = Qddb_RowListCopy(List1);
		tmp1 = tmp1->down;
		tmp1->down = NULL;
	    }
	    tmplist = tmp1;
	    while (tmplist->next != NULL)
		tmplist = tmplist->next;
	    tmplist->next = Qddb_RowListCopy(tmp2);
	    tmp2 = tmp2->down;
	}
	List1 = List1->down;
    }
    Qddb_Free(QDDB_TYPE_ROWLIST, OrigList1);
    Qddb_Free(QDDB_TYPE_ROWLIST, List2);
    return NewList;
}

Qddb_RowList *Qddb_RowListUnion(List1, List2)
    Qddb_RowList	*List1, *List2;
{
    Qddb_RowList	*tmp;

    if (List1 == NULL)
	return List2;
    if (List2 == NULL)
	return List1;
    tmp = List1;
    while (tmp->down != NULL)
	tmp = tmp->down;
    tmp->down = List2;
    return List1;
}

void Qddb_RowListPruneUntrueRows(List)
    Qddb_RowList	**List;
{
    Qddb_RowList	**list, *tmp;
    Qddb_Boolean	true_flag;

    if (List == NULL || *List == NULL)
	return;
    list = List;
    while (*list != NULL) {
	true_flag = False;
	tmp = *list;
	while (tmp != NULL) {
	    if ((tmp->Flags & QDDB_ROWLIST_CONTAINS_TRUE) != 0) {
		true_flag = True;
		break;
	    }
	    tmp = tmp->next;
	}
	if (true_flag == False) {
	    tmp = *list;
	    *list = tmp->down;
	    tmp->down = NULL;
	    Qddb_Free(QDDB_TYPE_ROWLIST, tmp);
	} else {
	    list = &(*list)->down;
	}
    }
}


#if defined(USE_TCL)
static void QueryPrintRow _ANSI_ARGS_((Tcl_Interp *, Qddb_Schema *, Qddb_KeyList *, Qddb_InCoreEntry *, \
				       Qddb_RowList *, char *, size_t *, size_t *, size_t, \
				       size_t *, size_t, Qddb_Boolean, Qddb_Boolean));
static char *PrintRowInstances _ANSI_ARGS_((Qddb_Schema *, Qddb_InCoreEntry *, Qddb_RowList *));
static void QueryPrintRowAsElements _ANSI_ARGS_((Tcl_Interp *, Qddb_Schema *, Qddb_KeyList *, \
						 Qddb_InCoreEntry *, \
						 Qddb_RowList *, char *, size_t *, size_t *, size_t, \
						 size_t *, size_t, Qddb_Boolean, Qddb_Boolean));
#else
static void QueryPrintRow _ANSI_ARGS_((Qddb_Schema *, Qddb_KeyList *, Qddb_InCoreEntry *, \
				       Qddb_RowList *, char *, \
				       size_t *, size_t *, size_t, size_t *, size_t, \
				       Qddb_Boolean, Qddb_Boolean));
#endif

#if defined(USE_TCL)
int Qddb_RowListQuery(interp, schema, list, SepString, 
		      AttrNames, AttrWidth, AttrTop, AttrsUsed, AttrsUsedTop, 
		      Any, TCLprint, DontPrintRows, print_tcl_elements, suppress_printing)
    Tcl_Interp		*interp;
    Qddb_Schema		*schema;
    Qddb_KeyList	*list;
    char		*SepString;
    /* printing attributes */
    char		*AttrNames[];
    size_t		*AttrWidth, AttrTop;
    /* attributes used to generate 'list' */
    char		*AttrsUsed[];
    size_t		AttrsUsedTop;
    /* was this an 'Any' search? */
    Qddb_Boolean	Any;
    Qddb_Boolean	TCLprint;
    Qddb_Boolean	DontPrintRows;
    Qddb_Boolean	print_tcl_elements;
    Qddb_Boolean	suppress_printing;
#else
int Qddb_RowListQuery(schema, list, SepString, 
		      AttrNames, AttrWidth, AttrTop, AttrsUsed, AttrsUsedTop, 
		      Any, TCLprint)
    Qddb_Schema		*schema;
    Qddb_KeyList		*list;
    char		*SepString;
    /* printing attributes */
    char		*AttrNames[];
    size_t		*AttrWidth,AttrTop;
    /* attributes used to generate 'list' */
    char		*AttrsUsed[];
    size_t		AttrsUsedTop;
    /* was this an 'Any' search? */
    Qddb_Boolean	Any;
    Qddb_Boolean	TCLprint;
#endif
{
    Qddb_Entry		ThisEntry = NULL;
    Qddb_InCoreEntry	*CoreEntry = NULL;
    Qddb_RowList	*Row;
    Qddb_KeyList	**SplitList, **OrigSplitList, *tmplist, Key;
    char		*RelationFN = schema->RelationName;
    int			DBFile, tmpattr;
    size_t		*AttrNums, i;
    size_t		*AttrNumsUsed;

    AttrNums = (size_t *)Malloc(sizeof(size_t)*AttrTop);
    AttrNumsUsed = (size_t *)Malloc(sizeof(size_t)*AttrsUsedTop);
    if (list == NULL)
	return 0;
    DBFile = schema->database_fd;
    for (i = 0; i < AttrTop; i++) {
	tmpattr = Qddb_ConvertAttributeNameToNumber(schema, AttrNames[i]);
	if (tmpattr == -1) {
	    Free(AttrNums);
	    Free(AttrNumsUsed);
	    return -1;
	}
	AttrNums[i] = (size_t)tmpattr;
    }
    for (i = 0; i < AttrsUsedTop; i++) {
	tmpattr = Qddb_ConvertAttributeNameToNumber(schema, AttrsUsed[i]);
	if (tmpattr == -1) {
	    Free(AttrNums);
	    Free(AttrNumsUsed);
	    return -1;
	}
	AttrNumsUsed[i] = (size_t)tmpattr;
    }
    /* Here we copy the keylist into a split list; each element of the
     * split list is deleted by MARKFROMINCORE
     */
    (void)Qddb_KeyListProcess(schema, list, &SplitList, QDDB_KEYLIST_PROC_SPLIT, QDDB_KEYLIST_FLAG_COPY);
    OrigSplitList = SplitList;
    while (*SplitList != NULL) {
	tmplist = *SplitList;
#if defined(DEBUG)
	printf("*SplitList == (%d,%d,%d,%s)\n", tmplist->Start, tmplist->Length, tmplist->Attribute,
	       tmplist->Instance);
#endif
	ThisEntry = NULL;
	if (Qddb_ReadEntryByKeyList(DBFile, RelationFN, &ThisEntry, tmplist, True) == -1) { 
	    /* invalid entry */
	    Qddb_Free(QDDB_TYPE_KEYLIST, *SplitList);
	    SplitList++;
	    continue;
	}
	Qddb_ReducedAttrToFullAttr(schema, ThisEntry);
	CoreEntry = (Qddb_InCoreEntry *)Qddb_Convert(schema, QDDB_ENTRYTYPE_EXTERNAL, 
						ThisEntry, QDDB_ENTRYTYPE_INTERNAL);
	Qddb_Free(QDDB_TYPE_ENTRY, ThisEntry);
	Key = **SplitList;
	Key.Instance = NULL;
	if (Any == False)
	    (void)Qddb_KeyListProcess(schema, *SplitList, CoreEntry, QDDB_KEYLIST_PROC_MARK, 
				       QDDB_KEYLIST_FLAG_MARKFROMINCORE);
	else { 
	    Qddb_InCoreEntry 	*e;
	    
	    (void)Qddb_KeyListProcess(schema, *SplitList, CoreEntry, QDDB_KEYLIST_PROC_MARK, 
				       QDDB_KEYLIST_FLAG_MARKFROMINCORE);
	    e = CoreEntry;
	    while (e->SequenceNumber != 0) {
		e->Marked = True;
		e++;
	    }
	}
	/* *SplitList was deleted by Qddb_KeyListProcess */
	Row = Qddb_RowListProcessEntry(schema, CoreEntry);
#if defined(DEBUG)
	if (Row == NULL) fprintf(stderr, "Row is NULL!!\n");
	else fprintf(stderr, "Row is *not* NULL!!\n");
#endif
	Qddb_RowListDeleteDuplicateRows(CoreEntry, Row, AttrNums, AttrTop);
#if defined(USE_TCL)
	if (print_tcl_elements || suppress_printing) {
	    QueryPrintRowAsElements(interp, schema, &Key, CoreEntry, Row, SepString, 
				    AttrNums, AttrWidth, AttrTop, 
				    AttrNumsUsed, AttrsUsedTop, DontPrintRows, 
				    suppress_printing);
	} else {
	    QueryPrintRow(interp, schema, &Key, CoreEntry, Row, SepString, AttrNums, AttrWidth, AttrTop, 
			  AttrNumsUsed, AttrsUsedTop, TCLprint, DontPrintRows);
	}
#else
	QueryPrintRow(schema, &Key, CoreEntry, Row, SepString, AttrNums, AttrWidth, AttrTop, 
		 AttrNumsUsed, AttrsUsedTop, TCLprint, False);
#endif
	Qddb_Free(QDDB_TYPE_INCOREENTRY, CoreEntry);
	Qddb_Free(QDDB_TYPE_ROWLIST, 	 Row);
	SplitList++;
    }
    Free(AttrNums);
    Free(AttrNumsUsed);
    Free(OrigSplitList);
    return 0;
}

#if defined(USE_TCL)
static void QueryPrintRow(interp, schema, Key, ThisEntry, Row, SepString, 
			  AttrNums, AttrWidth, AttrTop, AttrNumsUsed, AttrUsedTop,
			  TCLprint, DontPrintRows)
    Tcl_Interp		*interp;
    Qddb_Schema		*schema;
    Qddb_KeyList	*Key;
    Qddb_InCoreEntry	*ThisEntry;
    Qddb_RowList	*Row;
    char		*SepString;
    size_t		*AttrNums, *AttrWidth, AttrTop;
    size_t		*AttrNumsUsed, AttrUsedTop;
    Qddb_Boolean	TCLprint;
    Qddb_Boolean	DontPrintRows;
#else
static void QueryPrintRow(schema, Key, ThisEntry, Row, SepString, 
			  AttrNums, AttrWidth, AttrTop, AttrNumsUsed, AttrUsedTop,
			  TCLprint, DontPrintRows)
    Qddb_Schema		*schema;
    Qddb_KeyList	*Key;
    Qddb_InCoreEntry	*ThisEntry;
    Qddb_RowList	*Row;
    char		*SepString;
    size_t		*AttrNums, *AttrWidth, AttrTop;
    size_t		*AttrNumsUsed, AttrUsedTop;
    Qddb_Boolean	TCLprint;
    Qddb_Boolean	DontPrintRows;
#endif
{
    Qddb_RowList	*tmp, *thisrow;
    char		*buf_ptr;
    int			i;

    while (Row != NULL) {
	if (Qddb_RowListShouldSkipRow(ThisEntry, Row, AttrNumsUsed, AttrUsedTop) == True) {
	    Row = Row->down;
	    continue;
	}
	thisrow = Row;
	Qddb_InitBuffer();
	for (i = 0; i < AttrTop; i++) {

	    tmp = Row;
	    while (tmp != NULL) {
		if (AttrNums[i] == ThisEntry[tmp->Attribute].AttributeNumber) {
		    break;
		}
		tmp = tmp->next;
	    }
	    if (tmp == NULL) {
		if (AttrWidth[i] == 0) {
		    Qddb_ConcatBuffer(SepString);
		} else {
		    buf_ptr = Malloc(AttrWidth[i]+strlen(SepString)+2);
		    sprintf(buf_ptr, "%*s%s", (int)AttrWidth[i], " ", SepString);
		    Qddb_ConcatBuffer(buf_ptr);
		    Free(buf_ptr);
		}
	    } else {
		if (AttrWidth[i] != 0) {
		    ThisEntry[tmp->Attribute].Marked = True;
		    Qddb_ConcatBuffer(buf_ptr = Qddb_FieldString(ThisEntry[tmp->Attribute].Data, 
								 (size_t)AttrWidth[i]));
		    Free(buf_ptr);
		    Qddb_ConcatBuffer(SepString);
		} else {
		    ThisEntry[tmp->Attribute].Marked = True;
		    Qddb_ConcatBuffer(buf_ptr = Qddb_FieldString(ThisEntry[tmp->Attribute].Data, 
								 strlen(ThisEntry[tmp->Attribute].Data)));
		    Free(buf_ptr);
		    Qddb_ConcatBuffer(SepString);
		}
	    }
	}
	buf_ptr = Qddb_GetBuffer();
	if (TCLprint == True) {
#if defined(USE_TCL)
	    char		key_buf[BUFSIZ], *row_buf = NULL, *merged_buf;
	    int			nargc = 0;
	    char		*nargv[4];

	    nargv[nargc++] = buf_ptr;
	    sprintf(key_buf, "%d %d %d %d",
		    (int)Key->Start, (int)Key->Length, (int)Key->Number, (int)QDDB_KEYLIST_TYPE(Key));
	    nargv[nargc++] = key_buf;
	    if (DontPrintRows == False) {
		row_buf = nargv[nargc++] = PrintRowInstances(schema, ThisEntry, Row);
	    }
	    nargv[nargc] = NULL;
	    merged_buf = Tcl_Merge(nargc, (CONST char * CONST *)nargv);
	    Free(buf_ptr);
	    if (row_buf != NULL)
		Free(row_buf);
	    if (interp == NULL)
		printf("{ %s }\n", merged_buf);
	    else
		Tcl_AppendElement(interp, merged_buf);
	    Free(merged_buf);
#else
	    fprintf(stderr, "-tcl option is unusable unless qddb is compiled with \
--with-tcl or --with-tk\n");
	    exit(1);
#endif
	} else {
	    printf("%s\n", buf_ptr);
	    Free(buf_ptr);
	}
	Row = Row->down;
    }
}

#if defined(USE_TCL)
static void QueryPrintRowAsElements(interp, schema, Key, ThisEntry, Row, SepString, 
			  AttrNums, AttrWidth, AttrTop, AttrNumsUsed, AttrUsedTop,
			  DontPrintRows, suppress_printing)
    Tcl_Interp		*interp;
    Qddb_Schema		*schema;
    Qddb_KeyList	*Key;
    Qddb_InCoreEntry	*ThisEntry;
    Qddb_RowList	*Row;
    char		*SepString;
    size_t		*AttrNums, *AttrWidth, AttrTop;
    size_t		*AttrNumsUsed, AttrUsedTop;
    Qddb_Boolean	DontPrintRows;
    Qddb_Boolean	suppress_printing;
{
    Qddb_RowList	*tmp, *thisrow;
    char		**merge_buf;
    char		*buf_ptr;
    int			i;
    char		key_buf[BUFSIZ], *row_buf = NULL, *merged_buf;
    int			nargc = 0;
    char		*nargv[4];

    while (Row != NULL) {
	if (Qddb_RowListShouldSkipRow(ThisEntry, Row, AttrNumsUsed, AttrUsedTop) == True) {
	    Row = Row->down;
	    continue;
	}
	thisrow = Row;
	if (!suppress_printing) {
	    merge_buf = (char **)Malloc(sizeof(char *)*(AttrTop+1));
	    for (i = 0; i < AttrTop; i++) {
		tmp = Row;
		while (tmp != NULL) {
		    if (AttrNums[i] == ThisEntry[tmp->Attribute].AttributeNumber) {
			break;
		    }
		    tmp = tmp->next;
		}
		if (tmp == NULL) {
		    if (AttrWidth[i] == 0) {
			merge_buf[i] = Malloc(QDDB_MIN_MALLOC_SIZE);
			*(merge_buf[i]) = '\0';
		    } else {
			buf_ptr = Malloc((size_t)AttrWidth[i]+1);
			sprintf(buf_ptr, "%*s%s", (int)AttrWidth[i], " ", SepString);
			merge_buf[i] = buf_ptr;
		    }
		} else {
		    if (AttrWidth[i] != 0) {
			ThisEntry[tmp->Attribute].Marked = True;
			buf_ptr = Qddb_FieldString(ThisEntry[tmp->Attribute].Data, (size_t)AttrWidth[i]);
			merge_buf[i] = buf_ptr;
		    } else {
			ThisEntry[tmp->Attribute].Marked = True;
			buf_ptr = Qddb_FieldString(ThisEntry[tmp->Attribute].Data, 
						   strlen(ThisEntry[tmp->Attribute].Data));
			merge_buf[i] = buf_ptr;
		    }
		}
	    }
	    merge_buf[AttrTop] = NULL;
	    buf_ptr = Tcl_Merge((int)AttrTop, (CONST char * CONST *)merge_buf);
	    for (i = 0; i < AttrTop; i++)
		Free(merge_buf[i]);
	    Free(merge_buf);
	} else {
	    buf_ptr = Calloc(QDDB_MIN_MALLOC_SIZE);
	}
	nargv[nargc++] = buf_ptr;
	sprintf(key_buf, "%d %d %d %d",
		(int)Key->Start, (int)Key->Length, (int)Key->Number, (int)QDDB_KEYLIST_TYPE(Key));
	nargv[nargc++] = key_buf;
	if (DontPrintRows == False) {
	    row_buf = nargv[nargc++] = PrintRowInstances(schema, ThisEntry, Row);
	}
	nargv[nargc] = NULL;
	merged_buf = Tcl_Merge(nargc, (CONST char * CONST *)nargv);
	Free(buf_ptr);
	if (row_buf != NULL)
	    Free(row_buf);
	Tcl_AppendElement(interp, merged_buf);
	Free(merged_buf);
	Row = Row->down;
    }
}


static char *PrintRowInstances(schema, ThisEntry, Row)
    Qddb_Schema			*schema;
    Qddb_InCoreEntry		*ThisEntry;
    Qddb_RowList		*Row;
{
    char			*name, *instance;
    char			buf[BUFSIZ];

    Qddb_InitBuffer();
    while (Row != NULL) {
	name = Qddb_ConvertAttributeNumberToName(schema, (int)ThisEntry[Row->Attribute].AttributeNumber);
	instance = Qddb_InstanceToString(ThisEntry[Row->Attribute].Instance, 
		 (size_t)schema->Entries[ThisEntry[Row->Attribute].AttributeNumber].Level);
	sprintf(buf, "%s,%s ", name, instance);
	Free(name);
	Free(instance);
	Qddb_ConcatBuffer(buf);
	Row = Row->next;
    }
    return Qddb_GetBuffer();
}
#endif
