
%{

/* $RCSfile: SchParse.y,v $ $Revision: 2.0.0.1 $ $Date: 2000/11/09 15:06:04 $
 *
 * Copyright (C) 1993-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "Qddb.h"

/* Define yyparse() to be SchemaParse() so that we can have multiple parsers
 * in a single executable.
 */
int yyerror _ANSI_ARGS_((char *));
int Qddblex _ANSI_ARGS_(());

#define yyparse		Qddb_SchemaParse
#define yylex   	Qddblex

static Qddb_Schema	*CurrentSchema;

void Qddb_SchemaParse_Set(schema) 
    Qddb_Schema	*schema;
{
    CurrentSchema = schema;
}

Qddb_Schema *Qddb_SchemaParse_Get() 
{
    return CurrentSchema;
}

static int	CurrentLevel;
static int	CurrentAttribute;
static int	CurrentAttributeNumber;
static int	CurrentAncestors[MAXIMUM_NUMBER_OF_SCHEMA_LEVELS];

static void 	FillInSchemaEntry _ANSI_ARGS_((Qddb_SchemaNode *, char *, char *, char *, \
					       char *, char *, char *, Qddb_Boolean, char *, \
					       char *, char *));
static int      SchemaFormatCheck _ANSI_ARGS_((char *, char *));

%}
%pure_parser
%expect 1	

%token STRING CSTRING
%token ALIAS VERBOSENAME
%token LPAREN RPAREN ASTERISK COMMA DOUBLEQUOTE SINGLEQUOTE EQUALS
%token BLANKLINE NUMBER TYPE DEFAULTVALUE FORMAT DATEFORMAT
%token HASHSIZE USE CACHED HASHING REDUCED ATTRIBUTE IDENTIFIERS 
%token CONDENSED INDEXING
%token CACHESIZE HASHTYPE
%token SEPARATORS EXCLUDE
%token INTEGER 
%token SECONDARY SEARCH MAXMEM EXCLUDEWORDS MAXFILES
%token LARGE DATABASE OPTIMIZATIONS PROXIMITY
%token AUTOINCREMENT

%type <String>	STRING CSTRING
%type <Integer> Identifier IdentifierLparen
%type <Integer> Options INTEGER
%type <Option> AttrOption AttrOptions

%union {
    char		*String;
    int			Integer;
    double		Real;
    Qddb_AttrOption	*Option;
};


%%
SchemaFile:
    OptionList Record {
    } |
    Record {
    };

OptionList:
    Options OptionList {
    } |
    {
    };

Options:
    MAXMEM EQUALS INTEGER {
	qddb_maxmem = $3; /* Only used for stabilization anyway */
    } |
    MAXFILES EQUALS INTEGER {
	qddb_maxfiles = $3;
    } |
    DATEFORMAT EQUALS CSTRING {
        Free(CurrentSchema->default_date_format);
        CurrentSchema->default_date_format = $3;
    } |
    HASHSIZE EQUALS INTEGER {
	CurrentSchema->HashSize = $3;
    } |
    HASHTYPE EQUALS INTEGER {
        CurrentSchema->HashType = $3;
    } |
    CACHESIZE EQUALS INTEGER {
	CurrentSchema->CacheSize = $3;
    } |
    USE CACHED HASHING {
	CurrentSchema->UseCachedHashing = True;
    } |
    USE REDUCED ATTRIBUTE IDENTIFIERS {
	CurrentSchema->UseReducedAttributeIdentifiers = True;
    } |
    USE CACHED SECONDARY SEARCH {
        CurrentSchema->UseCachedSecondarySearch = True;
    } |
    USE EXCLUDEWORDS {
	CurrentSchema->UseExcludeWords = True;
    } |
    USE CONDENSED INDEXING {
	CurrentSchema->UseCondensedIndexing = True;
    } |
    USE LARGE DATABASE OPTIMIZATIONS {
	CurrentSchema->UseLargeDatabaseOptimizations = True;
    } |
    USE PROXIMITY INDEXING {
	CurrentSchema->UseProximityIndexing = True;
    };
    

Record:
    IndividualRecord {
    } |
    IndividualRecord Record {
    };

IndividualRecord:
    Identifier {
	CurrentSchema->Entries[$1].IsLeaf = True;
    } |
    IdentifierLparen Record RPAREN {
	CurrentLevel--;
	CurrentAttributeNumber = CurrentAncestors[CurrentLevel]+1;
	CurrentSchema->Entries[$1].IsLeaf = False;
#ifdef DEBUG
fprintf(stderr, "ID ( : Name=%s; Alias=%s; IsLeaf=%s;\n",
	CurrentSchema->Entries[$1].Name,
	CurrentSchema->Entries[$1].Alias,
	CurrentSchema->Entries[$1].IsLeaf==True? "True":"False");
#endif
    } |
    Identifier ASTERISK {
	CurrentSchema->Entries[$1].AsteriskPresent = True;
	CurrentSchema->Entries[$1].IsLeaf = True;
    } |
    IdentifierLparen Record RPAREN ASTERISK {
	CurrentSchema->Entries[$1].AsteriskPresent = True;
	CurrentLevel--;
	CurrentAttributeNumber = CurrentAncestors[CurrentLevel]+1;
	CurrentSchema->Entries[$1].IsLeaf = False;
    };

IdentifierLparen:
    Identifier LPAREN {
	CurrentAncestors[CurrentLevel] = CurrentSchema->Entries[$1].Number;
	CurrentLevel++;
	$$ = $1;
	CurrentAttributeNumber = 1;
    };

Identifier:
    STRING {
	$$ = CurrentAttribute;
	FillInSchemaEntry(CurrentSchema->Entries+CurrentAttribute, 
			  $1, NULL, NULL, NULL, NULL, NULL, False, NULL,
			  NULL, NULL);
    } | 
    STRING AttrOptions {
        char		*Name, *Alias, *VerboseName, *Type, *Format;
	char            *Separators, *DefaultValue;
	char            *Autoincrement1, *Autoincrement2;
        Qddb_AttrOption *ptr, *tmp;
        Qddb_Boolean    exclude = False;

	$$ = CurrentAttribute;
        Name = Alias = VerboseName = Type = Format = Separators = DefaultValue = NULL;
	Autoincrement1 = Autoincrement2 = NULL;
        for (ptr = $2; ptr != NULL; ptr = ptr->next) {
	    switch (ptr->option) {
	    case AUTOINCREMENT:
	        Autoincrement1 = ptr->value;
		Autoincrement2 = ptr->value2;
		Type = Malloc(strlen("autoincrement")+1);
		strcpy(Type, "autoincrement");
		break;
	    case DEFAULTVALUE:
		DefaultValue = ptr->value;
		break;
            case VERBOSENAME:
		VerboseName = ptr->value;
		break;
	    case TYPE:
		Type = ptr->value;
		break;
	    case FORMAT:
		Format = ptr->value;
		break;
	    case SEPARATORS:
		Separators = ptr->value;
		break;
	    case ALIAS:
		Alias = ptr->value;
		break;
	    case EXCLUDE:
		exclude = True;
                break;
	    default:
		fprintf(stderr, "SchemaParse: internal error: cannot recognize attribute option %d\n", 
			ptr->option);
	    }
	}
        for (ptr = $2; ptr != NULL;) {
	    tmp = ptr->next;
	    Free(ptr);
	    ptr = tmp;
	}
	FillInSchemaEntry(CurrentSchema->Entries+CurrentAttribute, 
			  $1, Alias, VerboseName, Type, Format, Separators,
			  exclude, DefaultValue, Autoincrement1, Autoincrement2);
    };

AttrOptions:
    AttrOption {
        $$ = $1;
    } |
    AttrOption AttrOptions {
        $$ = $1;
        $$->next = $2;
    };

AttrOption:
    EXCLUDE {
        $$ = (Qddb_AttrOption *)Malloc(sizeof(Qddb_AttrOption));
        $$->option = EXCLUDE;
        $$->value = NULL;
        $$->next = NULL;
    } |
    SEPARATORS CSTRING {
	$$ = (Qddb_AttrOption *)Malloc(sizeof(Qddb_AttrOption));
        $$->option = SEPARATORS;
        $$->value = $2;
        $$->next = NULL;
    } |
    FORMAT CSTRING {
	$$ = (Qddb_AttrOption *)Malloc(sizeof(Qddb_AttrOption));
        $$->option = FORMAT;
        $$->value = $2;
        $$->next = NULL;
    } |
    TYPE STRING {
	$$ = (Qddb_AttrOption *)Malloc(sizeof(Qddb_AttrOption));
        $$->option = TYPE;
        $$->value = $2;
        $$->next = NULL;
    } |
    ALIAS STRING {
	$$ = (Qddb_AttrOption *)Malloc(sizeof(Qddb_AttrOption));
        $$->option = ALIAS;
        $$->value = $2;
        $$->next = NULL;
    } |
    VERBOSENAME CSTRING {
	$$ = (Qddb_AttrOption *)Malloc(sizeof(Qddb_AttrOption));
        $$->option = VERBOSENAME;
        $$->value = $2;
        $$->next = NULL;
    } |
    DEFAULTVALUE CSTRING {
        $$ = (Qddb_AttrOption *)Malloc(sizeof(Qddb_AttrOption));
        $$->option = DEFAULTVALUE;
        $$->value = $2;
        $$->next = NULL;
    } |
    AUTOINCREMENT LPAREN INTEGER COMMA INTEGER RPAREN {
        char            *intbuf;

        $$ = (Qddb_AttrOption *)Malloc(sizeof(Qddb_AttrOption));
        $$->option = AUTOINCREMENT;
	intbuf = Malloc(sizeof(int)*16);
	sprintf(intbuf, "%d", $3);
        $$->value = intbuf;
	intbuf = Malloc(sizeof(int)*16);
	sprintf(intbuf, "%d", $5);
        $$->value2 = intbuf;
        $$->next = NULL;
    } |
    AUTOINCREMENT LPAREN INTEGER RPAREN {
        char            *intbuf;

        $$ = (Qddb_AttrOption *)Malloc(sizeof(Qddb_AttrOption));
        $$->option = AUTOINCREMENT;
	intbuf = Malloc(sizeof(int)*16);
	strcpy(intbuf, "0");
        $$->value = intbuf;
	intbuf = Malloc(sizeof(int)*16);
	sprintf(intbuf, "%d", $3);
        $$->value2 = intbuf;
        $$->next = NULL;
    };
%%

static int SchemaFormatCheck(form, val)
    char		*form, *val;
{
    Qddb_Regexp		*RegexpVal;

    RegexpVal = Qddb_Regcomp(form);
    if (RegexpVal == NULL)
	return 0;
    return Qddb_Regexec(RegexpVal, val);
}

static void FillInSchemaEntry(TheEntry, Name, Alias, VerboseName, Type,
			      Format, Separators, Exclude, DefaultValue,
			      Autoincrement1, Autoincrement2)
    Qddb_SchemaNode	*TheEntry;
    char		*Name, *Alias, *VerboseName, *Type, *Format, *Separators;
    Qddb_Boolean	Exclude;
    char		*DefaultValue, *Autoincrement1, *Autoincrement2;
{
    int			i;
#ifdef DEBUG
fprintf(stderr, "Filling :%s:%s:%s:%s:\n", Name, Alias, VerboseName, Type);
#endif
    CurrentSchema->NumberOfAttributes++;
    if (Name == NULL) {
	TheEntry->Name = Malloc(1);
	TheEntry->Name[0] = '\0';
    } else {
        TheEntry->Name = Name;
    }
    if (Alias == NULL) {
	TheEntry->Alias = Malloc(1);
	TheEntry->Alias[0] = '\0';
    } else {
	TheEntry->Alias = Alias;
    }
    if (VerboseName == NULL) {
	TheEntry->VerboseName = Malloc(1);
	TheEntry->VerboseName[0] = '\0';
    } else {
	TheEntry->VerboseName = VerboseName;
    }
    TheEntry->Level = CurrentLevel;
    for (i = 1; i < CurrentLevel; i++) {
	TheEntry->AncestorNumber[i] = CurrentAncestors[i];
    }
#ifdef DEBUG
    {int j;
for (j = 1; j < CurrentLevel; j++)
printf("%d.", TheEntry->AncestorNumber[j]);
printf("%d", CurrentAttributeNumber);
printf("\n");
 }
#endif
    CurrentAncestors[CurrentLevel] = CurrentAttributeNumber;
    TheEntry->AsteriskPresent = False;
    TheEntry->Number = CurrentAttributeNumber;
    TheEntry->Options = QAO_None;
    CurrentAttribute++;
    CurrentAttributeNumber++;
    if (Type == NULL || Type[0] == '\0') {
	TheEntry->Type = QAT_Typeless;
    } else if (Qddb_StrcmpNoupper(Type, "Typeless") == 0) {
	TheEntry->Type = QAT_Typeless;
    } else if (Qddb_StrcmpNoupper(Type, "Real") == 0) {
	TheEntry->Type = QAT_Real;
    } else if (Qddb_StrcmpNoupper(Type, "Autoincrement") == 0) {
	TheEntry->Type = QAT_Integer;
	TheEntry->Options = QAO_Autoincrement;
	TheEntry->AutoStart = (Autoincrement1 != NULL) ? strtol(Autoincrement1, NULL, 10) : 0;
	TheEntry->AutoIncrement = (Autoincrement2 != NULL) ? strtol(Autoincrement2, NULL, 10) : 1;
    } else if (Qddb_StrcmpNoupper(Type, "Integer") == 0) {
	TheEntry->Type = QAT_Integer;
    } else if (Qddb_StrcmpNoupper(Type, "String") == 0) {
	TheEntry->Type = QAT_String;
    } else if (Qddb_StrcmpNoupper(Type, "Date") == 0) {
	TheEntry->Type = QAT_Date;
    } else if (Qddb_StrcmpNoupper(Type, "Time") == 0) {
	TheEntry->Type = QAT_Time;
    } else {
	fprintf(stderr, "Syntax Error: Type '%s'\n", Type);
	fprintf(stderr, "Redeclaring '%s' to 'Typeless'\n", Type);
	TheEntry->Type = QAT_Typeless;
    }
    if (Type != NULL) {
	Free(Type);
    }
    if (DefaultValue == NULL) {
	/* Add a reasonable default value */
	switch (TheEntry->Type) {
	case QAT_Real:
	    DefaultValue = Malloc(5);
	    strcpy(DefaultValue, "0.00");
	    break;
	case QAT_Integer:
	    DefaultValue = Malloc(2);
	    strcpy(DefaultValue, "0");
	    break;
	default:
	    DefaultValue = Malloc(1);
	    DefaultValue[0] = '\0';
	    break; 
	}
	TheEntry->DefaultValue = DefaultValue; /* "" is a valid value for any type of attribute */
    } else {
        char            *ch;
	int             l;
	double          d;

	/* Check out the default value for validity */

	switch (TheEntry->Type) {
	case QAT_Real:
	    if (DefaultValue[0] == '\0')
		break;
	    d = strtod(DefaultValue, &ch);
	    if (DefaultValue+strlen(DefaultValue) != ch) {
		fprintf(stderr, "Invalid default value '%s' for attribute '%s'\n", 
			DefaultValue, Name);
		fprintf(stderr, "\tvalue must be a valid real number, for example \"1.002\".\n");
		fprintf(stderr, "\t(defaulting to \"0.00\")\n");
		Free(DefaultValue);
		DefaultValue = Malloc(5);
		strcpy(DefaultValue, "0.00");
	    }
	    break;
	case QAT_Integer:
	    if (DefaultValue[0] == '\0')
		break;
	    l = strtol(DefaultValue, &ch, 0);
	    if (DefaultValue+strlen(DefaultValue) != ch) {
		fprintf(stderr, "Invalid default value '%s' for attribute '%s'\n", 
			DefaultValue, Name);
		fprintf(stderr, "\tvalue must be a valid integer, for example \"112\".\n");
		fprintf(stderr, "\t(defaulting to \"0\")\n");
		Free(DefaultValue);
		DefaultValue = Malloc(2);
		strcpy(DefaultValue, "0");
	    }
	    break;
	case QAT_Date:
	    if (DefaultValue[0] == '\0')
		break;
	    if (Qddb_CheckDate(DefaultValue) != 0) {
		fprintf(stderr, "Invalid default value '%s' for attribute '%s'\n", 
			DefaultValue, Name);
		fprintf(stderr, "\tvalue must be a valid date, for example \"today\".\n");
		fprintf(stderr, "\t(defaulting to \"\")\n");
		Free(DefaultValue);
		DefaultValue = Malloc(1);
		DefaultValue[0] = '\0';
	    }
	    break;
	case QAT_Time:
	    if (DefaultValue[0] == '\0')
	        break;
	    if (Qddb_CheckTime(DefaultValue) != 0) {
		fprintf(stderr, "Invalid default value '%s' for attribute '%s'\n", 
			DefaultValue, Name);
		fprintf(stderr, "\tvalue must be a valid time, for example \"01:13:13\".\n");
		fprintf(stderr, "\t(defaulting to \"\")\n");
		Free(DefaultValue);
		DefaultValue = Malloc(1);
		DefaultValue[0] = '\0';
	    }
	    break;
	default:
	    /* String and Typeless are both ok for any default value string */
	    break; 
	}
	TheEntry->DefaultValue = DefaultValue;
    }
    if (Format != NULL) {
	switch (TheEntry->Type) {
	case QAT_Integer:
	    if (SchemaFormatCheck("^\\%[\\-\\#0\\ \\+]*[0-9]*\\.?[0-9]*[dui]$", Format) == 0 &&
		SchemaFormatCheck("^0x\\%[\\-\\#0\\ \\+]*[0-9]*\\.?[0-9]*[xX]$", Format) == 0 &&
		SchemaFormatCheck("^0\\%[\\-\\#0\\ \\+]*[0-9]*\\.?[0-9]*o$", Format) == 0) {
		fprintf(stderr, "Invalid format specifier (%s) for type Integer\n", Format);
		Free(Format);
		Format = NULL;
	    }
	    break;
	case QAT_Real:
	    if (SchemaFormatCheck("^\\%[\\-\\#0\\ \\+]*[0-9]*\\.?[0-9]*[fegGE]$", Format) == 0) {
		fprintf(stderr, "Invalid format specifier (%s) for type Real\n", Format);
		Free(Format);
		Format = NULL;
	    }
	    break;
	case QAT_Date:
	    break;
	default:
	    ;
	}
    }
    if (Format == NULL) {
	switch (TheEntry->Type) {
	case QAT_Typeless:
	    TheEntry->Format = Malloc(3);
	    strcpy(TheEntry->Format, "%s");
	    break;
	case QAT_Real:
	    TheEntry->Format = Malloc(3);
	    strcpy(TheEntry->Format, "%f");
	    break;
	case QAT_Integer:
	    TheEntry->Format = Malloc(3);
	    strcpy(TheEntry->Format, "%d");
	    break;
	case QAT_String:
	    TheEntry->Format = Malloc(3);
	    strcpy(TheEntry->Format, "%s");
	    break;
	case QAT_Date:
	    TheEntry->Format = Malloc(strlen(CurrentSchema->default_date_format)+1);
	    strcpy(TheEntry->Format, CurrentSchema->default_date_format);
	    break;
	default:
	    TheEntry->Format = Malloc(1);
	    TheEntry->Format[0] = '\0';
	}
    } else {
	if (TheEntry->Type == QAT_String || TheEntry->Type == QAT_Typeless) {
	    fprintf(stderr, "format specifier in attribute of type string ignored\n");
	    TheEntry->Format = Malloc(3);
	    strcpy(TheEntry->Format, "%s");
	} else {
	    TheEntry->Format = Format;
	}
    }
    if (Separators == NULL) {
	switch (TheEntry->Type) {
	case QAT_Typeless:
	    TheEntry->Separators = Malloc(strlen(QDDB_DEFAULT_SEPARATORS)+1);
	    strcpy(TheEntry->Separators, QDDB_DEFAULT_SEPARATORS);
	    break;
	case QAT_Real:
	    TheEntry->Separators = Malloc(strlen(QDDB_DEFAULT_REAL_SEPARATORS)+1);
	    strcpy(TheEntry->Separators, QDDB_DEFAULT_REAL_SEPARATORS);
	    break;
	case QAT_Integer:
	    TheEntry->Separators = Malloc(strlen(QDDB_DEFAULT_INTEGER_SEPARATORS)+1);
	    strcpy(TheEntry->Separators, QDDB_DEFAULT_INTEGER_SEPARATORS);
	    break;
	case QAT_String:
	    TheEntry->Separators = Malloc(strlen(QDDB_DEFAULT_SEPARATORS)+1);
	    strcpy(TheEntry->Separators, QDDB_DEFAULT_SEPARATORS);
	    break;
	case QAT_Date:
	    TheEntry->Separators = Malloc(strlen(QDDB_DEFAULT_DATE_SEPARATORS)+1);
	    strcpy(TheEntry->Separators, QDDB_DEFAULT_DATE_SEPARATORS);
	    break;
	case QAT_Time:
	    TheEntry->Separators = Malloc(strlen(QDDB_DEFAULT_TIME_SEPARATORS)+1);
	    strcpy(TheEntry->Separators, QDDB_DEFAULT_TIME_SEPARATORS);
	    break;
	default:
	    TheEntry->Separators = Malloc(2);
	    TheEntry->Separators[0] = '\n';
	    TheEntry->Separators[1] = '\0';	    
	}
    } else {
	if (strchr(Separators, '\n') == NULL) {
	    char		*tmp = Separators;

	    Separators = Malloc(strlen(tmp)+2);
	    strcpy(Separators, tmp);
	    strcat(Separators, "\n");
	    Free(tmp);
	}
	TheEntry->Separators = Separators;
    }
    TheEntry->Exclude = Exclude;
}

void Qddb_ResetSchemaParser()
{
    CurrentAttribute = 1;
    CurrentLevel = 1;
    CurrentAttributeNumber = 1;
    Qddb_ResetSchemaLex();
}
