%{

/* $RCSfile: ScriptParse.y,v $ $Revision: 2.0.0.1 $ $Date: 2000/11/09 15:06:04 $
 *
 * Copyright (C) 1993-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "Qddb.h"

/* Define yyparse() to be EntryParse() so that multiple parsers can
 * occupy the same executable.
 */
#define yyparse	QddbScriptParse
#define yylex   QddbScriptlex

int yylex();

static Qddb_ScriptCell *cells = NULL, *lastcell = NULL;
static Qddb_ScriptCell current_cell;
static int current_values = 0;

int GetOption _ANSI_ARGS_((char *));
void AddValue _ANSI_ARGS_((char *));
void ClearCurrentCell _ANSI_ARGS_((void));

extern int Qddb_LineNumber;

%}

%pure_parser
%expect 2

%token STRING CSTRING INTEGER
%token LBRACKET RBRACKET EQUALS ENDOPTION

%type <String> STRING CSTRING string
%type <Integer> INTEGER
%type <ScriptCell> script script_items

%union {
    char               *String;
    int                Integer;
    Qddb_ScriptCell    *ScriptCell;
};

%%

script:
    script_items {
    };

script_items:
    LBRACKET STRING options RBRACKET {
	Qddb_ScriptCell    *newcell;

	newcell = (Qddb_ScriptCell *)Malloc(sizeof(Qddb_ScriptCell));
	bcopy(&current_cell, newcell, sizeof(current_cell));
	newcell->option = GetOption($2);
	newcell->linenumber = Qddb_LineNumber;
	if (lastcell != NULL) {
	    lastcell->next = newcell;
	} else {
	    cells = newcell;
	}
	lastcell = newcell;
	ClearCurrentCell();
	Free($2);
	$$ = cells;
    } |
    LBRACKET STRING RBRACKET {
	Qddb_ScriptCell    *newcell;

	newcell = (Qddb_ScriptCell *)Calloc(sizeof(Qddb_ScriptCell));
	newcell->option = GetOption($2);
	newcell->linenumber = Qddb_LineNumber;
	if (lastcell != NULL) {
	    lastcell->next = newcell;
	} else {
	    cells = newcell;
	}
	lastcell = newcell;
	ClearCurrentCell();
	Free($2);
	$$ = cells;
    } |
    LBRACKET ENDOPTION STRING RBRACKET {
	Qddb_ScriptCell    *newcell;

	newcell = (Qddb_ScriptCell *)Calloc(sizeof(Qddb_ScriptCell));
	newcell->option = GetOption($3);
	newcell->endoption = True;
	newcell->linenumber = Qddb_LineNumber;
	if (lastcell != NULL) {
	    lastcell->next = newcell;
	} else {
	    cells = newcell;
	}
	lastcell = newcell;
	ClearCurrentCell();
	Free($3);
	$$ = cells;
    } |
    script_items LBRACKET STRING options RBRACKET {
	Qddb_ScriptCell    *newcell;

	newcell = (Qddb_ScriptCell *)Malloc(sizeof(Qddb_ScriptCell));
	bcopy(&current_cell, newcell, sizeof(current_cell));
	newcell->option = GetOption($3);
	newcell->linenumber = Qddb_LineNumber;
	if (lastcell != NULL) {
	    lastcell->next = newcell;
	} else {
	    cells = newcell;
	}
	lastcell = newcell;
	ClearCurrentCell();
	Free($3);
	$$ = cells;
    } |
    script_items LBRACKET STRING RBRACKET {
	Qddb_ScriptCell    *newcell;

	newcell = (Qddb_ScriptCell *)Calloc(sizeof(Qddb_ScriptCell));
	newcell->option = GetOption($3);
	newcell->linenumber = Qddb_LineNumber;
	if (lastcell != NULL) {
	    lastcell->next = newcell;
	} else {
	    cells = newcell;
	}
	lastcell = newcell;
	ClearCurrentCell();
	Free($3);
	$$ = cells;
    } |
    script_items LBRACKET ENDOPTION STRING RBRACKET {
	Qddb_ScriptCell    *newcell;

	newcell = (Qddb_ScriptCell *)Calloc(sizeof(Qddb_ScriptCell));
	newcell->option = GetOption($4);
	newcell->endoption = True;
	newcell->linenumber = Qddb_LineNumber;
	if (lastcell != NULL) {
	    lastcell->next = newcell;
	} else {
	    cells = newcell;
	}
	lastcell = newcell;
	ClearCurrentCell();
	Free($4);
	$$ = cells;
    };

options:
    |
    options STRING EQUALS string {
	if (strncasecmp($2, "attr", (size_t)4) == 0) {
	    current_cell.attr = $4;
	} else if (strcasecmp($2, "label") == 0) {
	    current_cell.label = $4;
	} else if (strcasecmp($2, "anchor") == 0) {
	    if (strcasecmp($4, "w") == 0) {
		current_cell.anchor = QDDB_SCRIPT_ANCHOR_W;
	    } else if (strcasecmp($4, "e") == 0) {
		current_cell.anchor = QDDB_SCRIPT_ANCHOR_E;
	    } else if (strcasecmp($4, "n") == 0) {
		current_cell.anchor = QDDB_SCRIPT_ANCHOR_N;
	    } else if (strcasecmp($4, "s") == 0) {
		current_cell.anchor = QDDB_SCRIPT_ANCHOR_S;
	    } else if (strcasecmp($4, "nw") == 0) {
		current_cell.anchor = QDDB_SCRIPT_ANCHOR_NW;
	    } else if (strcasecmp($4, "ne") == 0) {
		current_cell.anchor = QDDB_SCRIPT_ANCHOR_NE;
	    } else if (strcasecmp($4, "sw") == 0) {
		current_cell.anchor = QDDB_SCRIPT_ANCHOR_SW;
	    } else if (strcasecmp($4, "se") == 0) {
		current_cell.anchor = QDDB_SCRIPT_ANCHOR_SE;
	    } else if (strcasecmp($4, "center") == 0) {
		current_cell.anchor = QDDB_SCRIPT_ANCHOR_CENTER;
	    }
	} else if (strcasecmp($2, "value") == 0) {
	    AddValue($4);
	} else if (strcasecmp($2, "config") == 0) {
	    current_cell.config = $4;
	} else if (strcasecmp($2, "side") == 0) {
	    if (strcasecmp($4, "left") == 0) {
		current_cell.side = QDDB_SCRIPT_SIDE_LEFT;
	    } else if (strcasecmp($4, "right") == 0) {
		current_cell.side = QDDB_SCRIPT_SIDE_RIGHT;
	    } else if (strcasecmp($4, "top") == 0) {
		current_cell.side = QDDB_SCRIPT_SIDE_TOP;
	    } else if (strcasecmp($4, "bottom") == 0) {
		current_cell.side = QDDB_SCRIPT_SIDE_BOTTOM;
	    }
	    Free($4);
	} else if (strcasecmp($2, "labelside") == 0) {
	    if (strcasecmp($4, "left") == 0) {
		current_cell.side = QDDB_SCRIPT_SIDE_LEFT;
	    } else if (strcasecmp($4, "right") == 0) {
		current_cell.side = QDDB_SCRIPT_SIDE_RIGHT;
	    } else if (strcasecmp($4, "top") == 0) {
		current_cell.side = QDDB_SCRIPT_SIDE_TOP;
	    } else if (strcasecmp($4, "bottom") == 0) {
		current_cell.side = QDDB_SCRIPT_SIDE_BOTTOM;
	    }
	    Free($4);
	} else if (strcasecmp($2, "relief") == 0) {
	    if (strcasecmp($4, "groove") == 0) {
		current_cell.relief = QDDB_SCRIPT_RELIEF_GROOVE;
	    } else if (strcasecmp($4, "sunken") == 0) {
		current_cell.relief = QDDB_SCRIPT_RELIEF_SUNKEN;
	    } else if (strcasecmp($4, "raised") == 0) {
		current_cell.relief = QDDB_SCRIPT_RELIEF_RAISED;
	    } else if (strcasecmp($4, "flat") == 0) {
		current_cell.relief = QDDB_SCRIPT_RELIEF_FLAT;
	    }
	    Free($4);
	} else if (strcasecmp($2, "bannergif") == 0) {
	    current_cell.bannergif = $4;
	} else if (strcasecmp($2, "bannerxbm") == 0) {
	    current_cell.bannerxbm = $4;
	} else if (strcasecmp($2, "bannertitle") == 0) {
	    current_cell.bannertitle = $4;
	}
	Free($2);
    } |
    options STRING EQUALS INTEGER {
	if (strcasecmp($2, "width") == 0) {
	    current_cell.width = $4;
	} else if (strcasecmp($2, "labelwidth") == 0) {
	    current_cell.labelwidth = $4;
	} else if (strcasecmp($2, "height") == 0) {
	    current_cell.height = $4;
	} else if (strcasecmp($2, "border") == 0) {
	    current_cell.border = $4;
	} else if (strcasecmp($2, "padx") == 0) {
	    current_cell.padx = $4;
	} else if (strcasecmp($2, "pady") == 0) {
	    current_cell.pady = $4;
	} else if (strcasecmp($2, "ipadx") == 0) {
	    current_cell.ipadx = $4;
	} else if (strcasecmp($2, "ipady") == 0) {
	    current_cell.ipady = $4;
	} else if (strcasecmp($2, "underline") == 0) {
	    current_cell.underline = $4;
	}
	Free($2);
    } |
    options STRING {
	if (strcasecmp($2, "noexpand") == 0) {
	    current_cell.noexpand = True;
	} else if (strcasecmp($2, "expand") == 0) {
	    current_cell.noexpand = False;
	} else if (strcasecmp($2, "autoalign") == 0) {
	    current_cell.autoalign = True;
	} else if (strcasecmp($2, "noautoalign") == 0) {
	    current_cell.autoalign = Inapplicable;
	}
	Free($2);
    }
    ;

string:
    STRING {
        $$ = $1;
    } |
    CSTRING {
        $$ = $1;
    };
%%

void ClearCurrentCell()
{
    bzero(&current_cell, sizeof(current_cell));
    current_values = 0;
}

int GetOption(option)
    char               *option;
{
    if (strcasecmp(option, "frame") == 0) {
	return QDDB_SCRIPT_OPTION_FRAME;
    } else if (strcasecmp(option, "entry") == 0) {
	return QDDB_SCRIPT_OPTION_ENTRY;
    } else if (strcasecmp(option, "text") == 0) {
	return QDDB_SCRIPT_OPTION_TEXT;
    } else if (strcasecmp(option, "radiobutton") == 0) {
	return QDDB_SCRIPT_OPTION_RADIOBUTTON;
    } else if (strcasecmp(option, "label") == 0) {
	return QDDB_SCRIPT_OPTION_LABEL;
    } else if (strcasecmp(option, "notebook") == 0) {
	return QDDB_SCRIPT_OPTION_NOTEBOOK;
    } else if (strcasecmp(option, "tab") == 0) {
	return QDDB_SCRIPT_OPTION_TAB;
    } else if (strcasecmp(option, "break") == 0) {
	return QDDB_SCRIPT_OPTION_BREAK;
    } else if (strcasecmp(option, "epilogue") == 0) {
	return QDDB_SCRIPT_OPTION_EPILOGUE;
    } else if (strcasecmp(option, "prologue") == 0) {
	return QDDB_SCRIPT_OPTION_PROLOGUE;
    } else if (strcasecmp(option, "code") == 0) {
	return QDDB_SCRIPT_OPTION_CODE;
    } else if (strcasecmp(option, "menubar") == 0) {
	return QDDB_SCRIPT_OPTION_MENUBAR;
    } else if (strcasecmp(option, "options") == 0) {
	return QDDB_SCRIPT_OPTION_OPTIONS;
    }
    return 0;
}

void AddValue(value)
    char               *value;
{
    if (current_values == 0) {
	current_cell.values = (char **)Malloc(sizeof(char *)*2);
	current_cell.values[0] = value;
	current_cell.values[1] = NULL;
	current_values = 1;
	return;
    }
    current_cell.values = (char **)Realloc(current_cell.values,
					   sizeof(char *) * (current_values + 2));
    current_cell.values[current_values] = value;
    current_cell.values[current_values+1] = NULL;
    current_values++;
}

void Qddb_ResetQddbScriptParser()
{
    cells = NULL;
    lastcell = NULL;
    ClearCurrentCell();
    Qddb_ResetQddbScriptLex();
}

Qddb_ScriptCell *Qddb_GetQddbScriptParserResults()
{
    return cells;
}
