%{

/* $RCSfile: TCLParse.y,v $ $Revision: 2.0.0.1 $ $Date: 2000/11/09 15:06:04 $
 *
 * Copyright (C) 1993-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "Qddb.h"

/* Define yyparse() to be EntryParse() so that multiple parsers can
 * occupy the same executable.
 */
#define yyparse	Qddb_TCLParse
#define yylex   Qddb_TCLlex

int yylex();

static Qddb_Schema	*CurrentSchema;
static Qddb_InCoreEntry	*CurrentEntries = NULL;
static unsigned int	CurrentEntriesLength;

static int TCL_FindAttribute _ANSI_ARGS_((char *));

void Qddb_TCLParse_Set(schema)
    Qddb_Schema		*schema;
{
    CurrentSchema = schema;
}

Qddb_InCoreEntry *Qddb_TCLParse_Get()
{
    return CurrentEntries;
}

static unsigned int	CurrentAttribute = 0;

static unsigned int	CurrentAncestors[MAXIMUM_NUMBER_OF_SCHEMA_LEVELS];
static unsigned int	CurrentInstances[MAXIMUM_NUMBER_OF_SCHEMA_LEVELS];
static unsigned int	CurrentLevel = 1;
%}

%pure_parser
%expect 2

%token STRING CSTRING
%token LBRACE RBRACE 
%token NUMBER

%type <String>		STRING CSTRING
%type <Integer>		String Cstring 

%union {
    char	*String;
    double	Real;
    int		Integer;
};


%%

FileOfEntries:
    WholeEntry {
    } |
    WholeEntry FileOfEntries {
    };

WholeEntry:
    UniqueSequenceNumber Entry {
	CurrentEntriesLength = CurrentAttribute;
	CurrentEntries[CurrentAttribute+1].SequenceNumber = 0;
	CurrentAttribute = 0;
    };

Entry:
    /* return an TEMPLATE */
    Record {
    } |
    Record Entry {
    };

Record:
    /* return an INSTANCE */
    IndividualRecord {
    } |
    IndividualRecord Record {

    };

IndividualRecord:
    /* return a TEMPLATE */
    LBRACE Cstring RBRACE {
    } |
    LbraceString Record RBRACE {
	CurrentLevel--;
    };

LbraceString:
    LBRACE String {
	CurrentLevel++;
	CurrentAttribute--;
    };

String:
    STRING {
	Qddb_SchemaNode	*ThisAttribute;
	int		AttrNo;
	int		i;

    	/* return an INSTANCE */
	AttrNo = TCL_FindAttribute($1);
	if (AttrNo < 0) {
	    fprintf(stderr, "  Attribute named '%s' is either misplaced or it is\n", $1);
	    fprintf(stderr, "not an attribute name in Schema\n");
	    YYERROR;
	}

	ThisAttribute = CurrentSchema->Entries+AttrNo;
	CurrentEntries[CurrentAttribute].BytesOfData = 0;
	CurrentEntries[CurrentAttribute].AttributeNumber = AttrNo;
	$$ = ThisAttribute->Number;
	if (CurrentAncestors[CurrentLevel] != ThisAttribute->Number) {
	    CurrentInstances[CurrentLevel] = 1;
	} else
	    CurrentInstances[CurrentLevel]++;
	CurrentAncestors[CurrentLevel] = ThisAttribute->Number;
	for (i = 1; i <= CurrentLevel; i++)
	    CurrentEntries[CurrentAttribute].Instance[i] = CurrentInstances[i];
	CurrentAttribute++;
    };

Cstring:
    STRING CSTRING {
	Qddb_SchemaNode	*ThisAttribute;
	int		AttrNo;
	int		i;

	AttrNo = TCL_FindAttribute($1);
	if (AttrNo < 0) {
	    fprintf(stderr, "  Attribute named '%s' is either misplaced or it is\n", $1);
	    fprintf(stderr, "not an attribute name in Schema\n");
	    YYERROR;
	}	
	ThisAttribute = CurrentSchema->Entries+AttrNo;
	CurrentEntries[CurrentAttribute].AttributeNumber =
	    AttrNo;
	CurrentEntries[CurrentAttribute].Data = $2;
	CurrentEntries[CurrentAttribute].BytesOfData = strlen($2);
	$$ = ThisAttribute->Number;
	if (CurrentAncestors[CurrentLevel] != ThisAttribute->Number) {
	    CurrentInstances[CurrentLevel] = 1;
	} else
	    CurrentInstances[CurrentLevel]++;
	CurrentAncestors[CurrentLevel] = ThisAttribute->Number;
	for (i = 1; i <= CurrentLevel; i++)
	    CurrentEntries[CurrentAttribute].Instance[i] = CurrentInstances[i];
	CurrentAttribute++;
    };

UniqueSequenceNumber:
    LBRACE NUMBER CSTRING RBRACE {
	Qddb_InCoreEntry	*NewEntry;

	NewEntry = (Qddb_InCoreEntry *)Malloc(sizeof(Qddb_InCoreEntry)*MAXIMUM_SCHEMA_ENTRY);
	CurrentEntries = NewEntry;
	CurrentEntriesLength = 0;
	CurrentAttribute = 0;
	NewEntry[0].SequenceNumber = atoi($3);
	NewEntry[0].BytesOfData = 0;
	NewEntry[0].AttributeNumber = 0;
    };
%%

static int TCL_FindAttribute(string)
    char		*string;
{
    unsigned int	i, j;

    for (i = 1; i <= CurrentSchema->NumberOfAttributes; i++) {
	if (strcmp(string, CurrentSchema->Entries[i].Name) == 0) {
	    /* this is a possibility.. */
	    if (CurrentLevel != CurrentSchema->Entries[i].Level) {
		continue;
	    }
	    /* Check out the ancestors */
	    if (CurrentSchema->Entries[i].Level != CurrentLevel)
		continue;
	    for (j = 0; j < CurrentLevel; j++) {
		if (CurrentSchema->Entries[i].AncestorNumber[j] != 
		    CurrentAncestors[j]) {
		    break;
		}
	    }
	    if (j != CurrentLevel) {
		continue;
	    }
	    return i;
	}
    }
    return -1;
}

void Qddb_ResetTCLParser()
{
    if (CurrentEntries != NULL) {
	Free(CurrentEntries);
	CurrentEntries = NULL;
    }
    CurrentAttribute = 0;
    CurrentLevel = 1;
    CurrentInstances[0] = 0;
    CurrentAncestors[0] = 0;
    Qddb_ResetTCLLex();
}
