
/* $RCSfile: Time.c,v $ $Revision: 2.1 $ $Date: 2003/03/23 17:16:54 $
 *
 * Copyright (C) 1993-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "Qddb.h"

int Qddb_CheckTime(time)
    CONST char              *time;
{
    CONST char              *ptr;
    int                     colon_count;

    for (ptr = time, colon_count = 0; *ptr != '\0'; ptr++) {
        if (!isdigit(*ptr) && *ptr != ':')
	    return -1;
        if (*ptr == ':')
            colon_count++;
    }
    if (colon_count > 2)
        return -1;
    return 0;
}

int Qddb_TimeStringToSeconds(time)
    CONST char              *time;
{
    char                    *hours, *minutes, *seconds, *ptr, *ntime;
    int                     state, result;

    if (time == NULL)
        return 0;
    ntime = Malloc(strlen(time)+1);
    strcpy(ntime, time);
    hours = ptr = ntime;
    minutes = seconds = NULL;
    state = 0;
    while (*ptr != '\0') {
       if (*ptr == ':') {
           *ptr = '\0';
	   switch (state) {
	   case 0:
	     minutes = ptr + 1;
	     state++;
	     break;
	   case 1:
	     seconds = ptr + 1;
	     state++;
	     break;
	   default:
	     /* too many colons */
	     Free(ntime);
	     return -1;
	   }
       }
       ptr++;
    }
    result = 0;
    if (hours != NULL && strlen(hours) > 0) {
       while (*hours == '0') hours++;
       result += strtol(hours, NULL, 10) * 60 * 60;
    }
    if (minutes != NULL && strlen(minutes) > 0) {
       while (*minutes == '0') minutes++;
       result += strtol(minutes, NULL, 10) * 60;
    }
    if (seconds != NULL && strlen(seconds) > 0) {
       while (*seconds == '0') seconds++;
       result += strtol(seconds, NULL, 10);
    }
    Free(ntime);
    return result;
}


char *Qddb_SecondsToTimeString(time, format)
    int                     time;
    CONST char              *format;
{
    int                     hours, minutes, seconds;
    char                    *result;

    if (time < 0)
       return NULL;
    seconds = time % 60;
    minutes = (time / 60) % 60;
    hours   = time / 3600;
    result = Malloc(MAXTIMELEN);
    if (strcmp(format, "%H") == 0) {
      sprintf(result, "%02d", hours + ((minutes + (seconds >= 30 ? 1:0)) >= 30? 1:0));
    } else if (strcmp(format, "%M") == 0) {
      sprintf(result, "%02d", hours*60 + minutes + (seconds >= 30 ? 1:0));
    } else if (strcmp(format, "%S") == 0) {
      sprintf(result, "%02d", time);
    } else if (strcmp(format, "%H:%M") == 0) {
      sprintf(result, "%02d:%02d", hours, minutes + (seconds >= 30 ? 1:0));
    } else if (strcmp(format, "%M:%S") == 0) {
      sprintf(result, "%02d:%02d", hours*60 + minutes, seconds);
    } else if (strcmp(format, "%H:%M:%S") == 0) {
      sprintf(result, "%02d:%02d:%02d", hours, minutes, seconds);
    } else {
      sprintf(result, "%02d:%02d:%02d", hours, minutes, seconds);
    }
    return result;
}






