
/* $RCSfile: Tree.c,v $ $Revision: 2.1 $ $Date: 2003/03/23 17:16:54 $
 *
 * Copyright (C) 1993-2000 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "Qddb.h"

/* EXPORTED:
 *	void *Qddb_DataTreeProcess(Qddb_Schema *, Qddb_DataTree **, void *, int, int);
 *
 *
 *
 */
static Qddb_DataTree **DataTreeNewInstance _ANSI_ARGS_((Qddb_Schema *, Qddb_SchemaTreeNode **));
static void AddNodeToDataTree _ANSI_ARGS_((Qddb_Schema *, Qddb_DataTree **, Qddb_InCoreEntry *));
static void DataTreeSetSequenceNumber _ANSI_ARGS_((Qddb_DataTree **, int));
static void DataTreeSetMark _ANSI_ARGS_((Qddb_DataTree **, int));
static void DataTreeSetAttr _ANSI_ARGS_((Qddb_DataTree **, char *, char *));
static char *TreeDataToString _ANSI_ARGS_((Qddb_DataTree *));
static char *GetDefaultValue _ANSI_ARGS_((Qddb_Schema *, Qddb_SchemaNode *));

void *Qddb_DataTreeProcess(schema, tree, args, op, flags)
     Qddb_Schema			*schema;
     Qddb_DataTree		**tree;
     CONST void			*args;
     int				op;
     int				flags;
{
  void			*retval = NULL;

  switch (op) {
  case QDDB_DATATREE_PROC_NEWINSTANCE:
    retval = DataTreeNewInstance(schema, (Qddb_SchemaTreeNode **)args);
    break;
  case QDDB_DATATREE_PROC_ADDNODE:
    AddNodeToDataTree(schema, tree, (Qddb_InCoreEntry *)args);
    break;
  case QDDB_DATATREE_PROC_SETMARK:
    DataTreeSetMark(tree, *(int *)args);
    break;
  case QDDB_DATATREE_PROC_SETATTR:
    DataTreeSetAttr(tree, "", (char *)args);
    break;
  case QDDB_DATATREE_PROC_SETSEQ:
    DataTreeSetSequenceNumber(tree, *(int *)args);
    break;
  case QDDB_DATATREE_PROC_GETDATA:
    retval = TreeDataToString((Qddb_DataTree *)args);
    break;
  default:
    ;
  }
  return retval;
}

static void AddNodeToDataTree(schema, tree, entry)
     Qddb_Schema			*schema;
     Qddb_DataTree		**tree;
     Qddb_InCoreEntry		*entry;
{
  Qddb_SchemaNode		*schema_entry;
  Qddb_DataTree		**where, *whereinst, **placeinst;
  size_t			maxinst;
  int				i, j, level;
    
  /* Assume that the entry is valid and follow the attribute down the DataTree
   * until a point is reached where a new instance is required.   We assume
   * that at least the first instance has been built.
   */
  schema_entry = schema->Entries+entry->AttributeNumber;
  level = schema_entry->Level;
  where = tree;
  for (i = 1; i <= level; i++) {
#if defined(DIAGNOSTIC)
    if (where == NULL) {
      fprintf(stderr, "AddNodeToDataTree: premature where == NULL\n");
      fflush(stderr);
    }
#endif
    if (i == level) {
#if defined(DIAGNOSTIC)
      for (j = 0; where[j] != NULL; j++);
      if (schema_entry->Number-1 > j) {
	fprintf(stderr, "1. Uh, there aren't enough pieces here\n");
	fflush(stderr);
      }
#endif
      placeinst = where+schema_entry->Number-1;
      whereinst = *placeinst;
    } else {
#if defined(DIAGNOSTIC)
      for (j = 0; where[j] != NULL; j++);
      if (schema_entry->AncestorNumber[i]-1 > j) {
	fprintf(stderr, "2. Uh, there aren't enough pieces here\n");
	fflush(stderr);
      }
#endif
      placeinst = where+schema_entry->AncestorNumber[i]-1;
      whereinst = *placeinst;
    }
    for (maxinst = 0; whereinst[maxinst].datatree_type != DATATREE_END; maxinst++);
    if (entry->Instance[i-1] > maxinst) {
      /* reallocate instance */
      if (entry->Instance[i-1] > maxinst+1) {
	int		newmax;
	char		*default_value;

	newmax = entry->Instance[i-1] - 1;
	whereinst = (Qddb_DataTree *)Realloc(whereinst, sizeof(Qddb_DataTree)*(newmax+2));
	if (whereinst[0].datatree_schema->schemanode->IsLeaf == True) {
	  for (j = maxinst; j < newmax; j++) {
	    whereinst[j].datatree_schema = whereinst[0].datatree_schema;
	    /* Recal default_value for autoincrement, etc. */
	    default_value = GetDefaultValue(schema, whereinst[j].datatree_schema->schemanode);
	    switch (whereinst[j].datatree_schema->schemanode->Type) {
	    case QAT_Real:
	      if (default_value[0] == '\0') {
		whereinst[j].modified = False;
	      } else {
		whereinst[j].datatree_real = strtod(default_value, NULL);
		whereinst[j].modified = True;
	      }
	      whereinst[j].datatree_type = DATATREE_REAL;
	      break;
	    case QAT_Integer:
	      if (default_value[0] == '\0') {
		whereinst[j].modified = False;
	      } else {
		whereinst[j].datatree_int = strtol(default_value, NULL, 0);
		whereinst[j].modified = True;
	      }
	      whereinst[j].datatree_type = DATATREE_INT;
	      break;
	    case QAT_Date:
	      whereinst[j].datatree_date = Malloc(strlen(default_value)+1);
	      strcpy(whereinst[j].datatree_date, default_value);
	      whereinst[j].modified = True;
	      whereinst[j].datatree_type = DATATREE_DATE;
	      break;
	    case QAT_Time:
	      whereinst[j].datatree_time = Malloc(strlen(default_value)+1);
	      strcpy(whereinst[j].datatree_time, default_value);
	      whereinst[j].modified = True;
	      whereinst[j].datatree_type = DATATREE_TIME;
	      break;
	    case QAT_String:
	    default:
	      whereinst[j].datatree_string = Malloc(strlen(default_value)+1);
	      strcpy(whereinst[j].datatree_string, default_value);
	      whereinst[j].modified = True;
	      whereinst[j].datatree_type = DATATREE_STRING;
	      break;
	    }
	    whereinst[j].marked = Inapplicable;
	  }
	} else {
	  for (j = maxinst; j < newmax; j++) {
	    whereinst[j].datatree_type = DATATREE_CHILDREN;
	    whereinst[j].datatree_schema = whereinst[0].datatree_schema;
	    whereinst[j].datatree_children =
	      DataTreeNewInstance(schema, whereinst[0].datatree_schema->children);
	  }
	}
	maxinst = entry->Instance[i-1] - 1;
      } else {
	whereinst = (Qddb_DataTree *)Realloc(whereinst, sizeof(Qddb_DataTree)*(maxinst+2));
      }
      *placeinst = whereinst;
      whereinst[maxinst].datatree_schema = whereinst[0].datatree_schema;
      if (whereinst[0].datatree_schema->schemanode->IsLeaf == True) {
#if defined(DIAGNOSTIC)
	if (whereinst[maxinst].datatree_schema->schemanode->IsLeaf == False)
	  PANIC("*** ISLEAF IS FALSE ***");
	if (i != level)
	  PANIC("*** BAD LEVEL, ISLEAF IS TRUE ***");
#endif
	switch (whereinst[maxinst].datatree_schema->schemanode->Type) {
	case QAT_Real:
	  whereinst[maxinst].datatree_type = DATATREE_REAL;
	  if (entry->Data[0] == '\0') {
	    whereinst[maxinst].datatree_real = 0.0;
	    whereinst[maxinst].modified = False;
	  } else {
	    whereinst[maxinst].datatree_real = strtod(entry->Data, NULL);
	    whereinst[maxinst].modified = True;
	  }
	  break;
	case QAT_Integer:
	  whereinst[maxinst].datatree_type = DATATREE_INT;
	  if (entry->Data[0] == '\0') {
	    whereinst[maxinst].datatree_int = 0;
	    whereinst[maxinst].modified = False;
	  } else {
	    whereinst[maxinst].datatree_int = Qddb_GetInt(entry->Data);
	    whereinst[maxinst].modified = True;
	  }
	  break;
	case QAT_Date:
	  whereinst[maxinst].datatree_type = DATATREE_DATE;
	  whereinst[maxinst].datatree_date = Malloc(strlen(entry->Data)+1);
	  strcpy(whereinst[maxinst].datatree_date, entry->Data);
	  whereinst[maxinst].modified = True;
	  break;
	case QAT_Time:
	  whereinst[maxinst].datatree_type = DATATREE_TIME;
	  whereinst[maxinst].datatree_time = Malloc(strlen(entry->Data)+1);
	  strcpy(whereinst[maxinst].datatree_time, entry->Data);
	  whereinst[maxinst].modified = True;
	  break;
	case QAT_String:
	default:
	  whereinst[maxinst].datatree_type = DATATREE_STRING;
	  whereinst[maxinst].datatree_string = Malloc(strlen(entry->Data)+1);
	  strcpy(whereinst[maxinst].datatree_string, entry->Data);
	  whereinst[maxinst].modified = True;
	  break;
	}
	whereinst[maxinst].marked = entry->Marked;
      } else {
	whereinst[maxinst].datatree_type = DATATREE_CHILDREN;
	whereinst[maxinst].datatree_children =
	  DataTreeNewInstance(schema, whereinst[maxinst].datatree_schema->children);
      }
      whereinst[maxinst+1].datatree_type = DATATREE_END;
      if (whereinst[maxinst].datatree_type == DATATREE_CHILDREN)
	where = whereinst[maxinst].datatree_children;
      else
	where = NULL;
    } else {
      int				myinst = entry->Instance[i-1] - 1;

      /* Use an existing instance */
      if ((whereinst[myinst].datatree_type != DATATREE_CHILDREN) && i == level) {
	myinst = entry->Instance[i-1]-1;
	if (myinst < 0)
	  myinst = 0;
	switch (whereinst[myinst].datatree_schema->schemanode->Type) {
	case QAT_Real:
	  if (entry->Data[0] == '\0') {
	    whereinst[myinst].datatree_real = 0.00;
	    whereinst[myinst].modified = False;
	  } else {
	    whereinst[myinst].datatree_real = strtod(entry->Data, NULL);
	    whereinst[myinst].modified = True;
	  }
	  whereinst[myinst].datatree_type = DATATREE_REAL;
	  break;
	case QAT_Integer:
	  if (entry->Data[0] == '\0') {
	    whereinst[myinst].datatree_int = 0;
	    whereinst[myinst].modified = False;
	  } else {
	    whereinst[myinst].datatree_int = Qddb_GetInt(entry->Data);
	    whereinst[myinst].modified = True;
	  }
	  whereinst[myinst].datatree_type = DATATREE_INT;
	  break;
	case QAT_Date:
	  whereinst[myinst].datatree_type = DATATREE_DATE;
	  if (whereinst[myinst].datatree_date != NULL)
	    Free(whereinst[myinst].datatree_date);
	  whereinst[myinst].datatree_date = Malloc(strlen(entry->Data)+1);
	  strcpy(whereinst[myinst].datatree_date, entry->Data);
	  whereinst[myinst].modified = True;
	  break;
	case QAT_Time:
	  whereinst[myinst].datatree_type = DATATREE_TIME;
	  if (whereinst[myinst].datatree_time != NULL)
	    Free(whereinst[myinst].datatree_time);
	  whereinst[myinst].datatree_time = Malloc(strlen(entry->Data)+1);
	  strcpy(whereinst[myinst].datatree_time, entry->Data);
	  whereinst[myinst].modified = True;
	  break;
	case QAT_String:
	default:
	  whereinst[myinst].datatree_type = DATATREE_STRING;
	  if (whereinst[myinst].datatree_string != NULL)
	    Free(whereinst[myinst].datatree_string);
	  whereinst[myinst].datatree_string = Malloc(strlen(entry->Data)+1);
	  strcpy(whereinst[myinst].datatree_string, entry->Data);
	  whereinst[myinst].modified = True;
	  break;
	}
	whereinst[myinst].marked = entry->Marked;
	where = NULL;
      } else {
	if (whereinst[myinst].datatree_type == DATATREE_CHILDREN)
	  where = whereinst[myinst].datatree_children;
	else
	  abort(); /* where = NULL; */
      }
    }
  }
}

static Qddb_DataTree **DataTreeNewInstance(schema, schema_tree)
     Qddb_Schema                 *schema;
     Qddb_SchemaTreeNode		**schema_tree;
{
  Qddb_DataTree		**retval, **children;
  int				i, len;
  char			*default_value;

  if (schema_tree == NULL || *schema_tree == NULL) {
    return NULL;
  }
  for (len = 0; schema_tree[len] != NULL; len++);
  len++;
  retval = (Qddb_DataTree **)Calloc(sizeof(Qddb_DataTree *)*(len+1));
  for (i = 0; schema_tree[i] != NULL; i++) {
#if defined(DIAGNOSTIC)
    if (i > len)
      abort();
#endif
    retval[i] = (Qddb_DataTree *)Calloc(sizeof(Qddb_DataTree)*2); /* new instances have two nodes */
    retval[i][0].datatree_schema = schema_tree[i];
    children = DataTreeNewInstance(schema, schema_tree[i]->children);
    retval[i][0].datatree_sequence_number = -1;
    if (children != NULL) {
      retval[i][0].datatree_type = DATATREE_CHILDREN;
      retval[i][0].datatree_children = children;
    } else {
      /* Recal default_value for autoincrement, etc. */
      default_value = GetDefaultValue(schema, schema_tree[i]->schemanode);
      switch (schema_tree[i]->schemanode->Type) {
      case QAT_Real:
	retval[i][0].datatree_type = DATATREE_REAL;
	if (default_value[0] == '\0') {
	  retval[i][0].modified = False;
	  retval[i][0].datatree_real = 0.0;
	} else {
	  retval[i][0].datatree_real = strtod(default_value, NULL);
	  retval[i][0].modified = True;
	}
	break;
      case QAT_Integer:
	retval[i][0].datatree_type = DATATREE_INT;
	retval[i][0].datatree_int = 0;
	if (default_value[0] == '\0') {
	  retval[i][0].modified = False;
	  retval[i][0].datatree_int = 0;
	} else {
	  retval[i][0].datatree_int = strtol(default_value, NULL, 0);
	  retval[i][0].modified = True;
	}
	break;
      case QAT_Date:
	retval[i][0].datatree_type = DATATREE_DATE;
	if (default_value[0] == '\0') {
	  retval[i][0].datatree_date = Calloc(1);
	} else {
	  retval[i][0].datatree_date = Malloc(strlen(default_value)+1);
	  strcpy(retval[i][0].datatree_date, default_value);
	}
	retval[i][0].modified = True;
	break;
      case QAT_Time:
	retval[i][0].datatree_type = DATATREE_TIME;
	if (default_value[0] == '\0') {
	  retval[i][0].datatree_time = Calloc(1);
	} else {
	  retval[i][0].datatree_time = Malloc(strlen(default_value)+1);
	  strcpy(retval[i][0].datatree_time, default_value);
	}
	retval[i][0].modified = True;
	break;
      case QAT_String:
      default:
	retval[i][0].datatree_type = DATATREE_STRING;
	if (default_value[0] == '\0') {
	  retval[i][0].datatree_date = Calloc(1);
	} else {
	  retval[i][0].datatree_string = Malloc(strlen(default_value)+1);
	  strcpy(retval[i][0].datatree_string, default_value);
	}
	retval[i][0].modified = True;
	break;
      }
      retval[i][0].marked = Inapplicable;
    }
    retval[i][1].datatree_type = DATATREE_END;
  }
  return retval;
}

static void DataTreeSetSequenceNumber(tree, seq)
    Qddb_DataTree		**tree;
    int				seq;
{
    int				i, j;

    for (i = 0; tree[i] != NULL; i++) {
	for (j = 0; tree[i][j].datatree_type != DATATREE_END; j++) {
	    tree[i][j].datatree_sequence_number = seq;
	    if (tree[i]->datatree_schema->schemanode->IsLeaf == False) {
		DataTreeSetSequenceNumber(tree[i][j].datatree_children, seq);
	    }
	}
    }
}

static void DataTreeSetMark(tree, mark)
    Qddb_DataTree		**tree;
    int				mark;
{
    int				i, j;

    for (i = 0; tree[i] != NULL; i++) {
	for (j = 0; tree[i][j].datatree_type != DATATREE_END; j++) {
	    tree[i][j].marked = mark;
	    if (tree[i]->datatree_schema->schemanode->IsLeaf == False) {
		DataTreeSetMark(tree[i][j].datatree_children, mark);
	    }
	}
    }
}

static void DataTreeSetAttr(tree, path, attr)
    Qddb_DataTree		**tree;
    char			*path, *attr;
{
    int				i, j, isleaf;
    char			newpath[BUFSIZ];


    for (i = 0; tree[i] != NULL; i++) { /* attributes */
	isleaf = tree[i]->datatree_schema->schemanode->IsLeaf;
	strcpy(newpath, path);
	if (path[0] != '\0')
	    strcat(newpath, ".");
	strcat(newpath, tree[i]->datatree_schema->schemanode->Name);
	for (j = 0; tree[i][j].datatree_type != DATATREE_END; j++) { /* instances */
	    if (isleaf == False) {
		DataTreeSetAttr(tree[i][j].datatree_children, newpath, attr);
	    } else if (strcmp(newpath, attr) == 0) {
		if (tree[i][j].marked != True)
		    tree[i][j].marked = False;
	    }
	}
    }
}


static char *TreeDataToString(tree_entry)
    Qddb_DataTree		*tree_entry;
{
    char			*retval;
    char			buf[BUFSIZ];

    switch(tree_entry->datatree_type) {
    case DATATREE_STRING:
        if (tree_entry->datatree_string == NULL) {
	    retval = Calloc(1);
	} else {
	    retval = Malloc(strlen(tree_entry->datatree_string)+1);
	    strcpy(retval, tree_entry->datatree_string);
	}
	break;
    case DATATREE_DATE:
	retval = Malloc(strlen(tree_entry->datatree_date)+1);
	strcpy(retval, tree_entry->datatree_date);
	break;
    case DATATREE_TIME:
	retval = Malloc(strlen(tree_entry->datatree_time)+1);
	strcpy(retval, tree_entry->datatree_time);
	break;
    case DATATREE_INT:
	if (tree_entry->modified == False) {
	    retval = Calloc(1);
	} else {
	    sprintf(buf, tree_entry->datatree_schema->schemanode->Format, tree_entry->datatree_int);
	    retval = Malloc(strlen(buf)+1);
	    strcpy(retval, buf);
	}
	break;
    case DATATREE_REAL:
	if (tree_entry->modified == False) {
	    retval = Calloc(1);
	} else {
	    sprintf(buf, tree_entry->datatree_schema->schemanode->Format, tree_entry->datatree_real);
	    retval = Malloc(strlen(buf)+1);
	    strcpy(retval, buf);
	}
	break;
    case DATATREE_NOVAL:
    default:
	retval = Calloc(1);
    }
    return retval;
}

/* GetDefaultValue - return either default value or autoinc
 * (NOTE: return value is static, do not attempt to write it!)
 * NOTHREADSAFE
 */
static char *GetDefaultValue(schema, node)
    Qddb_Schema              *schema;
    Qddb_SchemaNode          *node;
{
    int                      fd;
    static char              buf[MAXINTEGERLEN];
    char                     path[MAXPATHLEN+32], *attribute;
    off_t                    size_of_file;

    if ((node->Options & QAO_Autoincrement) != 0) {
        sprintf(path, "%s/%s", schema->RelationName, QDDB_AUTOINC_FILENAME);
        if ((fd = open(path, O_RDWR|O_CREAT, 0666)) != -1) {
	    attribute = Qddb_NodeToAttribute(schema, node);
	    size_of_file = SizeOfFile(fd);
	    if (Qddb_LockSection(fd, F_WRLCK, (off_t)0, size_of_file, True) == -1) {
		return "0";
	    }
	    Qddb_AutoIncrement(fd, buf, MAXINTEGERLEN, attribute, node->AutoStart, node->AutoIncrement);
	    if (Qddb_UnlockSection(fd, (off_t)0, size_of_file) == -1) {
		return "0";
	    }
	    close(fd);
	    return buf;
	}
    }
    return node->DefaultValue;
}
