'\"
'\" Copyright (C) 1997-2001 Herrin Software Development, Inc.
'\" All rights reserved.
'\" 
'\" This program is free software; you can redistribute it and/or
'\" modify it under the terms of the GNU General Public License
'\" as published by the Free Software Foundation; either version 2
'\" of the License, or (at your option) any later version.
'\" 
'\" This program is distributed in the hope that it will be useful,
'\" but WITHOUT ANY WARRANTY; without even the implied warranty of
'\" MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'\" GNU General Public License for more details.
'\" 
'\" You should have received a copy of the GNU General Public License
'\" along with this program; if not, write to the Free Software
'\" Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
.TH qddb_script 1 "21 Oct 2000" "QDDB Reference Manual" "Herrin Software Development, Inc."
.DA 21 Oct 2000
.SH NAME
qddb_script \- \fIQddbScript\fR automatic Fx interface generation
.SH SYNOPSIS
.ft B
.nf
qddb_script autogen <schema_desc>
qddb_script compile <schema_variable_name> <script>

.SH DESCRIPTION
\fIqddb_script\fR serves as a very fast and convenient way to 
generate custom Qddb forms.   \fIautogen\fR generates a simple
\fIQddbScript\fR description of the given schema 
that can be easily modified by
any user.   \fIcompile\fR compiles a \fIQddbScript\fR form
description into executable Fx code.

.TP
\fIautogen\fR
accepts one extra argument, the
schema descriptor as returned from a call to
\fIqddb_schema open\fR.   The return value is
a string consisting of a valid \fIQddbScript\fR
form, suitable for passing directly to
\fIqddb_script compile\fR.

.TP
\fIcompile\fR
accepts two arguments: 1) a valid variable name (not
the value) in the global context containing
a valid schema descriptor as returned from
\fIqddb_schema open\fR, and 2) a string consisting of a
\fIQddbScript\fR form.   The \fIQddbScript\fR
is translated directly into Fx code suitable
for use with Tcl eval(n).

.SH QDDBSCRIPT
The \fIQddbScript\fR form description language
is similar to HTML, but allows more precise layout
of fields.   Code generated by \fIqddb_script compile\fR
is based on the Fx toolkit (Tcl/Tk).   All \fIQddbScript\fR
commands are of the following form:

.nf
    [keyword options...]
        [keyword options...]
        [end keyword]
    [end keyword]
.fi

where \fIkeyword\fR is one of the commands declared below.
The \fIoptions\fR field is a collection of options
that have one of the following forms: 1) option="value",
or 2) optionname.   Options that are not understood
by QddbScript are ignored.  For example, a valid 
QddbScript might look like this:

.nf
    # this is a comment
    [prologue]
    [menubar]
    [frame attr=Address]
        [text attr=Address.Street labelwidth=10] [break]
        [entry attr=Address.City labelwidth=10]
        [entry attr=Address.State width=2 noexpand]
        [entry attr=Address.ZipCode width=10 noexpand]
    [end frame]
    [epilogue]
.fi


.PP
.TP 4
\fBframe\fR
provides a frame that encapsulates a set of entries in rows.
Each frame may have many entries (entry, text, radiobutton, etc.)
and the rows are separated by the \fB[break]\fR statement.
\fBframe\fR requires an end marker:

.nf
    [frame]
    ...
    [end frame]
.fi

.PP
.TP 12
\fIattr=attribute\fR
specifies that the frame is associated with a structured
attribute which may be expandable.  If the attribute
is expandable, Add/View/Del buttons are automatically
produced so the user can add new instances of the
attribute. (optional)
.TP 12
\fIrelief=string\fR
specifies the relief of the outer frame.   Must be one
of "groove", "flat", "raised", or "sunken".
.TP 12
\fIborder=integer\fR
specifies the width of the border around the outer frame in pixels.
.TP 12
\fIpadx=integer\fR
specifies the external horizontal padding for the frame.
.TP 12
\fIpady=integer\fR
specifies the external vertical padding for the frame.
.TP 12
\fIipadx=integer\fR
specifies the internal horizontal padding for the frame.
.TP 12
\fIipady=integer\fR
specifies the internal vertical padding for the frame.
.TP 12
\fInoexpand\fR
prevents expansion of the text box to fill the width
of the screen if there is room.   The default is to
expand the width to fill the screen.
.PP
.TP 4
\fBentry\fR
provides an entry box for attribute values.   \fBentry\fR
does not have an end marker:

.nf
    [entry attr=Address.Street labelwidth=10 width=20]
.fi

.PP
.TP 12
\fIattr=attribute\fR
specifies the database attribute to which the
entry is bound.  If the attribute
is expandable, Add/View/Del buttons are automatically
produced so the user can add new instances of the
attribute. (required)
.TP 12
\fIside=string\fR
specifies the side within the current frame where
the new element should reside.  Must be one of
"left", "right", "top", or "bottom".
.TP 12
\fIlabelwidth=integer\fR
specifies the width of the entry's label in characters.
The default is the exact width of the label's contents.
.TP 12
\fIlabel="string"\fR
specifies that the default verbosename of the attribute
should be overriden with the value of "string".  The default
label is the verbosename of the attribute or the last
dot-separated component of the attribute's name.
.TP 12
\fIwidth=integer\fR
specifies the width of the entry box in characters.
The default is 10.
.TP 12
\fIrelief=string\fR
specifies the relief of the outer frame.   Must be one
of "groove", "flat", "raised", or "sunken".
.TP 12
\fIborder=integer\fR
specifies the border around the outer frame in pixels.
.TP 12
\fIpadx=integer\fR
specifies the external horizontal padding for the frame.
.TP 12
\fIpady=integer\fR
specifies the external vertical padding for the frame.
.TP 12
\fIipadx=integer\fR
specifies the internal horizontal padding for the frame.
.TP 12
\fIipady=integer\fR
specifies the internal vertical padding for the frame.
.TP 12
\fInoexpand\fR
prevents expansion of the entry box to fill the width
of the screen if there is room.   The default is to
expand the width to fill the screen.

.PP
.TP 4
\fBtext\fR
provides a full-text entry box for attribute values.
\fBtext\fR does not have an end marker:

.nf
    [text attr=Address.Street width=10 height=4]
.fi

.PP
\fIattr=attribute\fR
specifies the database attribute to which the
text box is bound.  If the attribute
is expandable, Add/View/Del buttons are automatically
produced so the user can add new instances of the
attribute. (required)
.TP 12
\fIside=string\fR
specifies the side within the current frame where
the new element should reside.  Must be one of
"left", "right", "top", or "bottom".
.TP 12
\fIlabelwidth=integer\fR
specifies the width of the text's label in characters.
The default is the exact width of the label's contents.
.TP 12
\fIlabel="string"\fR
specifies that the default verbosename of the attribute
should be overriden with the value of "string".   The default
label is the verbosename of the attribute or the last
dot-separated component of the attribute's name.
.TP 12
\fIwidth=integer\fR
specifies the width of the text box in characters.
The default is 10.
.TP 12
\fIheight=integer\fR
specifies the height of the text box in characters.
The default is 4.
.TP 12
\fIrelief=string\fR
specifies the relief of the outer frame.   Must be one
of "groove", "flat", "raised", or "sunken".
.TP 12
\fIborder=integer\fR
specifies the border around the outer frame in pixels.
.TP 12
\fIpadx=integer\fR
specifies the external horizontal padding for the frame.
.TP 12
\fIpady=integer\fR
specifies the external vertical padding for the frame.
.TP 12
\fIipadx=integer\fR
specifies the internal horizontal padding for the frame.
.TP 12
\fIipady=integer\fR
specifies the internal vertical padding for the frame.
.TP 12
\fInoexpand\fR
prevents expansion of the text box to fill the width
of the screen if there is room.   The default is to
expand the width to fill the screen.
.PP
.TP 4
\fBradiobutton\fR
provides a collection of radiobuttons for constrained
attribute values.
\fBradiobutton\fR does not have an end marker:

.nf
    [radiobutton attr=Address.State
        value="KY"
        value="IN" noexpand]
.fi

.TP 12
\fIattr=attribute\fR
specifies the database attribute to which the
text box is bound.  If the attribute
is expandable, Add/View/Del buttons are automatically
produced so the user can add new instances of the
attribute. (required)
.TP 12
\fIvalue="string"\fR
specifies the value of one radiobutton in the group.
You probably want multiple \fIvalue\fR options in
each radiobutton group.
.TP 12
\fIside=string\fR
specifies the side within the current frame where
the new element should reside.  Must be one of
"left", "right", "top", or "bottom".
.TP 12
\fIlabelwidth=integer\fR
specifies the width of the text's label in characters.
The default is the exact width of the label's contents.
.TP 12
\fIlabel="string"\fR
specifies that the default verbosename of the attribute
should be overriden with the value of "string".   The default
label is the verbosename of the attribute or the last
dot-separated component of the attribute's name.
.TP 12
\fIwidth=integer\fR
specifies the width of the text box containing the
radiobuttons. The default is to expand as necessary
to hold the radiobuttons.
.TP 12
\fIheight=integer\fR
specifies the height of the text box containing the
radiobuttons.
The default is 1.
.TP 12
\fIrelief=string\fR
specifies the relief of the outer frame.   Must be one
of "groove", "flat", "raised", or "sunken".
.TP 12
\fIborder=integer\fR
specifies the border around the outer frame in pixels.
.TP 12
\fIpadx=integer\fR
specifies the external horizontal padding for the frame.
.TP 12
\fIpady=integer\fR
specifies the external vertical padding for the frame.
.TP 12
\fIipadx=integer\fR
specifies the internal horizontal padding for the frame.
.TP 12
\fIipady=integer\fR
specifies the internal vertical padding for the frame.
.TP 12
\fInoexpand\fR
prevents expansion of the text box to fill the width
of the screen if there is room.   The default is to
expand the width to fill the screen.

.PP
.TP 4
\fBlabel\fR
is non-attribute based label that you can place anywhere
in a frame.   You can also use it as a place-holder
for alignment purposes.   \fBlabel\fR does not have an
end marker:

.nf
    [label value="Some label" width=10 height=2]
.fi

.TP 12
\fIvalue="string"\fR
specifies the text for the label.   The default is
the empty string ("").
.TP 12
\fIwidth=integer\fR
specifies the width of the label.
.TP 12
\fIheight=integer\fR
specifies the height of the label.
The default is 1.
.TP 12
\fIrelief=string\fR
specifies the relief of the outer frame.   Must be one
of "groove", "flat", "raised", or "sunken".
.TP 12
\fIborder=integer\fR
specifies the border around the outer frame in pixels.
.TP 12
\fIpadx=integer\fR
specifies the external horizontal padding for the frame.
.TP 12
\fIpady=integer\fR
specifies the external vertical padding for the frame.
.TP 12
\fIipadx=integer\fR
specifies the internal horizontal padding for the frame.
.TP 12
\fIipady=integer\fR
specifies the internal vertical padding for the frame.
.TP 12
\fInoexpand\fR
prevents expansion of the label to fill the width
of the screen if there is room.   The default is to
expand the width to fill the screen.


.PP
.TP 4
\fBnotebook\fR
defines a new notebook frame to hold notebook tabs
to partition the fields on the screen.   \fBnotebook\fR
requires an end marker:

.nf
    [notebook]
    ...
    [end notebook]
.fi

.PP
.TP 4
\fBtab\fR
defines a notebook tab for use with the \fBnotebook\fR
option.  The \fBtab\fR option must be between notebook
beginning and end markers as shown below.  All
QddbScript commands must be within tabs inside a notebook
section.   \fBtab\fR requires an end marker:

.nf
    [notebook]
        [tab label=Person underline=1]
        [entry attr=First]
        [entry attr=Last][break]
        [text attr=Comments]
        [end tab]
        [tab label=Address underline=2]
            [frame attr=Address]
                [text attr=Address.Street][break]
                [entry attr=Address.City]
                [entry attr=Address.State]
                [entry attr=Address.ZipCode][break]
            [end frame]
        [end tab]
    [end notebook]
.fi

.TP 12
\fIlabel="string"\fR
defines a label for use in the notebook tab.
.TP 12
\fIunderline=integer\fR
defines an underlined character position (beginning
with 0) for use as an accelerator.

.PP
.TP 4
\fBlabel\fR
builds a label in the current location with the
given attributes.
.TP 12
\fIside=string\fR
specifies the side within the current frame where
the new element should reside.  Must be one of
"left", "right", "top", or "bottom".
.TP 12
\fIlabelwidth=integer\fR
specifies the width of the entry's label in characters.
The default is the exact width of the label's contents.
.TP 12
\fIlabel="string"\fR
specifies that the default verbosename of the attribute
should be overriden with the value of "string".  The default
label is the verbosename of the attribute or the last
dot-separated component of the attribute's name.
.TP 12
\fIwidth=integer\fR
specifies the width of the label in characters.
The default is 10.
.TP 12
\fIrelief=string\fR
specifies the relief of the outer frame.   Must be one
of "groove", "flat", "raised", or "sunken".
.TP 12
\fIborder=integer\fR
specifies the border around the outer frame in pixels.
.TP 12
\fIpadx=integer\fR
specifies the external horizontal padding for the frame.
.TP 12
\fIpady=integer\fR
specifies the external vertical padding for the frame.
.TP 12
\fIipadx=integer\fR
specifies the internal horizontal padding for the frame.
.TP 12
\fIipady=integer\fR
specifies the internal vertical padding for the frame.
.TP 12
\fInoexpand\fR
prevents expansion of the entry box to fill the width
of the screen if there is room.   The default is to
expand the width to fill the screen.


.PP
.TP 4
\fBbreak\fR
instructs QddbScript to break to the next line for
further entries within the current frame.   [break]
has no options.

.SH SEE ALSO
qddb_instance(n), qddb_keylist(n),  qddb_rows(n), 
qddb_search(n), qddb_schema(n), 
qddb_tuple(n), qddb_view(n)

.SH BUGS
None known.

.SH REFERENCES
\fBQddbScript in a NutShell\fR
.nf
http://www.hsdi.com/qddb/qddbscript
.fi
