'\"
'\" Copyright (C) 1997-2001 Herrin Software Development, Inc.
'\" All rights reserved.
'\" 
'\" This program is free software; you can redistribute it and/or
'\" modify it under the terms of the GNU General Public License
'\" as published by the Free Software Foundation; either version 2
'\" of the License, or (at your option) any later version.
'\" 
'\" This program is distributed in the hope that it will be useful,
'\" but WITHOUT ANY WARRANTY; without even the implied warranty of
'\" MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'\" GNU General Public License for more details.
'\" 
'\" You should have received a copy of the GNU General Public License
'\" along with this program; if not, write to the Free Software
'\" Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
.TH qddb_util n "20 Oct 2000" "QDDB Reference Manual" "Herrin Software Development, Inc."
.DA 20 Oct 2000
.SH NAME
qddb_util \- Miscellaneous utility functions.
.SH SYNOPSIS
.ft B
.nf
qddb_util isdate <date string>
qddb_util istime <time string>
qddb_util postscript ?option value ...? string
qddb_util formatdate <format string> <date string>
qddb_util formattime <format string> <time string>
qddb_util datestyle ?us|european?
qddb_util isregexp <regexp string>
qddb_util flock <file desc> LOCK_SH|LOCK_EX|LOCK_NB|LOCK_UN 
qddb_util catrelation <schema desc> ?offset length?
qddb_util pr ?options? string 


.SH DESCRIPTION
.TP
\fIisdate\fR accepts a date string and determines whether
Qddb can parse it.   If it is a valid date string, 
\fIqddb_util isdate\fR returns \fB1\fR; \fB0\fR is returned
otherwise.

.TP
\fIistime\fR accepts a time string and determines whether
Qddb can parse it.   If it is a valid date string, 
\fIqddb_util isdate\fR returns \fB1\fR; \fB0\fR is returned
otherwise.

.TP
\fIpostscript\fR accepts an ASCII string and produces
POSTSCRIPT output.    The following options are valid:
.nf
   -lines            lines per page
   -cpi              characters per inch
   -pagelength       page length
   -baseline         XX(i|m|c|p) (inch|mm|cm|point)
   -verticalspacing  XX(i|m|c|p) (inch|mm|cm|point)
   -leftmargin       XX(i|m|c|p) (inch|mm|cm|point)
   -orient portrait|landscape
   -papertype letter|legal|a4

Example: qddb_util postscript -leftmargin 1i "Hello world"
.fi

.TP
\fIformatdate\fR accepts a format string (see strftime(3))
and a date string and returns the formatted string.   If
Qddb cannot parse the date string, an error is returned.
\fBNote\fR: the format string "%s" returns the number of seconds
since the epoch, and the format string "days1900" returns
the number of days since January 1, 1900.

.TP
\fIformattime\fR accepts a format string plus an input time string
and returns the resulting time string.   If
Qddb cannot parse the input string, an error is returned.
Valid format strings are currently "%H", "%M", "%S", 
"%H:%M", "%M:%S", and "%H:%M:%S".

.TP
\fIdatestyle\fR with no argument returns the current
date style.   With an argument of "us" or "european",
the date style is set appropriately (us is "%m/%d/%y"
and european is "%d/%m/%y").

.TP
\fIisregexp\fR accepts a string and determintes whether the
string is a valid V8-style regular expression.   If it is a
valid regular expression, \fIqddb_util isregexp\fR returns
\fB1\fR; \fB0\fR is returned otherwise.

.TP
\fIflock\fR locks or unlocks a file associated with the
given file descriptor.   The file descriptor must be
the result of a previous Tcl "open" command.   The options are a 
set of tokens separated by '|'.   If LOCK_SH is given, a shared
(read) lock is attempted.   If LOCK_EX is given, an exclusive
(write) lock is attempted.   The presence of LOCK_NB in
the options list indicates that the operation should not
block.   LOCK_UN unlocks the file if the lock exists.   For
example:
.nf
    set fd [open /tmp/ho.txt w]
    if {![qddb_util flock $fd LOCK_EX|LOCK_NB]} {
        puts "Warning: cannot set lock on '/tmp/ho.txt'"
    }
.fi

.TP
\fIcatrelation\fR returns a keylist descriptor that
refers to a keylist containing exactly one item for each
tuple in the database.  If the \fIoffset\fR and \fIlength\fR
are specified, the keylist refers to
those tuples numbered from \fIoffset\fR
through \fIoffset + length - 1\fR.

.TP
\fIpr\fR accepts a strings and produces paginated output.
.nf
   -dateformat <date format>
   -headers
   -footers
   -formfeed
   -linesperpage <integer>
.fi


.SH SEE ALSO
qddb_instance(n), qddb_keylist(n),  qddb_rows(n), qddb_schema(n), 
qddb_search(n), qddb_table(n), qddb_tuple(n), qddb_view(n)

.SH REFERENCES
.nf
A Guide to QDDB
Eric H. Herrin II and Raphael A. Finkel

Qddb User's Guide

An ASCII Database for Fast Queries of Relatively Stable Data
Eric H. Herrin II and Raphael A. Finkel
Computing Systems, Volume 4 Number 2
University of California Press, Berkeley CA

Schema and Tuple Trees: An Intuitive Structure for 
Representing Relational Data
Eric H. Herrin, II and Raphael A. Finkel
Computing Systems, Volume 9, Number 2
MIT Press, Cambridge MA

TCL and the TK Toolkit
John K. Ousterhout
Addison-Wesley, 1994
ISBN 0-201-63337-X
.fi
.SH DIAGNOSTICS

.SH BUGS
None known.
