
/* $RCSfile: qedit.c,v $ $Revision: 2.2 $ $Date: 2001/08/10 01:01:22 $
 *
 * Copyright (C) 1993-2001 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* Modifications contributed by Raphael A. Finkel
 */

#include "Qddb.h"

int PrintKeyList _ANSI_ARGS_((Qddb_Schema *, Qddb_KeyList *));
void EditKeyList _ANSI_ARGS_((Qddb_Schema *, Qddb_KeyList *));

void usage(char *progName) {
        fprintf(stderr,
                "Usage: %s l|e|d RelationDir [[-a | -o | -! |] [-f field] item ... ]\n",
                progName);
        fprintf(stderr, "\t-a: and (default)\n\t-o: or\n\t-!: excluding\n");
        fprintf(stderr, "\t-f: name of field (attribute) to restrict search\n");
        fprintf(stderr,
                "\titem = [SearchString | -r RegExp | -n Number |\n\t\t-nr Number-Number | -sr SearchString-SearchString]\n");
        fflush(stderr);
        exit(1);
}

/* the SHIFTARGS macro is used to carefully shift arguments expecting more */
#define SHIFTARGS(amount) { \
        argc-=(amount); \
        argv+=(amount); \
        if (argc <= 0) usage(progName); \
}

int main(argc, argv)
        int                 argc;
        char                *argv[];
{
        Qddb_Schema         *schema;
        char                Relation[BUFSIZ], *ch;
        char                combiner; /* a for AND, o for OR */
        Qddb_KeyList        *NewList = NULL, *LastList = NULL;
        size_t              Length;
        int                 firstTime;
        enum {
                EDIT, LIST, DELETE
        }                   DoWhat;
        Qddb_SearchArg      arg;
        char                *progName;
        
        progName = argv[0];
        if (argc < 4) {
	    usage(progName);
        }
        if (argv[1][0] == 'l')
	    DoWhat = LIST;
        else if (argv[1][0] == 'e')
	    DoWhat = EDIT;
        else if (argv[1][0] == 'd')
	    DoWhat = DELETE;
        else 
	    usage(progName);
        Qddb_Init();
        
        if ((ch=Qddb_FindRelation(argv[2])) == NULL) {
	    fprintf(stderr, "Cannot open relation %s\nPlease set QDDBDIRS.\n", argv[2]);
	    fflush(stderr);
	    exit(1);
        }
        strcpy(Relation, ch);
        schema = Qddb_InitSchema(Relation);
        if (schema == NULL) {
	    if (qddb_errmsg != NULL) {
		fprintf(stderr, "%s", qddb_errmsg);
	    } else {
		fprintf(stderr, "Unknown error '%d' while reading schema\n",
                        qddb_errno);
	    }
	    fflush(stderr);
	    exit(1);
        }
        SHIFTARGS(3);
        firstTime = 1;
        while (argc) { /* treat a specifier in argv[0 .. ] */
	    int fieldNumber;

	    combiner = 'a'; /* AND by default */
	    if (strcmp(argv[0], "-a") == 0) {
		SHIFTARGS(1);
	    } else if (strcmp(argv[0], "-o") == 0) {
		combiner = 'o'; /* OR */
		SHIFTARGS(1);
	    } else if (strcmp(argv[0], "-!") == 0) {
		combiner = '!'; /* exclusion  */
		SHIFTARGS(1);
	    }
	    if (strcmp(argv[0], "-f") == 0) { /* field-specific item */
		SHIFTARGS(1);
		fieldNumber = Qddb_ConvertAttributeNameToNumber(schema, argv[0]);
		if (fieldNumber < 0) {
		    fprintf(stderr, "No such field name: %s.\n", argv[0]);
		    exit(-1);
		}
		SHIFTARGS(1);
	    } else {
		fieldNumber = -1;
	    }
	    if (argc == 0) usage(progName);
	    if (strcmp(argv[0], "-r") != 0 && strcmp(argv[0], "-n") != 0 &&
		strcmp(argv[0], "-sr") != 0 && strcmp(argv[0], "-nr") != 0) {
		arg.Type = ALPHA;
		arg.Alpha = argv[0];
		NewList = Qddb_Search(schema, &arg, &Length, fieldNumber);
	    } else if (strcmp(argv[0], "-r") == 0) {
		SHIFTARGS(1);
		arg.Type = REGEXP_V8;
		arg.Alpha = argv[0];
		NewList = Qddb_Search(schema, &arg, &Length, fieldNumber);
	    } else if (strcmp(argv[0], "-nr") == 0) {
		char            *range;

		SHIFTARGS(1);
		arg.LowerExists = (argv[0][0] != '-');
		arg.UpperExists = (argv[0][strlen(argv[0])-1] != '-');
		range = argv[0];
		if (arg.LowerExists == True) {
		    int         index, len;

		    len = strlen(range);
		    for (index = 0; index < len; index++)
			if (range[index] == '-')
			    break;
		    argv[0][index] = '\0';
		    arg.LowerNum = strtod(argv[0], NULL);
		    range = argv[0]+index+1;
		} else {
		    arg.LowerNum = 0.0;
		    range++;
		}
		if (arg.UpperExists == True)
		    arg.UpperNum = strtod(range, NULL);
		else
		    arg.UpperNum = 0.0;
		arg.Type = NUMERICRANGE;
		NewList = Qddb_Search(schema, &arg, &Length, fieldNumber);
	    } else if (strcmp(argv[0], "-n") == 0) {
		SHIFTARGS(1);
		arg.Type = NUMERIC;
		arg.Number = strtod(argv[0], NULL);
		NewList = Qddb_Search(schema, &arg, &Length, fieldNumber);
	    } else if (strcmp(argv[0], "-sr") == 0) {
		char            *range, *tmpstr;
		unsigned int    z, len;

		SHIFTARGS(1);
		len = strlen(argv[0]);
		tmpstr = Malloc(len);
		strcpy(tmpstr, argv[0]);
		for (z = 0; z < len; z++)
		    if (_isupper(tmpstr[z]))
			tmpstr[z] = _tolower(tmpstr[z]);
		arg.LowerExists = (tmpstr[0] != '-');
		arg.UpperExists = (tmpstr[len-1] != '-');
		range = tmpstr;
		if (arg.LowerExists == True) {
		    int         index, len;

		    len = strlen(tmpstr);
		    for (index = 0; index < len; index++)
			if (tmpstr[index] == '-')
			    break;
		    tmpstr[index] = '\0';
		    arg.LowerStr = tmpstr;
		    range = tmpstr+index+1;
		} else {
		    arg.LowerStr = NULL;
		    range++;
		}
		if (arg.UpperExists == True)
		    arg.UpperStr = range;
		else
		    arg.UpperStr = NULL;
		NewList = Qddb_Search(schema, &arg, &Length, fieldNumber);
	    } else {
		fprintf(stderr, "Unknown option: %s\n", argv[0]);
		PANIC("Cannot handle option");
	    }
	    if (firstTime) {
		LastList = NewList;
		firstTime = 0;
	    } else {
		LastList = Qddb_KeyListOp(schema, LastList, NewList,
					  (combiner == 'a') ?
					  QDDB_KEYLIST_OP_INTERSECTION
					  : (combiner == 'o') ?
					  QDDB_KEYLIST_OP_UNION :
					  QDDB_KEYLIST_OP_EXCLUSION,
					  QDDB_KEYLIST_FLAG_DELDUP_SAMEENTRY);
	    }
	    argc--; argv++;
        } /* treat one specifier */
        LastList = Qddb_KeyListProcess(schema, LastList, NULL,
				       QDDB_KEYLIST_PROC_NULLOP, QDDB_KEYLIST_FLAG_DELDUP_SAMEENTRY);
        if (DoWhat == LIST)
	    PrintKeyList(schema, LastList);
        else if (DoWhat == EDIT)
	    EditKeyList(schema, LastList);
        else
	    Qddb_DeleteEntryByKeyList(schema, LastList);
        return(0);
} /* main */

int PrintKeyList(schema, list)
    Qddb_Schema         *schema;
    Qddb_KeyList        *list;
{
    char                *RelationFN = schema->RelationName, *output;
    Qddb_Entry          ThisEntry;
    int                 DBFile;

    DBFile = OpenDatabase(RelationFN, 0);
    ThisEntry = NULL;
    while (list != NULL) {
	if (Qddb_ReadEntryByKeyList(DBFile, RelationFN, &ThisEntry, list, True) == -1) {
	    list = list->next;
	    continue;
	}
	Qddb_ReducedAttrToFullAttr(schema, ThisEntry);
	output = (char *)Qddb_Convert(schema, QDDB_ENTRYTYPE_EXTERNAL, ThisEntry, QDDB_ENTRYTYPE_READABLE);
	fprintf(stdout, "%s\n", output);
	list = list->next;
    }
    close(DBFile);
    return 0;
} /* PrintKeyList */
