#!/usr/local/bin/qtclsh

proc clear_cell {table row col} {
	if {[catch "qddb_table cell clear $table $row $col" err] != 0} {
		puts "\t---cell $row,$col produced an error"
		puts "\t--->$err"
	}
}
set t [qddb_table define 3 10 \
	-variable table_variable -name table_name \
	-title "Table Title" -comment "Table Comment" \
	-order row]
qddb_table configure $t -order row
foreach i {-variable -name -title -comment -order} {
	puts "Table ${i}: [qddb_table cget $t $i]"
}
for {set i [qddb_table row maxnum $t]} {$i >= 1} {incr i -1} {
	set l {}
	for {set j [qddb_table col maxnum $t]} {$j >= 1} {incr j -1} {
		lappend l [expr $i * $j]
	}
	qddb_table row setval $t $i $l
}
for {set i 1} {$i <= [qddb_table row maxnum $t]} {incr i} {
	puts [qddb_table row getval $t $i]
}
set t1 [qddb_table define \
	[qddb_table col maxnum $t] [qddb_table row maxnum $t]]
for {set i 1} {$i <= [qddb_table row maxnum $t1]} {incr i} {
	qddb_table row setval $t1 $i [qddb_table col getval $t $i]
}
qddb_table row insert $t1 -before 3
qddb_table row setval $t1 3 {
	100 100 100
}
puts "\t---UNSORTED"
for {set i 1} {$i <= [qddb_table row maxnum $t1]} {incr i} {
	puts [qddb_table row getval $t1 $i]
}
qddb_table sort $t1 -sortby 1
puts "\t---SORTED"
for {set i 1} {$i <= [qddb_table row maxnum $t1]} {incr i} {
	puts [qddb_table row getval $t1 $i]
}
set t2 [qddb_table copy $t1]

clear_cell $t1 1 1
clear_cell $t1 2 3
clear_cell $t1 3 2
clear_cell $t1 4 2
clear_cell $t1 8 4
clear_cell $t1 8 2

puts "\t---CLEARED CELLS"
for {set i 1} {$i <= [qddb_table row maxnum $t1]} {incr i} {
	puts [qddb_table row getval $t1 $i]
}

puts "\t---COPIED TABLE (no cleared cells)"
for {set i 1} {$i <= [qddb_table row maxnum $t2]} {incr i} {
	puts [qddb_table row getval $t2 $i]
}

puts "\t---COPIED TABLE (after appending column 1)"
qddb_table col insert $t2
qddb_table col copy $t2 1 [qddb_table col maxnum $t2]
for {set i 1} {$i <= [qddb_table row maxnum $t2]} {incr i} {
	puts [qddb_table row getval $t2 $i]
}
puts "\t---COPIED TABLE (after appending rows 1 & 2, then prepending 3 & 4)"
qddb_table row insert $t2 -numrows 2 ;# append
qddb_table row copy $t2 1 [expr [qddb_table row maxnum $t2] - 1]
qddb_table row copy $t2 2 [qddb_table row maxnum $t2]
qddb_table row insert $t2 -before 1 -numrows 2 
qddb_table row copy $t2 5 1
qddb_table row copy $t2 6 2
for {set i 1} {$i <= [qddb_table row maxnum $t2]} {incr i} {
	puts [qddb_table row getval $t2 $i]
}

puts "\t---COPIED TABLE (after appending cols 1 & 2, then prepending 3 & 4)"
qddb_table col insert $t2 -numcols 2 ;# append
qddb_table col copy $t2 1 [expr [qddb_table col maxnum $t2] - 1]
qddb_table col copy $t2 2 [qddb_table col maxnum $t2]
qddb_table col insert $t2 -before 1 -numcols 2 ;# prepend 
qddb_table col copy $t2 5 1
qddb_table col copy $t2 6 2
for {set i 1} {$i <= [qddb_table row maxnum $t2]} {incr i} {
	puts [qddb_table row getval $t2 $i]
}

puts "\t---Using getval/setval for rows"
flush stdout
set val [qddb_table row getval $t2 1]
puts "got: $val"
qddb_table row insert $t2
set max [qddb_table row maxnum $t2]
puts "max: $max"
flush stdout
qddb_table row setval $t2 $max $val
puts "set: $val"
set val2 [qddb_table row getval $t2 $max]
if {[string compare $val $val2] != 0} {
	puts "\t--->>row values not the same"
}
puts "\t---Using getval/setval for columns"
flush stdout
set val [qddb_table col getval $t2 1]
puts "got: $val"
qddb_table col insert $t2
set max [qddb_table col maxnum $t2]
qddb_table col setval $t2 $max $val
puts "set: $val"
set val2 [qddb_table col getval $t2 $max]
if {[string compare $val $val2] != 0} {
	puts "\t--->>column values not the same"
}
set t3 [qddb_table copy $t2]
qddb_table clear $t3
for {set i 1} {$i <= [qddb_table row maxnum $t2]} {incr i} {
	lappend vals [qddb_table row getval $t2 $i]
}
qddb_table setval $t3 $vals
puts "\t---TABLE before setval"
for {set i 1} {$i <= [qddb_table row maxnum $t2]} {incr i} {
	puts [qddb_table row getval $t2 $i]
}
puts "\t---TABLE with setval"
for {set i 1} {$i <= [qddb_table row maxnum $t3]} {incr i} {
	puts [qddb_table row getval $t3 $i]
}

puts "\t---Checking for qddb_table cget"
qddb_table configure $t3 -title "TABLE TITLE" \
	-name "TABLE NAME" -variable "TABLE VARIABLE" -comment "TABLE COMMENT"
puts "\t\ttitle-----> [qddb_table cget $t3 -title]"
puts "\t\tname------> [qddb_table cget $t3 -name]"
puts "\t\tvariable--> [qddb_table cget $t3 -variable]"
puts "\t\tcomment---> [qddb_table cget $t3 -comment]"

puts "\t---Checking for qddb_table row cget"
qddb_table row configure $t3 1 -title "ROW TITLE" \
	-name "ROW NAME" -variable "ROW VARIABLE" -comment "ROW COMMENT"
puts "\t\ttitle-----> [qddb_table row cget $t3 1 -title]"
puts "\t\tname------> [qddb_table row cget $t3 1 -name]"
puts "\t\tvariable--> [qddb_table row cget $t3 1 -variable]"
puts "\t\tcomment---> [qddb_table row cget $t3 1 -comment]"
puts "\t---Checking for qddb_table col cget"
qddb_table col configure $t3 1 -title "COLUMN TITLE" \
	-name "COLUMN NAME" -variable "COLUMN VARIABLE" \
	-comment "COLUMN COMMENT" -type real \
	-justify left -width 10 -precision 4
puts "\t\ttitle-----> [qddb_table col cget $t3 1 -title]"
puts "\t\tname------> [qddb_table col cget $t3 1 -name]"
puts "\t\tvariable--> [qddb_table col cget $t3 1 -variable]"
puts "\t\tcomment---> [qddb_table col cget $t3 1 -comment]"
puts "\t\ttype------> [qddb_table col cget $t3 1 -type]"
puts "\t\tjustify---> [qddb_table col cget $t3 1 -justify]"
puts "\t\twidth-----> [qddb_table col cget $t3 1 -width]"
puts "\t\tprecision-> [qddb_table col cget $t3 1 -precision]"
puts "\t---Checking for qddb_table cell cget"
qddb_table cell configure $t3 1 1 -title "CELL TITLE" \
	-name "CELL NAME" -variable "CELL VARIABLE" \
	-comment "CELL COMMENT" -type real \
	-justify left -width 20 -precision 8
puts "\t\ttitle-----> [qddb_table cell cget $t3 1 1 -title]"
puts "\t\tname------> [qddb_table cell cget $t3 1 1 -name]"
puts "\t\tvariable--> [qddb_table cell cget $t3 1 1 -variable]"
puts "\t\tcomment---> [qddb_table cell cget $t3 1 1 -comment]"
puts "\t\ttype------> [qddb_table cell cget $t3 1 1 -type]"
puts "\t\tjustify---> [qddb_table cell cget $t3 1 1 -justify]"
puts "\t\twidth-----> [qddb_table cell cget $t3 1 1 -width]"
puts "\t\tprecision-> [qddb_table cell cget $t3 1 1 -precision]"

set num 4.12345678
qddb_table cell setval $t3 1 1 $num
puts "cell 1,1 (should be $num): [qddb_table cell getval $t3 1 1]"

puts "\t---TABLE after setval of cell 1,1"
for {set i 1} {$i <= [qddb_table row maxnum $t3]} {incr i} {
	puts [qddb_table row getval $t3 $i]
}
qddb_table sort $t3 -sortby {2 1} -ascending 1
puts "\t---TABLE after sorting"
for {set i 1} {$i <= [qddb_table row maxnum $t3]} {incr i} {
	puts [qddb_table row getval $t3 $i]
}
qddb_table cell configure $t3 1 1 -title "CELL TITLE" \
	-name "CELL NAME" -variable "CELL VARIABLE" \
	-comment "CELL COMMENT" -type date \
	-justify left -width 10 -precision 8
qddb_table cell setval $t3 1 1 10/2/95
puts "cell 1,1: [qddb_table cell getval $t3 1 1]"
qddb_table cell configure $t3 1 1 -title "CELL TITLE" \
	-name "CELL NAME" -variable "CELL VARIABLE" \
	-comment "CELL COMMENT" -type calc \
	-justify left -width 10 -precision 8
qddb_table cell setval $t3 1 1 "1.23"
puts "cell 1,1: [qddb_table cell getval $t3 1 1]"

puts "\t---TABLE from qddb_table getval"
foreach i [qddb_table getval $t3 -format list] {
	puts $i
}
set env(QDDBDIRS) [glob ~/Databases]
set s [qddb_schema open family]
set k [qddb_search $s -prunebyattr Person.Name.LastName w Finkel]
set k1 [qddb_search $s -prunebyattr Person.Name.FirstName w Raphael]
set k [qddb_keylist op intersection $k $k1]
set t [qddb_rows select -format table -flush on -rowdescs on \
	-attrs Person.Name.LastName \
	-print {Person.Name.FirstName Person.Name.LastName} $k]

set tup [lindex [qddb_table row cget $t 1 -comment] 0]
set tuptab [qddb_rows all -format table -sortby Person.Residences.Street \
	-ascending Person.Residences.Street \
	-print {Person.Residences.Street Person.Residences.City} \
	-attrs {Person.Residences.Street Person.Residences.City} $tup]
set tuprows [qddb_rows all -format rows -sortby Person.Residences.Street \
	-ascending Person.Residences.Street -instance Person.Residences,1\
	-print {Person.Residences.Street Person.Residences.City} \
	-attrs {Person.Residences.Street Person.Residences.City} $tup]
puts "\t---ROWS from qddb_rows all"
puts "$tuprows"
puts "\t---TABLE from qddb_table getval in rows from tuple $tup"
puts [qddb_table getval $tuptab -format text]

puts [qddb_table getval $t -format list -coltitles on -rowtitles off]
puts [qddb_table getval $t -format list -coltitles on -rowtitles on]
puts [qddb_table getval $t -format text -coltitles on -rowtitles on]
proc print_table {t} {
	puts "table: $t"
	set res [qddb_table getval $t -format rows -coltitles on -rowtitles on]
	foreach i $res {
		puts $i
	}
}
proc print_table_off {t} {
	puts "table: $t"
	set res [qddb_table getval $t -format rows -coltitles on \
		-rowtitles off]
	foreach i $res {
		puts $i
	}
}
print_table_off $t
print_table $t
qddb_table col configure $t Person.Name.FirstName -justify right
print_table_off $t
print_table $t
qddb_table col configure $t Person.Name.FirstName -justify right
qddb_table col configure $t Person.Name.LastName -justify right
print_table_off $t
print_table $t
qddb_table col configure $t Person.Name.FirstName -justify left
qddb_table col configure $t Person.Name.LastName -justify left
print_table_off $t
print_table $t
qddb_table cell setval $t 1 1 "hello\nworld"
qddb_table col configure $t Person.Name.FirstName -justify center -separator " | "
qddb_table col configure $t Person.Name.LastName -justify center -separator " | "
print_table_off $t
print_table $t
