#!/usr/local/bin/qtclsh
set tcl_precision 17

proc testrangefunc {t func} {
	set val [qddb_table expr @${func}($t,@range(1,1:@maxrow,@maxcol))]
	puts "$func (all cells): $val"
	set maxrow [qddb_table row maxnum $t]
	set maxcol [qddb_table col maxnum $t]
	for {set i 1} {$i <= $maxrow} {incr i} {
		set val [qddb_table expr @${func}($t,@row($i))]
		puts "$func (row $i): $val"
	}
	for {set i 1} {$i <= $maxcol} {incr i} {
		set val [qddb_table expr @${func}($t,@col($i))]
		puts "$func (column $i): $val"
	}
	set expr "@${func}($t"
	for {set i 1} {$i <= $maxrow} {incr i} {
		for {set j 1} {$j <= $maxcol} {incr j} {
			append expr ",@cell($i,$j)"
		}
	}
	append expr ")"
	set val [qddb_table expr $expr]
	puts "$func (all cells by cell): $val"
}

proc testnumericfunc1 {t func} {
	set val [qddb_table expr "@${func}(
		@sum($t,@range(1,1:@maxrow,@maxcol))
	)"]
	puts "${func}(@sum(@range(1,1:@maxrow,@maxcol))): $val"
}

proc testnumericfunc2 {t func} {
	set val1 [qddb_table expr "@sum($t,
		@range(1,1:@maxrow,@maxcol),@row(1),@col(1)
	)"]
	set val1 [expr $val1 + 0.12435212122]
	set val2 [qddb_table expr "@${func}($val1,4.0)"]
	puts "${func}($val1,2): $val2"
}

# table 1 definition
set t [qddb_table define 3 3]
qddb_table col configure $t 1 -type real
qddb_table col configure $t 2 -type real
qddb_table col configure $t 3 -type real
qddb_table configure $t -variable tbl
for {set i 1} {$i <= 3} {incr i} {
	for {set j 1} {$j <= 3} {incr j} {
		set tbl($i,$j) [expr $i * $j]
	}
}
# table 2 definition
set t2 [qddb_table define 3 3]
qddb_table col configure $t2 1 -type real
qddb_table col configure $t2 2 -type real
qddb_table col configure $t2 3 -type real
qddb_table configure $t2 -variable tbl2
for {set i 1} {$i <= 3} {incr i} {
	for {set j 1} {$j <= 3} {incr j} {
		set tbl2($i,$j) [expr $i * $j * $i * $j]
	}
}
puts "---- table 1 ----"
puts [qddb_table getval $t -format text]
puts "sum is [qddb_table expr @sum($t,@range(1,1:3,3))]"
puts "cellval(1,1) is [qddb_table expr @cellval($t,1,1)]"
puts "cellval(1,2) is [qddb_table expr @cellval($t,1,2)]"
puts "cellval(1,3) is [qddb_table expr @cellval($t,1,3)]"
puts "cellval(2,1) is [qddb_table expr @cellval($t,2,1)]"
puts "cellval(2,2) is [qddb_table expr @cellval($t,2,2)]"
puts "cellval(2,3) is [qddb_table expr @cellval($t,2,3)]"
puts "cellval(3,1) is [qddb_table expr @cellval($t,3,1)]"
puts "cellval(3,2) is [qddb_table expr @cellval($t,3,2)]"
puts "cellval(3,3) is [qddb_table expr @cellval($t,3,3)]"
puts "sum(@row(1),@row(2)) is [qddb_table expr @sum($t,@row(1),@row(2))]"
puts "sum(@col(2),@col(3)) is [qddb_table expr @sum($t,@col($t,2),@col($t2,3))]"

puts [qddb_table expr @sum($t,@range(@if(1=1,1,2),@if(1=1,1,2):3,3))]

puts [qddb_table expr \
@sum($t,@range($t2,@if(1=1,1,2),@if(1=1,1,2):3,3),@range(1,1:3,3))]

puts [qddb_table expr "@if(@cellval($t,1,1)=@cellval($t2,1,1),10,20)"]

puts ""
puts [qddb_table getval $t -format text]
puts "----"
puts [qddb_table getval $t2 -format text]

set sum [qddb_table expr "
    @sum($t, # set implicit table $t
        @range(1,1:@maxrow,@maxcol),
        # override implicit table with $t2
        @cell($t2,@maxrow,@maxcol)
    ) * 42
"]
puts "new sum expr is: $sum"
puts "---testing range functions on table 1---"
testrangefunc $t sum
testrangefunc $t prod
testrangefunc $t avg
testrangefunc $t stddev
testrangefunc $t count
testrangefunc $t min
testrangefunc $t max

puts "---testing numeric functions with one arg---"
testnumericfunc1 $t sqrt
testnumericfunc1 $t exp
testnumericfunc1 $t ln
testnumericfunc1 $t log
testnumericfunc1 $t floor
testnumericfunc1 $t ceil
testnumericfunc1 $t rnd
testnumericfunc1 $t abs
testnumericfunc1 $t degrees
testnumericfunc1 $t radians
testnumericfunc1 $t sin
testnumericfunc1 $t cos
testnumericfunc1 $t tan
testnumericfunc1 $t asin
testnumericfunc1 $t acos
testnumericfunc1 $t atan

puts "---testing numeric functions with two args---"
testnumericfunc2 $t pow
testnumericfunc2 $t round
testnumericfunc2 $t hypot
testnumericfunc2 $t atan2

puts "----"
puts [qddb_table getval $t -format text -rowtitles on -coltitles on]
qddb_table col clear $t 2
qddb_table col configure $t 2 -title "new col values"
qddb_table col configure $t 2 -type real -justify center -precision 2
qddb_table row clear $t 2
qddb_table row configure $t 2 -title "row 2"
puts "----"
puts [qddb_table getval $t -format text -rowtitles on -coltitles on]
qddb_table col configure $t 2 -type string -justify center -precision 2
puts "----"
puts [qddb_table getval $t -format text -rowtitles on -coltitles on]
qddb_table cell configure $t 2 1 -type integer -justify left -precision 2
qddb_table cell configure $t 2 2 -type string -justify center
set tbl(1,2) "row 1"
set tbl(2,2) "row 2"
set tbl(3,2) "row 3"
puts "----"
puts [qddb_table getval $t -format text -rowtitles on -coltitles on]
qddb_table col configure $t 1 -separator " | "
qddb_table col configure $t 2 -separator " | "
qddb_table col configure $t 3 -separator " | "
puts "----"
puts [qddb_table getval $t -format text -rowtitles on -coltitles on]

qddb_table col configure $t 2 -title "new col values"
qddb_table col configure $t 2 -type real -justify center -precision 2
puts "----"
puts [qddb_table getval $t -format text -rowtitles on -coltitles on]

exit 0
